#include <stdlib.h>
#include <stdio.h>

#include "PdfObjects.hpp"
#include "FontReader.hpp"
#include "pdfc_base.h"
#include "pdfc_basicobjects.h"
#include "stringclass.h"
#include "PDFTokenizer.hpp"
#include "ImageReader.hpp"
#include "PsCommands.hpp"

#define DEBUG_PROGRESS_STAGES false

int main(int argc, char **argv) {
	if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Building main objects\n");
	PdfRoot *root = new PdfRoot();
	PdfPage *page = new PdfPage();
	PdfResourceRef *fontRef = new PdfResourceRef();
	PdfFont *font = new PdfFont();
	PdfResourceRef *imageRef = new PdfResourceRef();
	PdfImage *image = new PdfImage();
	PdfBookmark *bookmark;

	PObj *obj;

	root->author = "Steve";
	root->title = "Test Document";
	root->subject = "Testing";
	root->keywords = "pdf, test";

	// Bookmarks
	if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Creating Bookmarks\n");
	root->showBookmarks = true;
	root->addBookmark("First Section", page);
	bookmark = new PdfBookmark("Second Section", page);
	root->addBookmark(bookmark);
	bookmark->addBookmark("First SubSection", page);
	root->addBookmark("Third Section", page);
	bookmark = new PdfBookmark("Fourth Section", page);
	bookmark->isOpen = true;
	root->addBookmark(bookmark);
	bookmark->addBookmark("Second SubSection", page);

	// Add Page
	if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Adding Page to root object\n");
	root->addPage(page);

	// Add Font
	if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Creating Font\n");
	font->id = 1;
	font->subType = "TrueType";
	font->filename = "vineritc.ttf";
	font->loadData();
	FontReader::readFont(font);
	root->addResource(font);

	fontRef->resource = font;
	fontRef->type = font->type;
	fontRef->name = ("Font1");
	page->addResourceRef(fontRef);

	// Add Image
	if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Creating Image\n");
	image->id = 2;
	image->format = "external";
	image->filename = "crab_vlt.jpg";

	ImageReader iReader;
	iReader.readImage(image);
	root->addResource(image);

	imageRef->resource = image;
	imageRef->type = image->type;
	imageRef->name = ("Image1");
	page->addResourceRef(imageRef);

	// Add Content -------------------------------------------------
	if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Adding Page Content\n");
	init_pscommands();

	page->addContent(PsCommands::set_strokecolorRGB(1, 0, 0));
	page->addContent(PsCommands::add_rectangle(4,4,604,784));
	page->addContent(PsCommands::strokepath());
	page->addContent(PsCommands::set_linewidth(8));
	page->addContent(PsCommands::moveto(8,680));
	page->addContent(PsCommands::lineto(604,680));
	page->addContent(PsCommands::strokepath());

	page->addContent(PsCommands::drawimage(new PDF_box(505,12,96,144), "Image1")); 

	// page->addContent(PsCommands::set_font_and_size("Font1", 48));
	page->addContent(PsCommands::drawstring("Font1", 48, new PDF_box(12,772,600,772), "Interlan Voodoo Grande"));

	char *text = "\
This text is arranged in three columns.  Each column is 193 points wide, and they are separated by six points.  The are each 660 points tall, extending up from near the bottom of the page.  The large text above is positioned 20 points below the top-left of the page.  And the rest of this text is a bunch of mindless repeating text which is written solely to demonstrate the word wrapping feature of the PDF munger--er, creator--tool.  \
FIRST PARAGRAPH OF MINDLESS TEXT:\
What is normalization? --Normalization is a technique of database design that suggests that certain criteria be used when constructing a table layout (deciding what columns each table will have, and creating the key structure), where the idea is to eliminate redundancy of non-key data across tables. Normalization is usually referred to in terms of forms, and I will introduce only the first three, even though it is somewhat common to use other, more advanced forms (fourth, fifth, Boyce-Codd; see documentation).\
SECOND PARAGRAPH OF MINDLESS TEXT:\
What is normalization? --Normalization is a technique of database design that suggests that certain criteria be used when constructing a table layout (deciding what columns each table will have, and creating the key structure), where the idea is to eliminate redundancy of non-key data across tables. Normalization is usually referred to in terms of forms, and I will introduce only the first three, even though it is somewhat common to use other, more advanced forms (fourth, fifth, Boyce-Codd; see documentation).\
THIRD PARAGRAPH OF MINDLESS TEXT:\
What is normalization? --Normalization is a technique of database design that suggests that certain criteria be used when constructing a table layout (deciding what columns each table will have, and creating the key structure), where the idea is to eliminate redundancy of non-key data across tables. Normalization is usually referred to in terms of forms, and I will introduce only the first three, even though it is somewhat common to use other, more advanced forms (fourth, fifth, Boyce-Codd; see documentation).\
FOURTH PARAGRAPH OF MINDLESS TEXT:\
What is normalization? --Normalization is a technique of database design that suggests that certain criteria be used when constructing a table layout (deciding what columns each table will have, and creating the key structure), where the idea is to eliminate redundancy of non-key data across tables. Normalization is usually referred to in terms of forms, and I will introduce only the first three, even though it is somewhat common to use other, more advanced forms (fourth, fifth, Boyce-Codd; see documentation).\
FIFTH PARAGRAPH OF MINDLESS TEXT:\
What is normalization? --Normalization is a technique of database design that suggests that certain criteria be used when constructing a table layout (deciding what columns each table will have, and creating the key structure), where the idea is to eliminate redundancy of non-key data across tables. Normalization is usually referred to in terms of forms, and I will introduce only the first three, even though it is somewhat common to use other, more advanced forms (fourth, fifth, Boyce-Codd; see documentation).\
SIXTH PARAGRAPH OF MINDLESS TEXT:\
What is normalization? --Normalization is a technique of database design that suggests that certain criteria be used when constructing a table layout (deciding what columns each table will have, and creating the key structure), where the idea is to eliminate redundancy of non-key data across tables. Normalization is usually referred to in terms of forms, and I will introduce only the first three, even though it is somewhat common to use other, more advanced forms (fourth, fifth, Boyce-Codd; see documentation).\
SEVENTH PARAGRAPH OF MINDLESS TEXT:\
What is normalization? --Normalization is a technique of database design that suggests that certain criteria be used when constructing a table layout (deciding what columns each table will have, and creating the key structure), where the idea is to eliminate redundancy of non-key data across tables. Normalization is usually referred to in terms of forms, and I will introduce only the first three, even though it is somewhat common to use other, more advanced forms (fourth, fifth, Boyce-Codd; see documentation).\
This is the end of the mindless text.\
";

	PDF_collection *boxes = new PDF_collection();
	boxes->addsub(new PDF_box(12,672,193,660));
	boxes->addsub(new PDF_box(211,672,193,660));
	boxes->addsub(new PDF_box(409,672,193,660));

	// page->addContent(PsCommands::set_font_and_size("Font1", 12));
	page->addContent(PsCommands::drawstring("Font1", 12, boxes, text));

	// -------------------------------------------------------------

	// Print document

	if(1) { // PDF
		if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Outputing PDF\n");
		obj = root->renderBasicObjects();
		String result;
		obj->render_pdf(NULL, result);
		fwrite((void*)(result.getrawstring()), 1, result.getlength(), stdout);
	}
	else if(0) { // Low Level XML
		if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Outputing low level XML\n");
		String result;
		obj = root->renderBasicObjects();
		obj->render_xml(0, result);
		printf("%s", result.nullterm());
	}
	else { // High level XML
		if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Outputing high level XML\n");
		printf("%s", root->render_xml(0).nullterm());
	}
	
	if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Finished Program\n");

	return 0;
}

