#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "XMLModule_highLevel.hpp"
#include "XMLModule_lowLevel.hpp"
#include "XMLHandler.hpp"
#include "XMLReader.hpp"
#include "PDFReader.hpp"

#include "PdfObjects.hpp"
#include "pdfc_base.h"
#include "pdfc_basicobjects.h"

#include "ObjParser.hpp"

#include "stringclass.h"
#include "stringformatter.h"

#include "FontReader.hpp"
#include "PDFTokenizer.hpp"
#include "ImageReader.hpp"
#include "PsCommands.hpp"

#define MAX_PAGES 0 // a value of 0 means unlimited

#define DEBUG_PAGE_PROGRESS true
#define DEBUG_PROGRESS_STAGES true
#define DEBUG_PROGRESS_STAGES_PAUSE false

PdfRoot* readTemplate(char *filename);
void readCatalog(PdfRoot *pdfRoot, PdfPage *templatePage, PdfFont *font, PdfFont *fontBold);
void formatPrice(char *input, char *output);
void renderHeader(PdfPage *page, PdfFont *font, PdfFont *fontBold, char *dept, char* prod, int pos);

int main(int argc, char *argv) {
	PdfRoot *pdfRoot = new PdfRoot();
	PObj *root = NULL;
	char *outputFile = "sol.pdf";
	FILE *fout = NULL;

	PdfObject::COMPRESS_DATA = 3;

if(DEBUG_PROGRESS_STAGES) printf("Beginning\n");
if(DEBUG_PROGRESS_STAGES_PAUSE) {
	printf("Press any key to continue\n");
	getchar();
} 
	init_pscommands();

	PdfRoot *pdfCoverRoot = readTemplate("cover_page.xml");
	if(pdfCoverRoot == NULL) {
		printerr("xmlparser returned NULL\n");
	} 

	PdfPage *coverPageTemplate = pdfCoverRoot->getPage(1);
	PdfPage *coverPage = (PdfPage*) coverPageTemplate->copy(pdfRoot);

	if(coverPage == NULL) {
		printerr("Page NOT FOUND\n");
		delete pdfCoverRoot;
	}

	PdfRoot *pdfTemplateRoot = readTemplate("template.xml");
	if(pdfTemplateRoot == NULL) {
		printerr("xmlparser returned NULL\n");
	} 

	PdfPage *templatePage = pdfTemplateRoot->getPage(1);

	if(DEBUG_PROGRESS_STAGES) fprintf(stderr, "Building main objects\n");
	pdfRoot->author = "Steve";
	pdfRoot->title = "Test Document";
	pdfRoot->subject = "Testing";
	pdfRoot->keywords = "pdf, test";
	pdfRoot->showBookmarks = true;

	// Add Font
	PdfFont *font = new PdfFont();
	font->id = 3;
	font->setFontHelvetica();
	pdfRoot->addResource(font);

	PdfFont *fontBold = new PdfFont();
	fontBold->id = 4;
	fontBold->setFontHelveticaBold();
	pdfRoot->addResource(fontBold);

	if(templatePage->contentObj != NULL) {
		if(templatePage->contentObj->get_type() != t_collection &&
		   templatePage->contentObj->get_type() != t_shared_collection) {

			PDF_collection *temp = new PDF_collection();
			temp->addsub(templatePage->contentObj);
			templatePage->contentObj = temp;
		}
		((PDF_collection*)templatePage->contentObj)->sharedStream = true;
	}

	readCatalog(pdfRoot, templatePage, font, fontBold);

if(DEBUG_PROGRESS_STAGES) printf("Rendering Basic Objects\n");
if(DEBUG_PROGRESS_STAGES_PAUSE) {
	printf("Press any key to continue\n");
	getchar();
} 
	// Print document
	root = pdfRoot->renderBasicObjects();

if(DEBUG_PROGRESS_STAGES) printf("Deleting High Level Objects\n");
if(DEBUG_PROGRESS_STAGES_PAUSE) {
	printf("Press any key to continue\n");
	getchar();
} 
	delete pdfRoot;
	pdfRoot = NULL;

	if(templatePage != NULL && templatePage->contentObj != NULL) 
		((PDF_collection*)templatePage->contentObj)->sharedStream = false;

	if(pdfTemplateRoot != NULL) {
		delete pdfTemplateRoot;
	}

if(DEBUG_PROGRESS_STAGES) printf("Rendering PDF\n");
if(DEBUG_PROGRESS_STAGES_PAUSE) {
	printf("Press any key to continue\n");
	getchar();
} 
	String result;
	root->render_pdf(NULL, result);

if(DEBUG_PROGRESS_STAGES) printf("Deleting Basic Objects\n");
if(DEBUG_PROGRESS_STAGES_PAUSE) {
	printf("Press any key to continue\n");
	getchar();
} 
	delete root;
	root = NULL;

if(DEBUG_PROGRESS_STAGES) printf("Printing to file\n");
if(DEBUG_PROGRESS_STAGES_PAUSE) {
	printf("Press any key to continue\n");
	getchar();
} 
	fout = fopen(outputFile, "wb");
	if(fout == NULL) 
		printerr("Error - could not open %s for writing\n", outputFile);
	fwrite((void*)(result.getrawstring()), 1, result.getlength(), fout);
	if(fclose(fout) != 0)
		printerr("Error - could not close %s\n", outputFile);

if(DEBUG_PROGRESS_STAGES) printf("Finished\n");
if(DEBUG_PROGRESS_STAGES_PAUSE) {
	printf("Press any key to continue\n");
	getchar();
} 
	return 0;
}

PdfRoot* readTemplate(char *filename) {
	// Read Document
	XMLModule_highLevel module;
	XMLHandler handler(&module);
	XMLReader::parseXmlFile(&handler, filename);

	return module.getRoot();
}

void readCatalog(PdfRoot *pdfRoot, PdfPage *templatePage, PdfFont *font, PdfFont *fontBold) {
	FILE *data = fopen("xml_catalog", "r");
	char dataBuffer[1024];
	char field[1024];

	char prevDept[10];
	char prevProdLine[10];
	char prevItem[10];
	char price[20];

	char *ch;
	int fieldNum, i;
	int fieldSize = 100;

	int numPages = 0;
	int numRecords = 0;
	int numRecordsForPage = 0;

	PdfPage *page = NULL;

	if(templatePage == NULL)
		printerr("Error - no template page\n");

/*
	int xMin = 36;
	int xMax = 576;
	int yMin = 140;
	int yMax = 740;
*/
	int xMin = 36;
	int xMax = 576;
	int yMin = -730;
	int yMax = -150;

	int pos = yMin-1;

	int fontHeight = 12;
	int rowHeight = 0;

	prevDept[0] = '\0';
	prevProdLine[0] = '\0';
	prevItem[0] = '\0';

	PdfBookmark *deptBookmark = NULL;
	PdfBookmark *prodBookmark = NULL;

	using namespace StringFormatter;

	if(data == NULL) {
		printf("Could not open file\n");
		exit(1);
	}
	fgets(dataBuffer, 1024, data); // Remove header line
	while(fgets(dataBuffer, 1024, data) != NULL) {
		ch = dataBuffer;
		fieldNum = 0;
		i = 0;
		while(*ch != '\0') {
			if(*ch == '\t') {
				field[fieldNum*fieldSize+i] = '\0';
				i = 0;
				fieldNum++;
			}
			else if(*ch == '\n') {
				*ch = '\0';
				field[fieldNum*fieldSize+i] = '\0';
			}
			else {
				field[fieldNum*fieldSize+i] = *ch;
				i++;
			}
			ch++;
		}
		field[fieldNum*fieldSize+i] = '\0';
		fieldNum++;

		if(strncmp(prevDept, field+(0*fieldSize), 10) != 0) {
			// fprintf(stderr, "Dept different - %s\n", field+(0*fieldSize));
			pos = yMin-1;
			deptBookmark = NULL;
		}
		if(strncmp(prevProdLine, field+(2*fieldSize), 10) != 0) {
			// fprintf(stderr, "ProdLine different - %s\n", field+(2*fieldSize));
			pos = yMin-1;
			prodBookmark = NULL;
		}

		if(pos < yMin) {
			if(page != NULL) {
				page->addContent(PsCommands::restore_graphics_state());
			}

			if(MAX_PAGES != 0 && numPages > MAX_PAGES) {
				fclose(data);
				return;
			}

if(DEBUG_PAGE_PROGRESS) fprintf(stderr, "New Page %d\n", numPages);
			numPages++;
			numRecordsForPage = 0;

			pos = yMax;

			page = (PdfPage*) templatePage->copy(pdfRoot, false);
			// page = new PdfPage();
			// pdfRoot->addPage(page);

			if(page == NULL)
				printerr("Page NOT FOUND\n");

			page->addStream( (PDF_collection*) templatePage->contentObj);
			page->switchStream();

			renderHeader(page, font, fontBold, field+(1*fieldSize), field+(3*fieldSize), pos);

			if(deptBookmark == NULL) {
				deptBookmark = new PdfBookmark(field+(1*fieldSize), page);
				pdfRoot->addBookmark(deptBookmark);
			}
			if(prodBookmark == NULL) {
				prodBookmark = new PdfBookmark(field+(3*fieldSize), page);
				deptBookmark->addBookmark(prodBookmark);
			}

			page->addContent(PsCommands::set_strokecolorCMYK(0,0,0,1));

			rowHeight = 0;

			strncpy(prevDept, field+(0*fieldSize), 10);
			strncpy(prevProdLine, field+(2*fieldSize), 10);
		}
/*
		page->addContent(PsCommands::moveto(xMin, yMin));
		page->addContent(PsCommands::lineto(xMin, yMax));
		page->addContent(PsCommands::lineto(xMax, yMax));
		page->addContent(PsCommands::lineto(xMax, yMin));
		page->addContent(PsCommands::lineto(xMin, yMin));
		page->addContent(PsCommands::strokepath());
*/

		if(strncmp(prevItem, field+(4*fieldSize), 10) != 0) {
			pos -= rowHeight;

			if( (numRecordsForPage % 5) == 0 && numRecordsForPage != 0) {
				page->addContent(PsCommands::moveto(xMin, pos-2));
				page->addContent(PsCommands::lineto(xMax, pos-2));
				page->addContent(PsCommands::strokepath());
				pos -= 10;
			}

			rowHeight = getWrappedStringHeight(field+(5*fieldSize), font, fontHeight, 340);

			page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin, pos, 45, fontHeight), field+(4*fieldSize)));
			page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin+50, pos, 260, rowHeight), field+(5*fieldSize)));

			numRecordsForPage++;
			// fprintf(stderr, "Done Record %d\n", numRecords);
			numRecords++;
		}

		if( *(field+(6*fieldSize)) == '-') {
			formatPrice(field+(7*fieldSize), price);
			page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin+348, pos, 60, fontHeight), price));
		}
		else if( *(field+(6*fieldSize)) == 'A') {
			formatPrice(field+(7*fieldSize), price);
			page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin+413, pos, 60, fontHeight), price));
		}
		else if( *(field+(6*fieldSize)) == 'B') {
			formatPrice(field+(7*fieldSize), price);
			page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin+480, pos, 60, fontHeight), price));
		}
		else if( *(field+(6*fieldSize)) == 'W') {
		}
		else {
			printf("Unknown price code - %s\n", field+6*fieldSize);
		}

		strncpy(prevItem, field+(4*fieldSize), 10);
	}
	fclose(data);

}

void formatPrice(char *input, char *output) {
	double price = 0.0;

	if(strlen(input) > 0) {
		price = atof(input);
	}

	sprintf(output, "$%8.2f", price);
}

void renderHeader(PdfPage *page, PdfFont *font, PdfFont *fontBold, char *dept, char* prod, int pos) {
	int fontHeight; 
	int xMin = 36;
	int xMax = 576;
	time_t currentTime;
	struct tm* currentDate = NULL;
	char formatedDate[512];

	PdfResourceRef *fontRef = new PdfResourceRef();
	fontRef->resource = font;
	fontRef->type = font->type;
	fontRef->name = ("Font1");
	page->addResourceRef(fontRef);

	PdfResourceRef *fontRefBold = new PdfResourceRef();
	fontRefBold->resource = fontBold;
	fontRefBold->type = fontBold->type;
	fontRefBold->name = ("Font2");
	page->addResourceRef(fontRefBold);

	page->addContent(PsCommands::save_graphics_state());

	page->addContent(PsCommands::moveto(xMin, pos));
	page->addContent(PsCommands::lineto(xMax, pos));
	page->addContent(PsCommands::strokepath());

	page->addContent(PsCommands::set_strokecolorCMYK(0,0,0,0));
	page->addContent(PsCommands::set_fillcolorCMYK(0,0,0,0));

	fontHeight = 14;
	// page->addContent(PsCommands::set_font_and_size("Font2", fontHeight));

	page->addContent(PsCommands::drawstring("Font2", fontHeight, new PDF_box(250, -95, 352, 17), dept));
	page->addContent(PsCommands::drawstring("Font2", fontHeight, new PDF_box(250, -112, 352, 17), prod));

	fontHeight = 10;
	currentTime = time(NULL);
	currentDate = localtime( &currentTime );
	strftime(formatedDate, 512, "%m/%d/%Y", currentDate);
	page->addContent(PsCommands::drawstring("Font2", fontHeight, new PDF_box(65, -115, 352, 17), formatedDate));

	page->addContent(PsCommands::set_strokecolorCMYK(0,0,0,1));
	page->addContent(PsCommands::set_fillcolorCMYK(0,0,0,1));

	fontHeight = 12;
	// page->addContent(PsCommands::set_font_and_size("Font1", fontHeight));

	page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin, pos+fontHeight+2, 600, fontHeight), "Part#"));
	page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin+50, pos+fontHeight+2, 600, fontHeight), "Description"));
	page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin+348, pos+fontHeight+2, 600, fontHeight), "End User"));
	page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin+423, pos+fontHeight+2, 600, fontHeight), "Junkie"));
	page->addContent(PsCommands::drawstring("Font1", fontHeight, new PDF_box(xMin+490, pos+fontHeight+2, 600, fontHeight), "Guru"));

}
