#include <stdlib.h>
#include <stdio.h>

#include "XMLModule_highLevel.hpp"
#include "XMLModule_lowLevel.hpp"
#include "XMLHandler.hpp"
#include "XMLReader.hpp"
#include "PDFReader.hpp"

#include "PdfObjects.hpp"
#include "pdfc_base.h"
#include "pdfc_basicobjects.h"

#include "ObjParser.hpp"

#include "stringclass.h"

#define HIGH_LEVEL 1
#define LOW_LEVEL 2
#define PDF 3

int main(int argc, char *argv) {
	int parseType = HIGH_LEVEL;
	int writeType = PDF;
	PdfRoot *pdfRoot = NULL;
	PObj *root = NULL;
	char *filename = "examples/mytest.xml";
	struct stat statStruct;
	char *buffer;
	long bufferLen;

	FILE *fp;

	stat(filename, &statStruct);
	bufferLen = statStruct.st_size;
	buffer = new char[bufferLen+1];
	buffer[bufferLen] = '\0';
	// printf("Stat: %ld\n", statStruct.st_size);
 
	fp = fopen(filename, "rb");
	if(fp == NULL)
		printerr("Error - could not open %s\n", filename);
	fread(buffer, 1, bufferLen, fp);
	if(fclose(fp) != 0)
		printerr("Error - could not close %s\n", filename);
 
	init_pscommands();

	// Read Document
	if(parseType == HIGH_LEVEL) {
		XMLModule_highLevel module;
		XMLHandler handler(&module);
		XMLReader::parseXmlMemory(&handler, buffer, bufferLen);
		pdfRoot = module.getRoot();
	}
	else if(parseType == LOW_LEVEL) {
		XMLModule_lowLevel module;
		XMLHandler handler(&module);
		XMLReader::parseXmlMemory(&handler, buffer, bufferLen);
		root = module.getRoot();
	}
	else if(parseType == PDF) { 
		StringReader data(buffer, bufferLen);
		PDFReader pReader;
		root= pReader.parse(&data);
	}

	if(pdfRoot == NULL) {
		ObjParser objParser;
		if(root == NULL) {
			printerr("xmlparser returned NULL (new error) \n");
		}
		pdfRoot = objParser.parse(root);

		if(pdfRoot == NULL) {
			printerr("ObjParser returned NULL\n");
		}

		if(root != NULL) 
			delete root;
	} 

	// Print document
	if(writeType == PDF) { // PDF
		root = pdfRoot->renderBasicObjects();
		String result;
		root->render_pdf(NULL, result);
		fwrite((void*)(result.getrawstring()), 1, result.getlength(), stdout);
	}
	else if(writeType == LOW_LEVEL) { // Low Level XML
		root = pdfRoot->renderBasicObjects();
		String __censored;
		root->render_xml(0, __censored);
		printf("%s", __censored.nullterm());
	}
	else if(writeType == HIGH_LEVEL) { // High level XML
		printf("%s", pdfRoot->render_xml(0).nullterm());
	}

	return 0;
}

