#include "dll_export.h"


#ifndef PDFC_BASICOBJECTS_ALREADY_INCLUDED
#define PDFC_BASICOBJECTS_ALREADY_INCLUDED

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h> // For time() function.
#include "pdfc_base.h"
#include "DataFilter.hpp"

/*
The PDF spec defines nine basic types of objects: booleans, numbers, strings, hexstrings, names, 
arrays, dictonaries, streams and nulls.  Numbers come in two varieties: ints and reals.  This 
header defines all eight types as subclasses of class PObj. A trailer type is also used.
*/

#define saveStreamToFile false

class DllExport PDF_blob: public PObj {
public:
    char* rawdata; 
    uint length; // Size (in bytes) of the blob.

    static int filenum;

    PDF_blob();
    virtual ~PDF_blob();
	virtual PObj* copy();
	virtual void copyFields(PDF_blob* alpha);
	virtual bool isBinary();
    virtual void setValue(char *str);
    virtual void setValue(char *str, int len);
    virtual void setValue(String string);
    virtual uint getlength();
    virtual void putlength(uint newlength);
    virtual void render_xml_start(int depth, String& dest);
    virtual void render_xml_end(int depth, String& dest);
    virtual void render_pdf_start(Xreftable* x, String& dest);
    virtual void render_pdf_end(Xreftable* x, String& dest);
    virtual ObjectType get_type();
    virtual char* get_printable_type();
};

class DllExport PDF_inline_image : public PDF_blob {
public:
    PDF_inline_image();
    virtual ~PDF_inline_image();
    virtual void render_xml_start(int depth, String& dest);
    virtual void render_xml_end(int depth, String& dest);
    virtual void render_pdf_start(Xreftable* x, String& dest);
    virtual void render_pdf_end(Xreftable* x, String& dest);
};

class DllExport PDF_null: public PObj {
public:
	virtual PObj* copy();
	virtual uint getlength();
	virtual void putlength(uint newlength);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
};

class DllExport PDF_boolean: public PObj {
public:
	bool value;

	PDF_boolean();
	PDF_boolean(bool initVal);
	virtual PObj* copy();
	virtual void copyFields(PDF_boolean* alpha);
	virtual uint getlength();
	virtual void putlength(uint newlength);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
};

class DllExport PDF_numeric: public PObj {
public:
	PDF_numeric();
	virtual bool isNumeric() { return true; } 
	virtual void setNumericValue(double newVal) = 0;
	virtual double getNumericValue() = 0;
private:
};

class DllExport PDF_int: public PDF_numeric {
public:
	int value;

	PDF_int();
	PDF_int(int i);
	virtual PObj* copy();
	virtual void copyFields(PDF_int* alpha);
	virtual void setNumericValue(double newVal);
	virtual double getNumericValue();
	virtual uint getlength();
	virtual void putlength(uint newlength);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
};

class DllExport PDF_real: public PDF_numeric {
public:
	double value;

	PDF_real();
	PDF_real(double d);
	virtual PObj* copy();
	virtual void copyFields(PDF_real* alpha);
	virtual uint getlength();
	virtual void setNumericValue(double newVal);
	virtual double getNumericValue();
	virtual void putlength(uint newlength);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
};

class DllExport PDF_string: public PObj {
public:
	String value;

	PDF_string();
	PDF_string(String newValue);
	PDF_string(char *newValue);
	virtual ~PDF_string();
	virtual PObj* copy();
	virtual void copyFields(PDF_string* alpha);
	virtual uint getlength();
	virtual void putlength(uint newlength);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *data);
	virtual void setValue(String data);
};

class DllExport PDF_hexstring: public PDF_string {
public:
	virtual PObj* copy();
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual void setValue(char *data);
	virtual void setValue(String data);
};

class DllExport PDF_name: public PDF_string {
public:
	PDF_name();
	PDF_name(String name);
	virtual PObj* copy();
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual void trimWhiteSpace();
	virtual bool isValidName();
};

class DllExport PDF_array: public PObj {
public:
	virtual PObj* copy();
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf_start(Xreftable* x, String& dest);
	virtual void render_pdf_end(Xreftable* x, String& dest);
};

class DllExport PDF_dictionary: public PObj {
public:
	virtual PObj* copy();
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf_start(Xreftable* x, String& dest);
	virtual void render_pdf_end(Xreftable* x, String& dest);
	virtual void showObject();
	virtual String** getKeys();
	virtual PObj* getValueForKey(String key);
	virtual void setValueForKey(String key, PObj *value, bool deleteObjectOnDestruct=true);
};

class DllExport PDF_stream: public PObj {
public:
	PDF_dictionary *dictionary;
	PDF_blob *dataObject;

	PDF_stream();
	PDF_stream(PDF_dictionary *newDict, PDF_blob *newBlob);
	PDF_stream(String newData);
	virtual ~PDF_stream();
	virtual PObj* copy();
    virtual void copyFields(PDF_stream* alpha);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf_start(Xreftable* x, String& dest);
	virtual void render_pdf_end(Xreftable* x, String& dest);
};

class DllExport PDF_ref: public PObj { // Reference type.
public:
	uint ref; // Reference to some other object, per the PDF spec.

	PDF_ref();
	PDF_ref(int newId);
	PDF_ref(PObj *obj);
    virtual PObj* copy();
    virtual void copyFields(PDF_ref* alpha);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual PObj* getObject(PObj* root);
private:
	virtual void PDF_ref_init();
};

class DllExport PDF_trailer: public PObj {
public:
	PDF_dictionary *dictionary;

	PDF_trailer();
	virtual ~PDF_trailer();
    virtual PObj* copy();
    virtual void copyFields(PDF_trailer* alpha);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf_start(Xreftable* x, String& dest);
	virtual void render_pdf_end(Xreftable* x, String& dest);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
private:
	uint* sortXrefEntries(Xreftable *xrefTable);
};

class DllExport PDF_root: public PObj { // Root type.

public:
	PObj *rootNode;
	PDF_trailer *trailer;
	Hashtable idHash;

	PDF_root();
    virtual PObj* copy();
    virtual void copyFields(PDF_root* alpha);
	virtual ~PDF_root();
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf_start(Xreftable* x, String& dest);
	virtual void render_pdf_end(Xreftable* x, String& dest);
	virtual void updateTrailer(PDF_trailer *newTrailer);
};

class DllExport PDF_deletion: public PObj {
public:
    virtual PObj* copy();
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf_start(Xreftable* x, String& dest);
	virtual void render_pdf_end(Xreftable* x, String& dest);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(char *str);
};

class DllExport PDF_shared_collection : public PDF_collection {
public:
	PDF_collection *currentStream;

    PDF_shared_collection();

	virtual void addsub(PObj* newsub, bool deleteObjectOnDestruct=true);
	virtual void addNewStream(bool deleteObjectOnDestruct=true);
	virtual void addNewStream(PDF_collection *collection, bool deleteObjectOnDestruct=true);

    virtual PObj* copy();
    virtual ObjectType get_type();
    virtual char* get_printable_type();
private:
};

class DllExport PDF_pscommand: public PObj {
public:
	String value;
	// These attributes are only used for complex commands
	String *attrName; 
	String *attrValue;
	int numAttr;
	char* contentstring;

	PDF_pscommand();
	PDF_pscommand(String name);
    virtual PObj* copy();
    virtual void copyFields(PDF_pscommand* alpha);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual void render_pdf_start(Xreftable* x, String& dest);
	virtual void render_pdf_end(Xreftable* x, String& dest);
	virtual void setValue(char *str);
	virtual void setValue(String data);
	virtual uint getlength();
	virtual void putlength(uint newlength);
};

class DllExport PDF_box: public PObj {
public:
	int x, y, width, height;

	PDF_box();
	PDF_box(int newX, int newY, int newWidth, int newHeight);
    virtual PObj* copy();
    virtual void copyFields(PDF_box* alpha);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf_start(Xreftable* x, String& dest);
	virtual void render_pdf_end(Xreftable* x, String& dest);
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void setValue(int newx, int newy, int newwidth, int newheight);
};

#define NUMBER_OF_COMPLEX_COMMANDS 4

void DllExport init_pscommands(); // Initialize the pscommands array (below).
int DllExport get_number_of_args_primitive(String& arg); // Get number of arguments for given arg (or arg alias).
char* DllExport get_realname_primitive(String& alias);
char* DllExport get_alias_primitive(char* realname);
char* DllExport get_alias_primitive(String& realname);
bool DllExport name_isvalid_primitive(char* name);
bool DllExport name_isvalid_complex(char* name);
bool DllExport name_isvalid_complex(String& name);
PObj* DllExport render_complex_command_to_primitives(PObj* o, PdfPage* currentpage);

extern char* DllExport pscommands_complex[];

#endif
