#include "dll_export.h"


#ifndef STRINGFORMATTER_ALREADY_INCLUDED
#define STRINGFORMATTER_ALREADY_INCLUDED

#include "PdfObjects.hpp"

/*
Wrap the specified source string into a box so that no line width exceeds boxwidth when the string is rendered
in the specified font.  If wordwrap is true, then rather fitting the maximum possible number of characters
onto each line, wrapstring() will instead only fit the maximum possible number of words onto each line
(where words are delimited by spaces (hex 20)).
If during the construction of the wrapped string based on source, the wrapped string fills the maximum number
of lines in the box (calculated as the floor of boxheight / fontsize (in points)), then the returned String
object contains only that subset of the source that actually fit in the box, and leftovers is set to point
to the first character in the source which didn't fit.  If the entire source fits in the box, then leftovers
is set to NULL.
The units for boxwidth and boxheight are taken to be points.
source must be null terminated.
*/

namespace StringFormatter {
uint DllExport countCharsThatFit(char* source, PdfFont* font, uint boxwidth, bool wordwrap, int fontSize); // Return length of source,
	// or the maximum number of characters in source that fit within boxwidth, or the number of chars to the next
	// newline already embedded in the source, whichever is least.

String DllExport wrapstring(char* source, PdfFont* font, uint boxwidth, uint boxheight, bool wordwrap, char*& leftovers, int fontSize);

String DllExport stripCR(char* source); // Convert CRLF pairs to LFs, then convert individual CRs to LFs.

PObj* renderLF(char* source, int x, int y, PdfPage* currentpage, String fontId, int fontSize); // Render a single source string containing linefeeds into a list of
	// postscript commands to draw each linefeed-delimited line and replace linefeeds with cursor-move commands.
	// The returned list is null-terminated.  Drawing begins at the specified coordinates (specified in points),
	// and proceeds downward and rightward (specified coordinates are the top-left of the box).

int DllExport getUnwrappedStringWidth(String text, PdfFont *font, int size); // Get the width of the text for the specified font if no wrapping is done
int DllExport getWrappedStringHeight(String text, PdfFont *font, int size, int width); // Get the height of the block of text for the specified font wrapping at the designated width
};


#endif
