#include "DataFilter.hpp"
 
#include <stdlib.h>
#include <stdio.h>

#ifdef WIN32 // Zlib needs this.
#ifndef _WINDOWS
#define _WINDOWS
#endif
//#define ZLIB_DLL
#ifndef off_t
#define off_t long
#endif
#endif


#include "zlib.h"

String DataFilter::encodeXML(String input) {
	String output;
	int iCnt, oCnt;
	char *buffer = new char[input.getlength()+1];
	char token[32];

	const char *iPtr = input.getrawstring();
	const char *iEnd = iPtr+input.getlength();
	char *oPtr = buffer;
	char *oEnd = oPtr+input.getlength();

	iCnt = 0;
	oCnt = 0;
	while(iPtr < iEnd) {
		switch(*iPtr)
		{
		case '&':
			strcpy(token, "&amp;");
			oCnt = pdfc_strlen("&amp;");
			iCnt = 1;
			break;

		case '<':
			strcpy(token, "&lt;");
			oCnt = pdfc_strlen("&lt;");
			iCnt = 1;
			break;

		case '>':
			strcpy(token, "&gt;");
			oCnt = pdfc_strlen("&gt;");
			iCnt = 1;
			break;

		case '\"':
			strcpy(token, "&quot;");
			oCnt = pdfc_strlen("&quot;");
			iCnt = 1;
			break;

		case '\'':
			strcpy(token, "&apos;");
			oCnt = pdfc_strlen("&apos;");
			iCnt = 1;
			break;

		default:
		  if( ((uchar)(*iPtr) > 0x7f || (uchar)(*iPtr) < 0x20) && *iPtr != '\n' && *iPtr != '\r') {
			sprintf(token, "&#x%0X;", (unsigned char) *iPtr);
			oCnt = 6;
			iCnt = 1;

			// printerr("Error - binary data encoutered in XML stream - '%s'\n", nullterm() );
			break;
		 }
		 else {
			*oPtr++ = *iPtr++;
		 }
		}

		if(oPtr + oCnt >= oEnd) {
			*oPtr = 0;
			output.append( (char*) buffer, oPtr-buffer);
			oPtr = buffer;
		}
		if(iCnt > 0 || oCnt > 0) {
			strcpy(oPtr, token);
			iPtr += iCnt;
			oPtr += oCnt;
			iCnt = 0;
			oCnt = 0;
		}
	}
	*oPtr = 0;

	output.append( (char*) buffer, oPtr-buffer);
	
	delete [] buffer;

	return output;
}

//------------------------------------------------------

String DataFilter::encodeHex(String input) {
	String output;
	uint i;
	char *buffer = new char[2*input.getlength()+(input.getlength()/35)+2+1+12+1+1];
	char *optr = buffer;
	const char *data = input.getrawstring();
	char buf;
	uint len = input.getlength();
	long cnt = 0;

	for(i=0; i<len; i++) {
		buf=(data[i]&0xF0)>>4; // Get high nybble of character.
		if(buf>9) buf+=0x7; // Deal with noncontinuity in ASCII chart between 9 and A.
		buf+=0x30; // Convert buf to genuine ASCII character from 0 to 9 or A to F.
		*optr++ =buf;
		buf=(data[i]&0xF); // Get low nybble of character.
		if(buf>9) buf+=0x7;
		buf+=0x30;
		*optr++ =buf;

		cnt++;
		if( (cnt % 35) == 0)
			*optr++ = '\n';
	}
	// *optr++ = '>';
	*optr = 0;

	output.append(buffer, optr-buffer);
	delete [] buffer;

	return output;
}

//----------------------------------------------

String DataFilter::decodeHex(String input) {
	String output;
	char temp[3];
	long length;

	unsigned char *buffer = new unsigned char[(input.getlength()/2)+2];
	unsigned char *optr = buffer;
	unsigned char *data = reinterpret_cast<unsigned char*>(const_cast<char*>(input.getrawstring()));
	length = removeWhiteSpace(data, input.getlength() );
	input.reduceLengthTo(length);
	data = reinterpret_cast<unsigned char*>(const_cast<char*>(input.getrawstring()));

	unsigned char *iptr = data;
	unsigned char *dataEnd = data + length;
	temp[2] = 0;

	if(dataEnd > data && *(dataEnd-1) == '>')
		dataEnd--;

	while(iptr+1 < dataEnd) {
		temp[0] = *iptr++;
		temp[1] = *iptr++;
		*optr++ = strtol(temp, NULL, 16);
	}
	if(iptr < dataEnd) {
		temp[0] = *iptr;
		temp[1] = 0;
		*optr++ = strtol(temp, NULL, 16);
	}
	*optr = 0;

	output.append( (char*)buffer, optr-buffer);
	delete [] buffer;

	return output;
}

//------------------------------------------------------

String DataFilter::encode85(String input) {
	String output;
	uint len;
	unsigned long remain, val, total;
	
	char *buffer = new char[ ((input.getlength() / 4) + 1) * 5 + (5 * (input.getlength() / 70) / 4) + 1 + 2];
	char *optr = buffer;
	const unsigned char *data = (const unsigned char*)(input.getrawstring());
	unsigned char *iptr = const_cast<unsigned char*>(data);
	unsigned char *dataEnd = const_cast<unsigned char*>(data + input.getlength());

	len = 0;
	while(iptr+3 < dataEnd) {
		total = 256*256*256* *iptr++;
		total += 256*256* *iptr++;
		total += 256* *iptr++;
		total += *iptr++;
		remain = total;

		if(total == 0) { 
			*optr++ = 'z';
			len++;
		}
		else {
			val = remain / (85*85*85*85);
			*optr++ = val + 33;
			remain -= val * (85*85*85*85);

			val = remain / (85*85*85);
			*optr++ = val + 33;
			remain -= val * (85*85*85);

			val = remain / (85*85);
			*optr++ = val + 33;
			remain -= val * (85*85);

			val = remain / (85);
			*optr++ = val + 33;
			remain -= val * (85);

			val = remain;
			*optr++ = val + 33;

			len+=5;
		}

		if( len > 70 ) {
			*optr++ = '\n';
			len = 0;
		}
	}

	total = 0;
	len = 0;
	while(iptr < dataEnd) {
		total = total << 8;
		total += *iptr++;
		len++;
	}
	total = total << (8*(4 - len));

	if(total != 0) {
		remain = total;

		val = remain / (85*85*85*85);
		*optr++ = val + 33;
		remain -= val * (85*85*85*85);

		val = remain / (85*85*85);
		*optr++ = val + 33;
		remain -= val * (85*85*85);

		if(len > 1) {
			val = remain / (85*85);
			*optr++ = val + 33;
			remain -= val * (85*85);

			if(len > 2) {
				val = remain / (85);
				*optr++ = val + 33;
				remain -= val * (85);

				// *optr++ = remain + 33; // Unused for last partial quartet
			}
		}
	}
	// *optr++ = '~';
	// *optr++ = '>';

	output.append(buffer, optr-buffer);
	delete [] buffer;

	return output;
}

//------------------------------------------------------

String DataFilter::decode85(String input) {
	String output;
	long i, numZs;
	unsigned long val, total, remain;
	long length;
	long numLeft;
	
	const unsigned char *data = reinterpret_cast<const unsigned char*>(input.getrawstring());
	length = removeWhiteSpace(const_cast<unsigned char*>(data), input.getlength() );
	input.reduceLengthTo(length);
	data = reinterpret_cast<const unsigned char*>(input.getrawstring());
	unsigned char *iptr = const_cast<unsigned char*>(data);
	unsigned char *dataEnd = iptr + length;

	numZs = 0;
	for(i=0; i<length; i++) {
		if(data[i] == 'z')
			numZs++;
	}

	char *buffer = new char[ (( (length-numZs) / 5) + numZs + 1) * 4 ];
	char *optr = buffer;

	if(dataEnd-1 > data &&  *(dataEnd-2) == '~' && *(dataEnd-1) == '>') {
		dataEnd-=2;
	}

	while(iptr+4 < dataEnd || (iptr < dataEnd && *iptr == 'z')) {
		if(0) { // Turn on or off strict error checking (at the expense of speed)
			for(i=0; i<4; i++) {
				if( (iptr[i] < '!' || iptr[i] > 'u') && iptr[i] != 'z') {
					printerr("Illegal character %c in base 85 encoding\n", iptr[i]);
				}
			}
		}

		if(*iptr == 'z') { 
			*optr++ = 0;
			*optr++ = 0;
			*optr++ = 0;
			*optr++ = 0;
			iptr++;
		}
		else {
			total =  (*iptr++ - 33) * (85*85*85*85);
			total += (*iptr++ - 33) * (85*85*85);
			total += (*iptr++ - 33) * (85*85);
			total += (*iptr++ - 33) * (85);
			total += (*iptr++ - 33);
			remain = total;

			val = remain / (256*256*256);
			*optr++ = val;
			remain -= val * (256*256*256);

			val = remain / (256*256);
			*optr++ = val;
			remain -= val * (256*256);

			val = remain / (256);
			*optr++ = val;
			remain -= val * (256);

			*optr++ = remain;
		}
	}
	numLeft = dataEnd - iptr;

	if(numLeft > 0) {
		if(0) { // Turn on or off strict error checking (at the expense of speed)
			for(i=0; i<numLeft; i++) {
				if( (iptr[i] < '!' || iptr[i] > 'u') && iptr[i] != 'z') {
					printerr("Illegal Character %c in base 85 encoding\n", iptr[i]);
				}
			}
		}

		total = 0;
		while(iptr < dataEnd) {
			total *= 85;
			total += *iptr++ - 33;
		}
		total++; // Compensate for missing bytes

		for(i=0; i<(5 - numLeft); i++) {
			total *= 85;
		}

		remain = total;

		val = remain / (256*256*256);
		*optr++ = val;
		remain -= val * (256*256*256);

		if(numLeft > 2) {
			val = remain / (256*256);
			*optr++ = val;
			remain -= val * (256*256);

			if(numLeft > 3) {
				val = remain / (256);
				*optr++ = val;
				remain -= val * (256);

				// *optr++ = remain; // Never used
			}
		}
	}

	output.append(buffer, optr-buffer);
	delete [] buffer;

	return output;
}

//------------------------------------------------------

String DataFilter::encodeFlate(String input) {
	z_stream c_stream; // compression stream 
	int err;
	char *buffer;
	String output;
	int amtSaved = 0;
	int bufferSize;

	// printf("Initial size = %d\n", input.getlength() );

	c_stream.zalloc = (alloc_func)0;
	c_stream.zfree = (free_func)0;
	c_stream.opaque = (voidpf)0;
 
	err = deflateInit(&c_stream, Z_DEFAULT_COMPRESSION);
	if (err != Z_OK)
		printerr("Error - deflateInit failed - %d\n", err);

	c_stream.next_in  = (Bytef*) input.getrawstring();
	c_stream.avail_in  = input.getlength();
 
	// Allocate enough space to hopefully comlete in 1 step (size recomended in zlib spec)
	bufferSize  = ((int)(c_stream.avail_in*1.001))+12;

	buffer = new char[bufferSize]; 
	c_stream.next_out  = (Bytef*) buffer;
	c_stream.avail_out  = bufferSize;
 
	while( (err = deflate(&c_stream, Z_FINISH)) == Z_OK) {
		output.append(buffer, c_stream.total_out - amtSaved);
		amtSaved = c_stream.total_out;
		c_stream.next_out  = (Bytef*) buffer;
		c_stream.avail_out  = bufferSize;
	}
	if(err != Z_STREAM_END) {
		if (err == Z_DATA_ERROR)
			printerr("Error - deflating data - data error\n");
		else
			printerr("Error deflating data - %d\n", err);
	}

	output.append(buffer, c_stream.total_out - amtSaved);

	err = deflateEnd(&c_stream);
	if (err != Z_OK) {
		if (err == Z_DATA_ERROR)
			printerr("Error - deflateEnd failed - data error\n");
		else
			printerr("Error - deflateEnd failed - %d\n", err);
	}

	delete [] buffer;

	// printf("Final size = %d\n", output.getlength() );

	return output;
}

//------------------------------------------------------


String DataFilter::decodeFlate(String input) {
	int err;
	z_stream d_stream; // decompression stream 
	char *buffer;
	int bufferSize = input.getlength();
	String output;
	int amtSaved = 0;

	d_stream.zalloc = (alloc_func)0;
	d_stream.zfree = (free_func)0;
	d_stream.opaque = (voidpf)0;
 
	d_stream.next_in  = (unsigned char*)(input.getrawstring());
	d_stream.avail_in = input.getlength();

	err = inflateInit(&d_stream);
	if (err != Z_OK)
		printerr("Error - inflateInit failed - %d\n", err);

	buffer = new char[bufferSize]; 
	d_stream.next_out  = (Bytef*) buffer;
	d_stream.avail_out  = bufferSize;
 
	while( (err = inflate(&d_stream, Z_SYNC_FLUSH)) == Z_OK) {
		output.append(buffer, d_stream.total_out - amtSaved);
		amtSaved = d_stream.total_out;
		d_stream.next_out  = (Bytef*) buffer;
		d_stream.avail_out  = bufferSize;
	}
	if(err != Z_STREAM_END) {
		if (err == Z_DATA_ERROR)
			printerr("Error flate decoding data - data error\n");
		else
			printerr("Error flate decoding data - %d\n", err);
	}

	output.append(buffer, d_stream.total_out - amtSaved);

	err = inflateEnd(&d_stream);
	if (err != Z_OK) {
		if (err == Z_DATA_ERROR)
			printerr("Error - inflateEnd failed - data error\n");
		else
			printerr("Error - inflateEnd failed - %d\n", err);
	}

	delete [] buffer;

	return output;
}

//------------------------------------------------------

String DataFilter::convertEOLs(String input) {
	String output;
	uint i;
	uint length = input.getlength();
	uint lastCnt = 0;

	uint foundLen;
	char *buffer = new char[ length ];
	char *optr = buffer;
	const char *data = input.getrawstring();

	for(i=0; i<length; i++) {
		if(data[i] == '\r') {
			foundLen = i-lastCnt + 1;

			memcpy(optr, data+lastCnt, foundLen);
			optr[foundLen-1] = '\n';
			optr += foundLen;

			if(i+1 < length && data[i+1] == '\n') 
				i++;

			lastCnt = i + 1;
		}
	}
	foundLen = length - lastCnt;
	if(foundLen > 0) {
		memcpy(optr, data+lastCnt, foundLen);
		optr += foundLen;
	}

 	output.append(buffer, optr-buffer);

	delete [] buffer;

	return output;
}

String DataFilter::encodePDF(String input) {
	String output;
	const char* in=input.getrawstring();
	for(uint c=0; c<input.getlength(); c++) {
		if((in[c]!='(')&&(in[c]!=')')&&(in[c]!='\\')) output.append(in+c, 1);
		else {
			output.append("\\");
			output.append(in+c, 1);
		}
	}
	return output;
}

String DataFilter::decodePDF(String input) {
	String output;
	const char* in=input.getrawstring();
	uint inputlength=input.getlength();
	for(uint c=0; c<inputlength; c++) {
		if(in[c]!='\\') output.append(in+c, 1);
		else {
			if(c+1==inputlength) return output+String("\\"); // Backslash found at the end of input.

			c++;
			output.append(in+c, 1);

			// Old Code
			// if((in[c+1]!='(')&&(in[c+1]!=')')) output.append(in+c, 1); // Character following backslash will be output on next pass through loop.
			// else {} // Intentionally nop; ignore backslashes preceding parentheses.  Parenthesis will be output on next pass through loop.
		}
	}
	return output;
}

//----------------------------------------------

long DataFilter::removeWhiteSpace(unsigned char *data, long length) {
	unsigned char *ptr = data;
	unsigned char *dataEnd = data+length;

	unsigned char *lastEnd = data;
	unsigned char *startSection = data;
	long sectionLength = 0;

	while(ptr < dataEnd) {
		if(*ptr == ' ' || *ptr == '\n') {
			sectionLength = ptr-startSection;

			if(startSection != data) {
				memmove(lastEnd, startSection, sectionLength);

			}
			lastEnd += sectionLength;

			while(ptr < dataEnd && (*ptr == ' ' || *ptr == '\n')) {
				ptr++;
			}
			startSection = ptr;
		}
		else {
			ptr++;
		}
	}
	sectionLength = ptr-startSection;
	if(lastEnd != NULL) {
		memmove(lastEnd, startSection, sectionLength);
	}

	return ((lastEnd + sectionLength) - data);
}

//----------------------------------------------

bool DataFilter::isSupportedEncoding(String encoding) {
	if(encoding == "ASCIIHexDecode" || encoding == "ASCII85Decode" || encoding == "FlateDecode")
		return true;
	else
		return false;
} 

//----------------------------------------------
 
String DataFilter::decode(String data, String encoding) {
    // printf("Decoding %s\n", encoding.nullterm() );
 
    if(encoding == "ASCIIHexDecode") {
        data = DataFilter::decodeHex( data );
    }
    else if(encoding == "ASCII85Decode") {
        data = DataFilter::decode85( data );
    }
    else if(encoding == "FlateDecode") {
        data = DataFilter::decodeFlate( data );
    }
    else {
        printerr("Error - unknown decoding method - %s\n", encoding.nullterm() );
    }
 
	return data;
}

String DataFilter::encode(String data, String encoding) {
    // printf("Encoding %s\n", encoding.nullterm() );
 
    if(encoding == "ASCIIHexDecode") {
        data = DataFilter::encodeHex( data );
    }
    else if(encoding == "ASCII85Decode") {
        data = DataFilter::encode85( data );
    }
    else if(encoding == "FlateDecode") {
        data = DataFilter::encodeFlate( data );
    }
    else {
        printerr("Error - unknown encoding method - %s\n", encoding.nullterm() );
    }
 
	return data;
} 

//----------------------------------------------

