#include "FileReader.hpp"

#include <string.h>
#include "stringclass.h"
#include "pdfc_errors.h"

FileReader::FileReader(char *newFile) {
	bufferCnt = 0;
	nextChar = buffer;
	bufferEnd = buffer;
	fp = NULL;
	bytesRead = 0;
	linesRead = 0;

	if(newFile == NULL) {
		fp = stdin;
		filename = NULL;
	}
	else {
		filename = new char[pdfc_strlen(newFile)+1];
		strcpy(filename, newFile);
		fp = fopen(filename, "rb"); // The b is critical to keep fopen from modifying EOLs
	}

	if(fp == NULL) {
		printerr("ERROR - cannot open file '%s'\n", filename);
	}
}

FileReader::~FileReader() {
	closeFile();
	delete [] filename;
}

void FileReader::closeFile() {
	if(fp != NULL) {
		// printf("Total file size is %d\n", bytesRead);
		if(fclose(fp) == EOF)
			printerr("ERROR - could not close file\n");
		fp = NULL;
	}
}

void FileReader::getNextBlock() {
	if(fp != NULL) { 
		bufferCnt = fread(buffer, 1, MAX_LINE_LENGTH, fp);
		nextChar = buffer;
		bufferEnd = buffer + bufferCnt;
		bytesRead += bufferCnt;
		// printf("bufferCnt: %d nextChar: %x bufferEnd: %x\n", bufferCnt, (int)nextChar, (int)bufferEnd);

		if(bufferCnt == 0) {
			closeFile();
		}
	}
}

int FileReader::getLine(char* dest, int maxChars, bool showEOL ) {
	char *iptr = nextChar;
	char *optr = dest;
	int dataLen;
	int cnt = 0;

	while(fp != NULL) {
		while(iptr<bufferEnd) {
			if(*iptr == '\r') {
				dataLen = iptr - nextChar;
				memcpy(optr, nextChar, dataLen);

				optr+= dataLen;

				if(showEOL) {
					*optr++ = *iptr;
					cnt++;
				}
				else {
					*optr++ = '\n';
					cnt++;
				}
				iptr++;

				if(iptr == bufferEnd) {
					getNextBlock();
					iptr = nextChar;
				}

				if(*iptr == '\n') {
					if(showEOL) {
						*optr++ = *iptr;
						cnt++;
					}
					iptr++;
				}

				nextChar = iptr;

				linesRead++;
				return cnt;
			}
			else if(*iptr == '\n') {
				dataLen = iptr - nextChar;
				memcpy(optr, nextChar, dataLen);

				optr+= dataLen;

				*optr++ = *iptr;
				cnt++;
				iptr++;

				nextChar = iptr;

				linesRead++;
				return cnt;
			}
			else {
				cnt++;
				iptr++;
			}

			if(cnt >= maxChars) {
				// printf("Line Seperated - %d\n", maxChars);
				dataLen = iptr - nextChar;
				memcpy(optr, nextChar, dataLen);
				nextChar = iptr;

				// if(! showEOL && dest[cnt-1] != '\n')
					// printf("Long Line in File Data\n");

				return cnt;
			}
		}
		dataLen = iptr - nextChar;
		memcpy(optr, nextChar, dataLen);
		getNextBlock();
		optr += dataLen;
		iptr = nextChar;
	}

	if(cnt < 1) {
		return -1;
	}
	else {
		return cnt;
	}
}

int FileReader::getByteOffset() {
	return bytesRead - bufferCnt + (nextChar-buffer);
}

////////////

StringReader::StringReader(const char* const newstring, const long dataLength) : thestring(newstring), totalsize(dataLength) {
	bufferCnt = 0;
	buffer = (const char* ) thestring;
	nextChar = buffer;
	bufferEnd = buffer;
	dataLeft = true;
	bytesRead = 0;
	linesRead = 0;
	currposition=0;

	// Debugging - show content blocks
	// printf("Data:\n");
	// fwrite(thestring, 1, dataLength, stdout);
	// printf("\n");
}

StringReader::~StringReader() {}

int StringReader::getLine(char* dest, int maxChars, bool showEOL ) {
	char *iptr = const_cast<char*>(nextChar);
	char *optr = dest;
	int dataLen;
	int cnt = 0;

	while(dataLeft) {
		while(iptr<bufferEnd) {
			if(*iptr == '\r') {
				dataLen = iptr - nextChar;
				memcpy(optr, nextChar, dataLen);

				optr+= dataLen;

				if(showEOL) {
					*optr++ = *iptr;
					cnt++;
				}
				else {
					*optr++ = '\n';
					cnt++;
				}
				iptr++;

				if(iptr == bufferEnd) {
					getNextBlock();
					iptr = const_cast<char*>(nextChar);
				}

				if(*iptr == '\n') {
					if(showEOL) {
						*optr++ = *iptr;
						cnt++;
					}
					iptr++;
				}

				nextChar = iptr;

				linesRead++;
				return cnt;
			}
			else if(*iptr == '\n') {
				dataLen = iptr - nextChar;
				memcpy(optr, nextChar, dataLen);

				optr+= dataLen;

				*optr++ = *iptr;
				cnt++;
				iptr++;

				nextChar = iptr;

				linesRead++;
				return cnt;
			}
			else {
				cnt++;
				iptr++;
			}

			if(cnt >= maxChars) {
				// printf("Line Seperated - %d\n", maxChars);
				dataLen = iptr - nextChar;
				memcpy(optr, nextChar, dataLen);
				nextChar = iptr;

				if(0 && ! showEOL) {
					if(dest[cnt-1] != '\n')
						printf("Long Line in Stream Data\n");
					else
						printf("Regular Line in Stream Data\n");
				}

				return cnt;
			}
		}
		dataLen = iptr - nextChar;
		memcpy(optr, nextChar, dataLen);
		getNextBlock();
		optr += dataLen;
		iptr = const_cast<char*>(nextChar);
	}

	if(cnt < 1) {
		return -1;
	}
	else {
		return cnt;
	}
}

int StringReader::getByteOffset() {
	return bytesRead - bufferCnt + (nextChar-buffer);
}

void StringReader::getNextBlock() {
	if(dataLeft) {
		// file-reader version of code: bufferCnt = fread(buffer, 1, MAX_LINE_LENGTH, fp);
		bufferCnt=totalsize-currposition;
		if(bufferCnt>MAX_LINE_LENGTH) bufferCnt=MAX_LINE_LENGTH;
		buffer = (char* const) thestring+currposition;
		currposition+=bufferCnt;

		nextChar = buffer;
		bufferEnd = buffer + bufferCnt;
		bytesRead += bufferCnt;
		// printf("bufferCnt: %d nextChar: %x bufferEnd: %x\n", bufferCnt, (int)nextChar, (int)bufferEnd);

		if(bufferCnt < 1) {
			dataLeft = false;
		}
	}
}

