#include "ImageReader.hpp"

#include <stdlib.h>
#include <stdio.h>
#include <iostream>
 
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
 
extern "C" {
#include <magick/api.h>
}

#ifdef WIN32
#ifndef off_t
#define off_t long
#endif
#endif
 
#include "DataFilter.hpp"
 
void ImageReader::readImage(PdfImage *pdfImage) {
	char *filename = NULL;
	unsigned int i;

	if(pdfImage->format == "") {
		pdfImage->format = "external";
	}
	else if(pdfImage->format != "external") {
		return;
	}

	Image *image;
	ImageInfo *imageInfo;

	if(pdfImage->filename.getlength() > 1) {
		filename = pdfImage->filename.nullterm();
		// printf("Image: %s\n", filename);

		pdfImage->loadData();
	}

	if(pdfImage->dataLength > 0) {
		imageInfo = CloneImageInfo(NULL);

		ExceptionInfo exceptionInfo;
		GetExceptionInfo( &exceptionInfo );
		image = BlobToImage(imageInfo,pdfImage->data, pdfImage->dataLength, &exceptionInfo);

		if(image == NULL)
			printerr("Error - could not convert image\n");
	}
	else {
		printerr("Error - no data available for image\n");
	}

	pdfImage->width = image->columns;
	pdfImage->height = image->rows;
	pdfImage->bitsPerPixel = image->depth;
	pdfImage->colorSpace = new PdfColorSpace("DeviceRGB");

	// printf("Image - width: %d height: %d\n", pdfImage->width, pdfImage->height);
	// printf("        depth: %d\n", pdfImage->bitsPerPixel);
	// cout << "Image type: " << image.magick() << "\n";

	if(image->depth == 8) {
		if(pdfImage->dataLength > 0)
			delete [] pdfImage->data;

		StringList *ptr = pdfImage->filters;
		StringList *filterStack = NULL;
		StringList *temp;

		while(ptr != NULL) {
			temp = new StringList(ptr->str);
			temp->next = filterStack;
			filterStack = temp;

			ptr = ptr->next;
		}

		if(filterStack != NULL && filterStack->str == "DCTDecode") {
			char *buffer;
			size_t length = 0;

			strcpy(image->magick, "JPEG");

			ExceptionInfo exceptionInfo;
			GetExceptionInfo( &exceptionInfo );
			buffer = (char*) ImageToBlob(imageInfo, image, &length, &exceptionInfo);

			if(buffer == NULL) {
				printerr("Error - could not convert image to JPEG - %s : %s\n", exceptionInfo.reason, exceptionInfo.description);
				// printerr("Error - could not convert image to JPEG - %s : %s\n", exceptionInfo.message, exceptionInfo.qualifier);
			}

			pdfImage->dataLength = length;
			pdfImage->data = new char[ length ];
			memcpy(pdfImage->data, buffer, length);

			temp = filterStack;
			filterStack = filterStack->next;
			delete temp;
		}
		else {
			PixelPacket* pixel = GetImagePixels(image, 0, 0, image->columns, image->rows);

			// for(i=0; i<image.columns()*image.rows(); i++) {
			//	printf("%d Pixel r: %d g: %d b: %d\n", i, pixel[i].red, pixel[i].green, pixel[i].blue );
			// }

			pdfImage->dataLength = pdfImage->width*pdfImage->height*3;
			pdfImage->data = new char[ pdfImage->dataLength ];
		
			for(i=0; i<(pdfImage->width*pdfImage->height); i++) {
				pdfImage->data[3*i] = pixel[i].red;
				pdfImage->data[3*i+1] = pixel[i].green;
				pdfImage->data[3*i+2] = pixel[i].blue;
			}
		}

		DestroyImage(image);
		DestroyImageInfo(imageInfo);

		while(filterStack != NULL) {
			// printf("Stack Filter: %s\n", filterStack->str.nullterm());
			if(! DataFilter::isSupportedEncoding(filterStack->str) && 
			   filterStack->str != "DCTDecode") 
				printerr("Error - unsupported encoding method\n");

			pdfImage->encode(filterStack->str);

			temp = filterStack;
			filterStack = filterStack->next;
			delete temp;
		}
	}
	else {
		printerr("Error - currently unsupported image color depth\n");
	}

	if(filename != NULL)
		delete [] filename;
}

//----------------------------------------------------------

/*
void ImageReader::convertImage(PdfImage *pdfImage) {
	Image image;
	char buffer[256];

	// FIX - unfinished

	Blob blob( pdfImage->data, pdfImage->dataLength ); 

	sprintf(buffer, "%dx%d", pdfImage->width, pdfImage->height);
	image.size( buffer );
	image.magick( "RGB" ); 
	image.read( blob);
	image.magick( "JPEG" ); 
	image.write(&blob);
	
	pdfImage->dataLength = blob.length();
	pdfImage->data = new char[ blob.length() ];
	memcpy(pdfImage->data, blob.data(), blob.length());
}
*/

