#include "PdfObjects.hpp"

#include <stdlib.h>
#include <stdio.h>

#include "DataFilter.hpp"

#ifdef WIN32
#define timezone _timezone
#endif

#define DEBUG_PROGRESS_STAGES false

int PdfObject::nextId = 1;
int PdfObject::fontNum = 1;
int PdfObject::imageNum = 1;
int PdfObject::functionNum = 1;

int PdfResource::nextResourceId = 1;

double PdfPage::DEFAULT_LLX = 0;
double PdfPage::DEFAULT_LLY = 0;
double PdfPage::DEFAULT_URX = 612;
double PdfPage::DEFAULT_URY = 792;

const int PdfObject::OUTPUT_LEVEL=1;
int PdfObject::COMPRESS_DATA=0;

//------------------------------------------

int PdfObject::getNextId() {
	return nextId++;
}
 
int PdfObject::getNumIds() {
	return nextId-1;
}
 
void PdfObject::resetId() {
	nextId = 0;
} 

PdfObject::PdfObject() : id(0), parent(NULL) {
	// printf("Creating PdfObject\n");
}

PdfObject::~PdfObject() {
	// printf("Destroying PdfObject\n");
}

PdfObject* PdfObject::copy() {
	PdfObject *beta = new PdfObject();

	beta->copyFields(this);
	return beta;
}

void PdfObject::copyFields(PdfObject *alpha) {
	id = alpha->id;
	parent = NULL;
}

String PdfObject::render_xml(int depth) {
	return ("<object></object>\n");
}

PObj* PdfObject::renderBasicObjects() {
	printf("renderBasicObjects\n");
	return new PDF_null();
}

//------------------------------------------

PdfRoot::PdfRoot(): PdfObject(), pages(NULL), resources(NULL), linkList(NULL), bookmarks(NULL), showBookmarks(false), defaultFont(NULL), colorSpaces(NULL), functions(NULL) {
}

PdfRoot::~PdfRoot() {
	PdfResource *ptr;
	PdfBookmark *bookmarkPtr;
	PdfLinkList *linkPtr, *tempLink;
	PdfList *listPtr, *tempListPtr;
	PdfObject *temp;

	ptr = pages;
	while(ptr != NULL) {
		// printf("Destroying page\n");
		temp = ptr;
		ptr = ptr->next;
		delete temp;
	}
		
	ptr = resources;
	while(ptr != NULL) {
		// printf("Destroying Resource\n");
		temp = ptr;
		ptr = ptr->next;
		delete temp;
	}
		
	linkPtr = linkList;
	while(linkPtr != NULL) {
		// printf("Destroying LinkList\n");
		tempLink = linkPtr;
		linkPtr = linkPtr->next;
		delete tempLink;
	}
		
	bookmarkPtr = bookmarks;
	while(bookmarkPtr != NULL) {
		// printf("Destroying Bookmark\n");
		temp = bookmarkPtr;
		bookmarkPtr = bookmarkPtr->next;
		delete temp;
	}
		
	listPtr = colorSpaces;
	while(listPtr != NULL) {
		// printf("Destroying ColorSpace\n");
		tempListPtr = listPtr;
		listPtr = listPtr->next;
		delete tempListPtr;
	}
		
	listPtr = functions;
	while(listPtr != NULL) {
		// printf("Destroying Function\n");
		tempListPtr = listPtr;
		listPtr = listPtr->next;
		delete tempListPtr;
	}
		
	if(defaultFont != NULL)
		delete defaultFont;
}

PdfObject* PdfRoot::copy() {
	PdfRoot *beta = new PdfRoot();

	beta->copyFields(this);
	return beta;
}

void PdfRoot::copyFields(PdfRoot *alpha) {
	PdfObject::copyFields(alpha);

	PdfPage *pagePtr = alpha->pages;
	while(pagePtr != NULL) {
		addPage((PdfPage*) pagePtr->copy());
		pagePtr = (PdfPage*) pagePtr->next;
	}
	PdfResource *resourcePtr = alpha->resources;
	while(resourcePtr != NULL) {
		addResource((PdfResource*)resourcePtr->copy());
		resourcePtr = resourcePtr->next;
	}
	PdfBookmark *bookmarkPtr = alpha->bookmarks;
	while(bookmarkPtr != NULL) {
		addBookmark((PdfBookmark*)bookmarkPtr->copy());
		bookmarkPtr = bookmarkPtr->next;
	}

	if(alpha->linkList != NULL)
		linkList = alpha->linkList->copy();

	author = alpha->author;
	creationDate = alpha->creationDate;
	modDate = alpha->modDate;
	creator = alpha->creator;
	producer = alpha->producer;
	title = alpha->title;
	subject = alpha->subject;
	keywords = alpha->keywords;
	title = alpha->title;
	showBookmarks = alpha->showBookmarks;

	PdfList *colorSpacePtr = alpha->colorSpaces;
	while(colorSpacePtr != NULL) {
		addColorSpace( (PdfColorSpace*) colorSpacePtr->object);
		colorSpacePtr = colorSpacePtr->next;
	}
}

String PdfRoot::render_xml(int depth) {
	String str;
	PdfResource *rPtr = resources;
	PdfBookmark *bPtr = bookmarks;
	PdfPage *pPtr;
	int pageId = 1;

	if(OUTPUT_LEVEL > 1) printf("Starting render\n");

	str = "<!DOCTYPE pdf SYSTEM \"high_level_pdf.dtd\">\n";
	str += "<pdf";

	if(author.getlength() > 0) {
		str += " author=\"";
		str += author;
		str += "\"";
	}
	if(creationDate.getlength() > 0) {
		str += " creationDate=\"";
		str += creationDate;
		str += "\"";
	}

	if(creator.getlength() == 0)
		creator = "InterLan PdfConverter";
	str += " creator=\"";
	str += creator;
	str += "\"";

	if(title.getlength() > 0) {
		str += " title=\"";
		str += title;
		str += "\"";
	}
	if(subject.getlength() > 0) {
		str += " subject=\"";
		str += subject;
		str += "\"";
	}
	if(keywords.getlength() > 0) {
		str += " keywords=\"";
		str += keywords;
		str += "\"";
	}

	if(showBookmarks)
		str += " mode=\"show_bookmarks\"";

	str += ">\n";

	if(OUTPUT_LEVEL > 1) printf("Assigning Page Ids\n");
	pPtr = pages;
	while(pPtr != NULL) {
		pPtr->refId = pageId++;
		pPtr = (PdfPage*) pPtr->next;
	}

	if(OUTPUT_LEVEL > 1) printf("Rendering Bookmarks\n");
	while(bPtr != NULL) {
		str += bPtr->render_xml(depth);
		bPtr = bPtr->next;
	}

	if(OUTPUT_LEVEL > 1) printf("Rendering Resources\n");
	while(rPtr != NULL) {
		rPtr->refId = PdfResource::getNextResourceId();
		str += rPtr->render_xml(depth);
		rPtr = rPtr->next;
	}

	if(OUTPUT_LEVEL > 1) printf("Rendering Pages\n");
	pPtr = pages;
	while(pPtr != NULL) {
		str += pPtr->render_xml(depth);
		pPtr = (PdfPage*) pPtr->next;
	}
	str += "</pdf>\n";

	if(OUTPUT_LEVEL > 1) printf("Ending render\n");
	return str;
}

PObj* PdfRoot::renderBasicObjects() {
	PdfPage *nextPage = pages;
	PdfResource *nextResource = resources;

	if(OUTPUT_LEVEL > 2) printf("Starting PdfRoot::renderBasicObjects\n");

	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - start\n");

	PDF_root *root = new PDF_root();
	PObj *catalog, *pageList, *pagesArray, *page, *resource, *trailer;
	PDF_dictionary *trailerDict, *info, *bookmarkRoot;
	PDF_ref *trailerRootRef, *infoRef;
	int pageCnt;

	// Add Catalog
	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - adding catalog\n");
	catalog = new PDF_dictionary();
	((PDF_dictionary*) catalog)->setValueForKey("Type", new PDF_name("Catalog"));
	root->addsub(catalog);
	catalog->id = getNextId();

	// Add Pages PObj
	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - adding pages object\n");
	pageList = new PDF_dictionary();
	((PDF_dictionary*) pageList)->setValueForKey("Type", new PDF_name("Pages"));
	root->addsub(pageList);
	pageList->id = getNextId();
	((PDF_dictionary*) catalog)->setValueForKey("Pages", new PDF_ref(pageList));

	// Pages Array
	pagesArray = new PDF_array();
	((PDF_dictionary*) pageList)->setValueForKey("Kids", pagesArray);

	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - start adding resources\n");
	// Resources - must be rendered before contents to assign ids to the resources
	// before they are referenced in the contents section
	while(nextResource != NULL) {
		if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - adding resource\n");
		nextResource->id = getNextId();
		resource = nextResource->renderBasicObjects(root);
		root->addsub(resource);

		nextResource = nextResource->next;
	}
	
	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - start adding pages\n");
	pageCnt = 0;
	while(nextPage != NULL) {
		if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - adding page\n");
		nextPage->id = getNextId();
		page = nextPage->renderBasicObjects(root, pageList, pagesArray);
		if(page != NULL) {
			root->addsub(page);
			pageCnt++;
		}
		nextPage = (PdfPage*) nextPage->next;
	}
	((PDF_dictionary*) pageList)->setValueForKey("Count", new PDF_int(pageCnt));
	
	// PdfLink -- setting page links
	PdfLinkList *lptr = linkList;
	PObj *destArray;
	PObj *pageRef;
	PObj *destMod;

	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - start adding links\n");
	while(lptr != NULL) {
		if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - adding link\n");
		// printf("Link: %d\n", lptr->link->destination->id );
		if(lptr->link == NULL || lptr->link->destination == NULL)
			printerr("Error - empty link destination\n");
		if(lptr->link->destination->objType() != p_page)
			printerr("Error - link destination is not a page\n");
		page = ((PDF_root*) root)->idHash.findobject(lptr->link->destination->id);

		if(page == NULL)
			printerr("Error - undefined page referenced by annotation link - %s\n", lptr->link->destKey.nullterm());

		destArray = new PDF_array();
		pageRef = new PDF_ref(page);
		destMod = new PDF_name(String("XYZ"));

		destArray->addsub(pageRef);
		destArray->addsub(destMod);
		destArray->addsub(new PDF_int(0) );
		destArray->addsub(new PDF_int( lptr->link->destination->mediaBox[3] ) );
		destArray->addsub(new PDF_null() );

		((PDF_dictionary*) lptr->obj)->setValueForKey("Dest", destArray);

		lptr = lptr->next;
	}

	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - adding document info\n");
	info = new PDF_dictionary();
	info->id = getNextId();
	root->addsub(info);

	producer = "InterLan PdfConverter";
	if(creator.getlength() == 0)
		creator = producer;
	modDate = pdfFormatDate(time(NULL));
	if(creationDate.getlength() == 0)
		creationDate = modDate;

	if(author.getlength() > 0)
		info->setValueForKey("Author", new PDF_string(author));
	if(creationDate.getlength() > 0)
		info->setValueForKey("CreationDate", new PDF_string(creationDate));
	if(modDate.getlength() > 0)
		info->setValueForKey( "ModDate", new PDF_string(modDate));
	if(creator.getlength() > 0)
		info->setValueForKey("Creator", new PDF_string(creator));
	if(producer.getlength() > 0)
		info->setValueForKey("Producer", new PDF_string(producer));
	if(title.getlength() > 0)
		info->setValueForKey("Title", new PDF_string(title));
	if(subject.getlength() > 0)
		info->setValueForKey("Subject", new PDF_string(subject));
	if(keywords.getlength() > 0)
		info->setValueForKey("Keywords", new PDF_string(keywords));

	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - start adding bookmarks\n");
	// Outline proccessing
	if(bookmarks != NULL) {
		int numBookmarks = 0;
		PObj *bookmark, *prevBookmark;

		if(showBookmarks)
			((PDF_dictionary*) catalog)->setValueForKey("PageMode", new PDF_name("UseOutlines"));

		bookmarkRoot = new PDF_dictionary();
		bookmarkRoot->id = getNextId();
		root->addsub(bookmarkRoot);

		PdfBookmark *ptr = bookmarks;
		prevBookmark = NULL;
		while(ptr != NULL) {
			if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - adding bookmark\n");
			bookmark = ptr->renderBasicObjects(root);
			((PDF_dictionary*)bookmark)->setValueForKey("Parent", new PDF_ref(bookmarkRoot));

			if(prevBookmark == NULL) {
				((PDF_dictionary*)bookmarkRoot)->setValueForKey("First", new PDF_ref(bookmark));
			}
			else {
				((PDF_dictionary*)prevBookmark)->setValueForKey("Next", new PDF_ref(bookmark));
				((PDF_dictionary*)bookmark)->setValueForKey("Prev", new PDF_ref(prevBookmark));
			}

			numBookmarks += ptr->count + 1;
			prevBookmark = bookmark;
			ptr = ptr->next;
		}
		((PDF_dictionary*)bookmarkRoot)->setValueForKey("Last", new PDF_ref(prevBookmark));

		bookmarkRoot->setValueForKey("Count", new PDF_int(numBookmarks));
		((PDF_dictionary*) catalog)->setValueForKey("Outlines", new PDF_ref(bookmarkRoot));
	}

	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - adding trailer\n");
	trailer = new PDF_trailer();
	trailerDict = new PDF_dictionary();

	trailerRootRef = new PDF_ref();
	trailerRootRef->ref = catalog->id;
	infoRef = new PDF_ref();
	infoRef->ref = info->id;

	trailerDict->setValueForKey("Root", trailerRootRef);
	trailerDict->setValueForKey("Size", new PDF_int(getNumIds()));
	trailerDict->setValueForKey("Info", infoRef);

	trailer->addsub(trailerDict);

	root->addsub(trailer);

	if(DEBUG_PROGRESS_STAGES) printf("PdfRoot:renderBasicObjects - done\n");
	if(OUTPUT_LEVEL > 2) printf("Ending PdfRoot::renderBasicObjects\n");

	return root;
}

void PdfRoot::addPage(PdfPage *page) {
	if(pages == NULL) {
		pages = page;
		page->prev = NULL;
		page->next = NULL;
	}
	else {
		PdfPage *last;

		last = pages;
		while(last->next != NULL) {
			last = (PdfPage*) last->next;
		}
		last->next = page;
		page->prev = last;
		page->next = NULL;
	}

	if(page->id != 0) {
		// printf("Found page id: %d\n", page->id);
		pageHash.add(page->id, page);
	}

	page->root = this;
}

PdfPage* PdfRoot::getPage(int pageId) {
	if(pages == NULL) 
		return NULL;

	PdfPage *currentPage;

	currentPage = pages;
	while(currentPage != NULL) {
		if(currentPage->id == pageId)
			return currentPage;
		currentPage = (PdfPage*) currentPage->next;
	}

	return NULL;
}

void PdfRoot::addResource(PdfResource *resource) {
	if(resources == NULL) {
		resources = resource;
		resource->prev = NULL;
		resource->next = NULL;
	}
	else {
		PdfResource *last;

		last = resources;
		while(last->next != NULL) {
			last = last->next;
		}
		last->next = resource;
		resource->prev = last;
		resource->next = NULL;
	}

	resourceIdHash.add(resource->resourceId, resource);
}

void PdfRoot::addLink(PdfLink *newLink) {
	addLink(newLink, NULL);
}

void PdfRoot::addLink(PdfLink *newLink, PObj *newObj) {
	PdfLinkList *link = new PdfLinkList(newLink, newObj);

	if(linkList == NULL) {
		linkList = link;
		link->next = NULL;
	}
	else {
		PdfLinkList *last;

		last = linkList;
		while(last->next != NULL) {
			last = last->next;
		}
		last->next = link;
		link->next = NULL;
	}
}

void PdfRoot::addBookmark(PdfBookmark *newBookmark) {
	if(newBookmark == NULL)
		return;

	if(bookmarks == NULL)
		bookmarks = newBookmark;
	else {
		PdfBookmark *ptr = bookmarks;
		while(ptr->next != NULL) 
			ptr = ptr->next;
		ptr->next = newBookmark;
	}
}

void PdfRoot::addBookmark(String newTitle, PdfPage *newPage) {
	addBookmark(new PdfBookmark(newTitle, newPage));
}

String PdfRoot::pdfFormatDate(time_t newTime) {
	char buffer[256];
	struct tm *instant = localtime(&newTime);
	long tz_hour, tz_min;
	char tz_mod;
	long tz;

/* XXX	This may also work with other *BSDs, but I don't know. -jf 
	This should be changed to a config.h #define like HAS_TIMEZONE.
	Activate the next commented out block when checks are added to
	configure to determine which is present.
*/

#if 0
#ifdef HAS_TIMEZONE
	/* Linux & Solaris way */
	tz = timezone;
#elif HAS_TM_OFF
	/* BSD way */
	tz = instant->tm_gmtoff;
#elif
	/* Is neither */
	tz = 0;
#endif /* !TIMEZONE & !HAS_TM_OFF */
#endif /* 0 */

#ifdef __NetBSD__
	tz = instant->tm_gmtoff;
#else
	tz = timezone;
#endif

	if(tz == 0) {
		tz_mod = 'Z';
		tz_hour = 0;
		tz_min = 0;
	}
	else if(tz < 0) {
		tz_mod = '-';
		tz_hour = -1 * (tz / (60*60));
		tz_min = -1 * ((tz-(tz_hour*60*60)) / (60));
	}
	else {
		tz_mod = '+';
		tz_hour = tz / (60*60);
		tz_min = (tz-(tz_hour*60*60)) / (60);
	}
		
	if(instant == NULL)
		return String();

	sprintf(buffer, "D:%04d%02d%02d%02d%02d%02d%c%02ld'%02ld'", instant->tm_year+1900, instant->tm_mon+1, instant->tm_mday, instant->tm_hour, instant->tm_min, instant->tm_sec, tz_mod, tz_hour, tz_min);
	return String(buffer);
}

void PdfRoot::addColorSpace(PdfColorSpace *newColorSpace) {
	if(newColorSpace == NULL)
		return;

	if(colorSpaces == NULL)
		colorSpaces = new PdfList(newColorSpace);
	else {
		PdfList *ptr = colorSpaces;
		while(ptr->next != NULL) 
			ptr = ptr->next;
		ptr->next = new PdfList(newColorSpace);
	}
}

PdfColorSpace* PdfRoot::getColorSpace(int id) {
	if(colorSpaces == NULL) 
		return NULL;

	PdfList *ptr = colorSpaces;
	while(ptr != NULL) {
		if( ((PdfColorSpace*)ptr->object)->id == id)
			return (PdfColorSpace*) ptr->object;
		ptr = ptr->next;
	}
	return NULL;
}

void PdfRoot::addFunction(PdfFunction *newFunction) {
	if(newFunction == NULL)
		return;

	if(functions == NULL)
		functions = new PdfList(newFunction);
	else {
		PdfList *ptr = functions;
		while(ptr->next != NULL) 
			ptr = ptr->next;
		ptr->next = new PdfList(newFunction);
	}
}

PdfFunction* PdfRoot::getFunction(int id) {
	if(functions == NULL) 
		return NULL;

	PdfList *ptr = functions;
	while(ptr != NULL) {
		if( ((PdfFunction*)ptr->object)->id == id)
			return (PdfFunction*) ptr->object;
		ptr = ptr->next;
	}
	return NULL;
}

//------------------------------------------

PdfPage::PdfPage(): PdfResource(),  contentObj(NULL), annotations(NULL), resourceRefList(NULL), root(NULL) {
	mediaBox[0] = DEFAULT_LLX;
	mediaBox[1] = DEFAULT_LLY;
	mediaBox[2] = DEFAULT_URX;
	mediaBox[3] = DEFAULT_URY;
}

PdfPage::~PdfPage() {
	PdfResourceRef *ptr;
	PdfObject *temp;

	ptr = resourceRefList;
	while(ptr != NULL) {
		temp = ptr;
		ptr = ptr->next;
		delete temp;
	}

	PdfAnnot *annotPtr = annotations;
	while(annotPtr != NULL) {
		temp = annotPtr;
		annotPtr = annotPtr->next;
		delete temp;
	}

	if(contentObj != NULL)
		delete contentObj;
}

PdfObject* PdfPage::copy() {
	PdfPage *beta = new PdfPage();

	beta->copyFields(this, true);
	return beta;
}

PdfObject* PdfPage::copy(PdfRoot *root, bool copyContent) {
	PdfPage *beta = new PdfPage();
	root->addPage(beta);

	beta->copyFields(this, copyContent);
	return beta;
}

void PdfPage::copyFields(PdfPage *alpha, bool copyContent) {
	PdfResource::copyFields(alpha);

	PdfResource *resource = NULL;
	PdfResourceRef *ref = NULL;

	int i;

	for(i=0; i<4; i++)
		mediaBox[i] = alpha->mediaBox[i];

	if(copyContent && alpha->contentObj != NULL) {
		contentObj = alpha->contentObj->copy();
		if(contentObj == NULL)
			printerr("Error copying content objects for PdfPage\n");
		contentObj->setParentPage(this);
	}

	PdfResourceRef *refPtr = alpha->resourceRefList;
	while(refPtr != NULL) {
		ref = (PdfResourceRef*) refPtr->copy();
		if(root != NULL) {
			resource = (PdfResource*) root->resourceIdHash.get(refPtr->resource->resourceId);
			if(resource != NULL) {
				// fprintf(stderr, "Found resource\n");
			}
			else {
				// fprintf(stderr, "Could not find resource\n");
				resource = (PdfResource*)refPtr->resource->copy();
				root->addResource(resource);
			}
			ref->resource = resource;
		}
		addResourceRef(ref);
		refPtr = refPtr->next;
	}
	PdfAnnot *annotPtr = alpha->annotations;
	while(annotPtr != NULL) {
		addAnnotation((PdfAnnot*)annotPtr->copy());
		annotPtr = annotPtr->next;
	}
}

String PdfPage::render_xml(int depth) {
	String str;
	PdfResourceRef *rPtr = resourceRefList;
	char buffer[256];
	PdfAnnot *nextAnnot;

	str += "<page";

	sprintf(buffer, " id=\"%d\"", refId);
	str += buffer;

	str += ">\n";

	while(rPtr != NULL) {
		str += rPtr->render_xml(depth);
		rPtr = rPtr->next;
	}

	if(mediaBox[0] != DEFAULT_LLX || 
	   mediaBox[1] != DEFAULT_LLY || 
	   mediaBox[2] != DEFAULT_URX || 
	   mediaBox[3] != DEFAULT_URY ) {
		sprintf(buffer, "<media_box llx=\"%.0f\" lly=\"%.0f\" urx=\"%.0f\" ury=\"%.0f\"></media_box>\n", mediaBox[0], mediaBox[1], mediaBox[2], mediaBox[3]);
		str += buffer;
	}

	if(annotations != NULL) {
		nextAnnot = annotations;
		while(nextAnnot != NULL) {
			str += nextAnnot->render_xml(depth);
			nextAnnot = nextAnnot->next;
		}
	}

	if(contentObj != NULL) {
		if(OUTPUT_LEVEL > 1) printf("Rendering Page Content\n");
		str += "<contents>";
		contentObj->render_xml(depth, str);
		str += "</contents>\n";
		if(OUTPUT_LEVEL > 1) printf("Done Rendering Page Content\n");
	}
	str += "</page>\n";

	return str;
}

PObj* PdfPage::renderBasicObjects(PObj *rootObj, PObj *parent, PObj *pagesArray) {
	PObj *collection = new PDF_collection();
	PObj *resourceDict = new PDF_dictionary();
	PdfResourceRef *resourceRef;
	PObj *ref, *typeDict, *mediaBoxObj, *contents, *procSet, *annot;
	int i;
	PObj *obj = new PDF_dictionary();
	bool hasImages = false;
	PdfAnnot *nextAnnot;
	PObj *annotRefs;

	if(OUTPUT_LEVEL > 2) printf("Starting PdfPage::renderBasicObjects\n");

	obj->id = id;

	// printf("Adding page %d to hash\n", id);
	((PDF_root*)rootObj)->idHash.add(obj, id);

	((PDF_dictionary*) obj)->setValueForKey("Type", new PDF_name("Page"));

	((PDF_dictionary*) obj)->setValueForKey("Parent", new PDF_ref(parent));
	pagesArray->addsub(new PDF_ref(obj));

	mediaBoxObj = new PDF_array();
	for(i=0; i<4; i++) {
		mediaBoxObj->addsub(new PDF_int(mediaBox[i]));
	}
	((PDF_dictionary*) obj)->setValueForKey("MediaBox", mediaBoxObj);
	
	resourceRef = resourceRefList;
	while(resourceRef != NULL) {
		if(resourceRef->resource == NULL)
			printerr("Error - resource reference pointing to null object\n");
		ref = new PDF_ref();
		((PDF_ref*) ref)->ref = resourceRef->resource->id;

		if(resourceRef->resource->objType() == p_image)
			hasImages = true;

		typeDict = ((PDF_dictionary*)resourceDict)->getValueForKey(resourceRef->type);
		if(typeDict == NULL) {
			typeDict = new PDF_dictionary();
			((PDF_dictionary*) resourceDict)->setValueForKey(resourceRef->type, typeDict);
		}
		((PDF_dictionary*) typeDict)->setValueForKey(resourceRef->name, ref);

		resourceRef = resourceRef->next;
	}
	((PDF_dictionary*) obj)->setValueForKey("Resources", resourceDict);

	procSet = new PDF_array();
	procSet->addsub(new PDF_name("PDF"));
	procSet->addsub(new PDF_name("Text"));
	if(hasImages)
		procSet->addsub(new PDF_name("ImageC"));
	((PDF_dictionary*) resourceDict)->setValueForKey("ProcSet", procSet);

	collection->addsub(obj);

	if(contentObj != NULL) {
		if(contentObj->get_type() == t_shared_collection) {
			PObjList *ptr = contentObj->subobjects;
			PDF_array *array = new PDF_array();
			((PDF_dictionary*) obj)->setValueForKey("Contents", array);

			while(ptr != NULL) {
				if(( ptr->theobject->get_type() == t_collection || 
				     ptr->theobject->get_type() == t_shared_collection) &&
				   ((PDF_collection*)ptr->theobject)->sharedStream) {

					if(((PDF_collection*)ptr->theobject)->streamAlreadyRendered) {
						contentId = ptr->theobject->id;
					}
					else {
						contentId = getNextId();
						ptr->theobject->id = contentId;

						contents = renderBasicObjectsForContents(ptr->theobject);
						collection->addsub(contents);

						((PDF_collection*)ptr->theobject)->streamAlreadyRendered = true; 
					}
				}
				else {
					contentId = getNextId();
					ptr->theobject->id = contentId;

					contents = renderBasicObjectsForContents(ptr->theobject);
					collection->addsub(contents);
				}
				ref = new PDF_ref();
				((PDF_ref*) ref)->ref = contentId;
				array->addsub(ref);

				ptr = ptr->next;
			}
		}
		else {
			if(( contentObj->get_type() == t_collection || 
				 contentObj->get_type() == t_shared_collection) &&
			   ((PDF_collection*) contentObj)->sharedStream) {

				if(((PDF_collection*)contentObj)->streamAlreadyRendered) {
					contentId = contentObj->id;
				}
				else {
					contentId = getNextId();
					contentObj->id = contentId;

					contents = renderBasicObjectsForContents(contentObj);
					collection->addsub(contents);

					((PDF_collection*) contentObj)->streamAlreadyRendered = true; 
				}
			}
			else {
				contentId = getNextId();
				contentObj->id = contentId;

				contents = renderBasicObjectsForContents(contentObj);
				collection->addsub(contents);
			}
			ref = new PDF_ref();
			((PDF_ref*) ref)->ref = contentId;
			((PDF_dictionary*) obj)->setValueForKey("Contents", ref);
		}
	}

	if(annotations != NULL) {
		annotRefs = new PDF_array();
		nextAnnot = annotations;
		while(nextAnnot != NULL) {
			annot = nextAnnot->renderBasicObjects();
			annot->id = getNextId();
			collection->addsub(annot);
			annotRefs->addsub(new PDF_ref(annot));

			if(nextAnnot->objType() == p_link)
				root->addLink( ((PdfLink*)nextAnnot), annot);
			nextAnnot = nextAnnot->next;
		}
		((PDF_dictionary*) obj)->setValueForKey("Annots", annotRefs);
	}
	
	if(OUTPUT_LEVEL > 2) printf("Ending PdfPage::renderBasicObjects\n");

	return collection;
}

PObj* PdfPage::renderBasicObjectsForContents(PObj* obj) {
	PDF_stream *stream = NULL;
	PDF_dictionary *dict;
	PDF_blob *blob;
	PObj *filterArray;
	StringList *nextFilter;

	if(OUTPUT_LEVEL > 2) printf("Starting PdfPage::renderBasicObjectsForContents\n");

	if(obj != NULL) {
		stream = new PDF_stream();
		stream->id = obj->id;

		dict = stream->dictionary;
		blob = stream->dataObject;

		String output;
		obj->render_pdf(NULL, output);

		// Add filter list
		if(filters == NULL) {
			// printf("Content Compression\n");
			if(COMPRESS_DATA > 2) {
				output = DataFilter::encode(output, "FlateDecode");

				filterArray = new PDF_array();
				dict->setValueForKey("Filter", filterArray);
				filterArray->addsub(new PDF_name("FlateDecode"));
			}
		}
		else {
			filterArray = new PDF_array();
			dict->setValueForKey("Filter", filterArray);
			nextFilter = filters;
			while(nextFilter != NULL) {
				filterArray->addsub(new PDF_name(nextFilter->str));
				nextFilter = nextFilter->next;
			}
		}

		blob->setValue(output);
		dict->setValueForKey("Length", new PDF_int( blob->getlength() ));
	}
	if(OUTPUT_LEVEL > 2) printf("Ending PdfPage::renderBasicObjectsForContents\n");
	return stream;
}

void PdfPage::addResourceRef(PdfResourceRef *ref) {
	PdfResourceRef *ptr;

	if(ref == NULL)
		return;

	if(resourceRefList == NULL) {
		resourceRefList = ref;
	}
	else {
		ptr = resourceRefList;
		while(ptr->next != NULL) {
			ptr = ptr->next;
		}
		ptr->next = ref;
	}	
}

PdfResource* PdfPage::getResource(String name) {
	PdfResourceRef *ptr;

	if(resourceRefList == NULL) {
		return NULL;
	}
	else {
		ptr = resourceRefList;
		while(ptr != NULL) {
			if(ptr->name == name)
				return ptr->resource;

			ptr = ptr->next;
		}

		return NULL;
	}	
}

void PdfPage::addAnnotation(PdfAnnot *annot) {
	PdfAnnot *ptr;

	if(annot == NULL)
		return;

	if(annotations == NULL) {
		annotations = annot;
	}
	else {
		ptr = annotations;
		while(ptr->next != NULL) {
			ptr = ptr->next;
		}
		ptr->next = annot;
	}	
}

void PdfPage::addCommand(String name) {
	PObj *obj;

	if(contentObj == NULL) {
		contentObj = new PDF_collection();
	}
	obj = new PDF_pscommand(name);
	contentObj->addsub(obj);
}

void PdfPage::addCommand(PDF_pscommand *command) {
	if(contentObj == NULL) {
		contentObj = new PDF_collection();
	}
	contentObj->addsub(command);
}

void PdfPage::addContent(PObj *node) {
	if(contentObj == NULL) {
		contentObj = new PDF_collection();
		contentObj->parentPage = this;
	}
	contentObj->addsub(node);
}

void PdfPage::enableSharedContent() {
	if(contentObj != NULL && contentObj->get_type() == t_shared_collection) 
		return;

	PDF_shared_collection *sharedData = new PDF_shared_collection();
	sharedData->parentPage = this;

	if(contentObj != NULL)
		sharedData->addsub(contentObj);

	contentObj = sharedData;
}

void PdfPage::switchStream() {
	if(contentObj == NULL)
		return;

	if(contentObj->get_type() != t_shared_collection) 
		enableSharedContent();

	if(contentObj == NULL || contentObj->get_type() != t_shared_collection) 
		printerr("Error creating multiple stream content object\n");

	((PDF_shared_collection*)contentObj)->addNewStream();
}

void PdfPage::addStream(PDF_collection *newStream) {
	enableSharedContent();

	((PDF_shared_collection*)contentObj)->addNewStream(newStream);
	
}

//------------------------------------------

PdfResource::PdfResource() : PdfObject(), prev(NULL), next(NULL), contentId(0), refId(0), data(NULL), dataLength(0), filters(NULL) {
	resourceId = getNextResourceId();
}

PdfResource::~PdfResource() {
	if(data != NULL)
		delete [] data;

	if(filters != NULL)
		delete filters;
}

int PdfResource::getNextResourceId() {
	return nextResourceId++;
}

PdfObject* PdfResource::copy() {
	PdfResource *beta = new PdfResource();

	beta->copyFields(this);
	return beta;
}

void PdfResource::copyFields(PdfResource *alpha) {
	PdfObject::copyFields(alpha);

    resourceId = alpha->resourceId;

    type = alpha->type;
    contentId = alpha->contentId;
    refId = alpha->refId;
 
	dataLength = alpha->dataLength;
	if(dataLength > 0) {
		data = new char[dataLength];
		memcpy(data, alpha->data, dataLength);
	}

	StringList *filterPtr = alpha->filters;
	while(filterPtr != NULL) {
    	addFilter(filterPtr->str);
		filterPtr = filterPtr->next;
	}
 
    filename = alpha->filename;
    encoding = alpha->encoding;
}

String PdfResource::render_xml(int depth) {
	return ("<resource></resource>\n");
}

PObj* PdfResource::renderBasicObjects(PDF_root *root) {
	printf("Unsupported renderBasicObjects Resource: %s\n", type.nullterm());
	return NULL;
}

//------------------------------------------

void PdfResource::decode(String encoding) {
	String output;

	// printf("Decoding %s\n", encoding.nullterm() );

	output.setval(data, dataLength);
	output = DataFilter::decode(output, encoding);

	if(data != NULL)
		delete [] data;
	data = new char[output.getlength()];
	dataLength = output.getlength();
	memcpy(data, output.getrawstring(), output.getlength() );
}

void PdfResource::encode(String encoding) {
	String output;

	// printf("Encoding %s\n", encoding.nullterm() );

	output.setval(data, dataLength);
	output = DataFilter::encode(output, encoding);

	if(data != NULL)
		delete [] data;
	data = new char[output.getlength()];
	dataLength = output.getlength();
	memcpy(data, output.getrawstring(), output.getlength() );
}

void PdfResource::addFilter(String newFilter) {
	StringList *item;

	if(filters == NULL) {
		filters = new StringList(newFilter);
		filters->next = NULL;
	}
	else {
		item = filters;
		while(item->next != NULL) {
			item = item->next;
		}
		item->next = new StringList(newFilter);
		item->next->next = NULL;
	}
}

void PdfResource::decode() {
	StringList *temp;

	while(filters != NULL && DataFilter::isSupportedEncoding( filters->str ) ) {
		decode(filters->str);

		temp = filters;
		filters = filters->next;

		temp->next = NULL;
		delete temp;
	}
}

//------------------------------------------

void PdfResource::loadData() {
	struct stat statInfo;
	size_t fSize;
	FILE *fp;

	if(filename.getlength() < 1)
		printerr("Error - no filename present\n");

	if(DEBUG_LOG_NULLTERM) printf("nullterm() #4.\n");
	char *fname = filename.nullterm();
 
	fp = fopen(fname, "rb");
 
	if(fp == NULL) {
		printerr("Error opening file: %s\n", fname );
	}
	else {
		if(stat(fname, &statInfo) != 0)
			printerr("Could not determine filesize for file %s\n", fname );
	
		fSize = statInfo.st_size;
		data = new char[fSize];
 
		if( fread( data, 1, fSize, fp) != fSize)
			printerr("Error - could not read font file %s\n", fname );

		dataLength = fSize;

		if(fclose(fp) != 0) {
			printerr("Error closing file: %s\n", fname );
		}
	}
 
	delete [] fname;
}

void PdfResource::saveData() {
	if(filename.getlength() < 1) 
		printerr("Error - cannot save data without a filename\n");

	if(DEBUG_LOG_NULLTERM) printf("nullterm() #5.\n");
	char *fbuffer = filename.nullterm();

	FILE *fp = fopen(fbuffer, "wb");

	if(fp == NULL) {
		fprintf(stderr, "Error while trying to open file %s\n", fbuffer);
		exit(1);
	}
	if(fwrite(data, 1, dataLength, fp)!=dataLength) {
		fprintf(stderr, "Error while trying to write file %s\n", fbuffer);
		exit(1);
	}
	fclose(fp);
	delete [] fbuffer;
}

//------------------------------------------

PdfFont::PdfFont() : PdfResource(), firstChar(-1), lastChar(-1), widths(NULL), 
					 numWidths(0), fontDescId(0), ascent(0), descent(0), flags(32) {
	int i;

	type = ("Font");

	for(i=0; i<4; i++) 
		fontBBox[i] = 0;
}

PdfFont::PdfFont(char * newName) : PdfResource(), firstChar(-1), lastChar(-1), widths(NULL), 
					 numWidths(0), fontDescId(0), ascent(0), descent(0), flags(32) {

	type = ("Font");

	int i;

	for(i=0; i<4; i++) 
		fontBBox[i] = 0;

	fontName = newName;
	initBase14Font();
}

PdfFont::~PdfFont() {
	if(widths != NULL)
		delete [] widths;
}

PdfObject* PdfFont::copy() {
	PdfFont *beta = new PdfFont();

	beta->copyFields(this);
	return beta;
}

void PdfFont::copyFields(PdfFont *alpha) {
	PdfResource::copyFields(alpha);

    int i;

    subType = alpha->subType;
    fontName = alpha->fontName;
    firstChar = alpha-> firstChar;
    lastChar = alpha->lastChar;

    numWidths = alpha->numWidths;
	if(numWidths > 0) {
		widths = new long[numWidths];
		memcpy(widths, alpha->widths, sizeof(long) * numWidths);
	}
 
    fontDescId = alpha->fontDescId;
    ascent = alpha->ascent;
    descent = alpha->descent;
    flags = alpha->flags;
    for(i=0; i<4; i++)
        fontBBox[i] = alpha->fontBBox[i];
 
    dataSubType = alpha->dataSubType;
}

void PdfFont::showWidths() {
	int i;

	for(i=0; i<numWidths; i++) {
		printf("Font width %d: %ld\n", i, widths[i]);
	}
}

String PdfFont::render_xml(int depth) {
	String str;
	char buffer[512];
	int i;

	str += "<font";

	sprintf(buffer, " id=\"%d\"", refId);
	str += buffer;

	str += " type=\"";
	str += subType;
	str += "\"";

	if(isBase14Font()) {
		str += " font_name=\"";
		str += fontName;
		str += "\"";

		str += ">";
	}
	else {
		if(dataSubType.getlength() > 0) {
			str += " data_subtype=\"";
			str += dataSubType;
			str += "\"";
		}

		if(filename.getlength() > 0 ) {
		}
		else if(decompose && dataLength > 0) {
			// Save Font to file
			char fbuffer[256];
			if(subType == "TrueType") {
				sprintf(fbuffer, "font_%ld-%d.ttf", time(NULL), fontNum++);
			}
			else {
				sprintf(fbuffer, "font_%ld-%d.fnt", time(NULL), fontNum++);
			}

			filename = fbuffer;
	 
			FILE *fp = fopen(fbuffer, "wb");

			if(fp == NULL) {
				fprintf(stderr, "Error while trying to open file %s\n", fbuffer);
				exit(1);
			}
			if(fwrite(data, 1, dataLength, fp)!=dataLength) {
				fprintf(stderr, "Error while trying to write file %s\n", fbuffer);
				exit(1);
			}
			fclose(fp);
		}

		if(filename.getlength() > 0) {
			str += " filename=\"";
			str += filename;
			str += "\"";
		}

		if(subType == "TrueType" && (filename.getlength() > 0 || dataLength > 0)) {
			str += ">";
			if(filename.getlength() < 1 && dataLength > 0) {
				str += "\n";
			}
		}
		else {
			str += " font_name=\"";
			str += fontName;
			str += "\"";

			str += ">";
		
			if(firstChar != -1) {
				str += "\n";

				str += "<font_metrics";

				sprintf(buffer, " first_char=\"%d\"", firstChar);
				str += buffer;

				sprintf(buffer, " last_char=\"%d\"", lastChar);
				str += buffer;

				sprintf(buffer, " ascent=\"%d\"", ascent);
				str += buffer;

				sprintf(buffer, " descent=\"%d\"", descent);
				str += buffer;

				sprintf(buffer, " flags=\"%d\"", flags);
				str += buffer;

				str += ">";

				if(numWidths > 0) {
					str += "\n<font_widths";
					str += " encoding=\"";
					str += encoding;
					str += "\"";
					str += ">\n";

					for(i=0; i<numWidths; i++) {
						sprintf(buffer, "<width>%ld</width>\n", widths[i]);
						str += buffer;
					}
					str += "</font_widths>\n";
				}

				str += "</font_metrics>\n";

				if(! (fontBBox[0] == 0 && fontBBox[1] == 0 && fontBBox[2] == 0 && fontBBox[3] == 0)) {
					sprintf(buffer, "<font_bounding_box llx=\"%.0f\" lly=\"%.0f\" urx=\"%.0f\" ury=\"%.0f\"></font_bounding_box>\n", fontBBox[0], fontBBox[1], fontBBox[2], fontBBox[3]);
					str += buffer;
				}
			}
		}

		if(filename.getlength() < 1 && dataLength > 0){
			if(0) {
				str += "<font_data encoding=\"Hex\">";
				// str += DataFilter::encodeXML(String(data, dataLength));
				str += DataFilter::encodeXML(DataFilter::encodeHex(String(data, dataLength)));
				str += "</font_data>\n";
			}
			else {
				str += "<font_data encoding=\"Base85\">";
				// str += DataFilter::encodeXML(String(data, dataLength));
				str += DataFilter::encodeXML(DataFilter::encode85(String(data, dataLength)));
				str += "</font_data>\n";
			}
		}
	}

	str += "</font>\n";

	return str;
}

PObj* PdfFont::renderBasicObjects(PDF_root *root) {
	int i;
	PObj *widthsArray, *fontDescRef;
	PObj *collection = new PDF_collection();
	PObj *fontDescriptor;
	PObj *obj = new PDF_dictionary();

	if(OUTPUT_LEVEL > 2) printf("Starting PdfFont::renderBasicObjects\n");

	obj->id = id;
	((PDF_dictionary*) obj)->setValueForKey("Type", new PDF_name("Font"));
	((PDF_dictionary*) obj)->setValueForKey("Subtype", new PDF_name(subType));
	((PDF_dictionary*) obj)->setValueForKey("BaseFont", new PDF_name(fontName));

	collection->addsub(obj);

	if(! isBase14Font()) {
		if(firstChar != -1)
			((PDF_dictionary*) obj)->setValueForKey("FirstChar", new PDF_int(firstChar));
		if(lastChar != -1)
			((PDF_dictionary*) obj)->setValueForKey("LastChar", new PDF_int(lastChar));

		if(encoding != "")
			((PDF_dictionary*) obj)->setValueForKey("Encoding", new PDF_name(encoding));

		if(widths != NULL) {
			widthsArray = new PDF_array();
			((PDF_dictionary*) obj)->setValueForKey("Widths", widthsArray);
			numWidths = lastChar - firstChar + 1;
			for(i=0; i<numWidths; i++) {
				widthsArray->addsub(new PDF_int(widths[i]));
			}
		}

		fontDescId = getNextId();
		fontDescriptor = renderBasicObjectsForFontDescriptor();
		collection->addsub(fontDescriptor);

		fontDescRef = new PDF_ref();
		((PDF_ref*) fontDescRef)->ref = fontDescId;
		((PDF_dictionary*) obj)->setValueForKey("FontDescriptor", fontDescRef);
	}

	if(OUTPUT_LEVEL > 2) printf("Ending PdfFont::renderBasicObjects\n");

	return collection;
}

bool PdfFont::isBase14Font() {

	if(fontName == "Courier" || 
	   fontName == "Courier-Bold" || 
	   fontName == "Courier-Oblique" || 
	   fontName == "Courier-BoldOblique" || 
	   fontName == "Helvetica" || 
	   fontName == "Helvetica-Bold" || 
	   fontName == "Helvetica-Oblique" || 
	   fontName == "Helvetica-BoldOblique" || 
	   fontName == "Symbol" || 
	   fontName == "Times-Roman" || 
	   fontName == "Times-Bold" || 
	   fontName == "Times-Italic" || 
	   fontName == "Times-BoldItalic" || 
	   fontName == "ZapfDingbats") {
		return true;
	}
	else {
		return false;
	}
}

PObj* PdfFont::renderBasicObjectsForFontDescriptor() {
	PObj *collection = new PDF_collection();
	PObj *obj = new PDF_dictionary();
	PObj *bBox;
	int i;
	PDF_dictionary *dict;
	PDF_blob *blob;
	PObj *filterArray;
	StringList *nextFilter;

	if(OUTPUT_LEVEL > 2) printf("Starting PdfFontDescriptor::renderBasicObjects\n");

	obj->id = fontDescId;

	((PDF_dictionary*) obj)->setValueForKey("Type", new PDF_name("FontDescriptor"));
	((PDF_dictionary*) obj)->setValueForKey("FontName", new PDF_name(fontName));
	((PDF_dictionary*) obj)->setValueForKey("Ascent", new PDF_int(ascent));
	((PDF_dictionary*) obj)->setValueForKey("CapHeight", new PDF_int(0));
	((PDF_dictionary*) obj)->setValueForKey("Descent", new PDF_int(descent));
	((PDF_dictionary*) obj)->setValueForKey("Flags", new PDF_int(flags));

	bBox = new PDF_array();
	for(i=0; i<4; i++) {
		bBox->addsub(new PDF_int(fontBBox[i]));
	}
	((PDF_dictionary*) obj)->setValueForKey("FontBBox", bBox);

	((PDF_dictionary*) obj)->setValueForKey("ItalicAngle", new PDF_int(0));
	((PDF_dictionary*) obj)->setValueForKey("StemV", new PDF_int(0));
	collection->addsub(obj);

	if(filename.getlength() > 0 && dataLength < 1) {
		loadData();
	}

	if(dataLength > 0) {
		PDF_stream* stream = new PDF_stream();

		stream->id = getNextId();
		contentId = stream->id;

		dict = stream->dictionary;
		blob = stream->dataObject;

		dict->setValueForKey("Length1", new PDF_int(dataLength));

		if(COMPRESS_DATA > 1 && filters == NULL) {
			// printf("Testing Font Compression\n");
			String output;
			output.setval(data, dataLength);
			output = DataFilter::encode(output, "FlateDecode");
			addFilter("FlateDecode");
	 
			blob->setValue((char*)(output.getrawstring()), output.getlength() );
			dict->setValueForKey("Length", new PDF_int(output.getlength()));
		}
		else {
			blob->setValue(data, dataLength);
			dict->setValueForKey("Length", new PDF_int(dataLength));
		}
	 
		// Add filter list
		if(filters != NULL) {
			filterArray = new PDF_array();
			dict->setValueForKey("Filter", filterArray);
			nextFilter = filters;
			while(nextFilter != NULL) {
				filterArray->addsub(new PDF_name(nextFilter->str));
				nextFilter = nextFilter->next;
			}
		} 

		collection->addsub(stream);

		if(subType == "Type1") {
			if(dataSubType.getlength() > 0) {
				dict->setValueForKey("Subtype", new PDF_name(dataSubType));
				((PDF_dictionary*) obj)->setValueForKey("FontFile3", new PDF_ref(stream));
			}
			else {
				((PDF_dictionary*) obj)->setValueForKey("FontFile", new PDF_ref(stream));
			}
		}
		else if(subType == "TrueType") {
			((PDF_dictionary*) obj)->setValueForKey("FontFile2", new PDF_ref(stream));
		}
		else {
			printerr("Error - unknown font subType - %s\n", subType.nullterm() );
		}
	}

	if(OUTPUT_LEVEL > 2) printf("Ending PdfFontDescriptor::renderBasicObjects\n");

	return collection;
}

//------------------------------------------

PdfColorSpace::PdfColorSpace() : PdfResource(), base(NULL), baseId(0), hival(0), altSpace(NULL), tintTransform(NULL), altSpaceId(0), tintTransformId(0), isPageResource(false) {
	int i, j;

	type = "ColorSpace";

	for(i=0; i<3; i++) 
		whitePoint[i] = 1;
	for(i=0; i<3; i++) 
		blackPoint[i] = 0;
	for(i=0; i<3; i++) 
		gamma[i] = 1;
	for(i=0; i<3; i++) {
		for(j=0; j<3; j++) {
			if(i==j)
				matrix[3*i+j] = 1;
			else
				matrix[3*i+j] = 0;
		}
	}
	for(i=0; i<4; i++) {
		if(i % 2) 
			range[i] = 100;
		else
			range[i] = -100;
	}
}

PdfColorSpace::PdfColorSpace(char *newCatagory) : PdfResource(), base(NULL), baseId(0), hival(0), altSpace(NULL), tintTransform(NULL), altSpaceId(0), tintTransformId(0), isPageResource(false) {
	int i, j;

	type = "ColorSpace";

	for(i=0; i<3; i++) 
		whitePoint[i] = 1;
	for(i=0; i<3; i++) 
		blackPoint[i] = 0;
	for(i=0; i<3; i++) 
		gamma[i] = 1;
	for(i=0; i<3; i++) {
		for(j=0; j<3; j++) {
			if(i==j)
				matrix[3*i+j] = 1;
			else
				matrix[3*i+j] = 0;
		}
	}
	for(i=0; i<4; i++) {
		if(i % 2) 
			range[i] = 100;
		else
			range[i] = -100;
	}

	catagory = newCatagory;
}

PdfColorSpace::~PdfColorSpace() {
	if(base != NULL)
		delete base;

	if(altSpace != NULL)
		delete altSpace;

	if(tintTransform != NULL)
		delete tintTransform;
}

PdfObject* PdfColorSpace::copy() {
	PdfColorSpace *beta = new PdfColorSpace();

	beta->copyFields(this);
	return beta;
}

void PdfColorSpace::copyFields(PdfColorSpace *alpha) {
	PdfResource::copyFields(alpha);

	int i;

    catagory = alpha->catagory;
    base = (PdfColorSpace*) alpha->base->copy();

    for(i=0; i<3; i++)
        whitePoint[i] = alpha->whitePoint[i];
    for(i=0; i<3; i++)
        blackPoint[i] = alpha->blackPoint[i];
    for(i=0; i<3; i++)
        gamma[i] = alpha->gamma[i];
    for(i=0; i<9; i++)
        matrix[i] = alpha->matrix[i];
    for(i=0; i<4; i++)
        range[i] = alpha->range[i];
 
    hival = alpha->hival;
    lookup = alpha->lookup;

    separationName = alpha->separationName;
    altSpace = (PdfColorSpace*) alpha->altSpace->copy();
    altSpaceId = alpha->altSpaceId;
    tintTransform = (PdfFunction*) alpha->tintTransform->copy();
    tintTransformId = alpha->tintTransformId;
}

String PdfColorSpace::render_xml(int depth) {
	char buffer[512];
	String result;

	result += "<colorspace";

	if(refId > 0) {
		sprintf(buffer, " id=\"%d\"", refId);
		result += buffer;
	}
		
	if(isPageResource == true)
		result += " is_page_resource=\"true\"";

	if(catagory.getlength() > 0) {
		result += " catagory=\"";
		result += catagory;
		result += "\"";
	}
		
	if(catagory == "CalGray" || catagory == "CalRGB" || catagory == "Lab") {
		sprintf(buffer, " white_point=\"%.2f,%.2f,%.2f\"", whitePoint[0], whitePoint[1], whitePoint[2]);
		result += buffer;
		sprintf(buffer, " black_point=\"%.2f,%.2f,%.2f\"", blackPoint[0], blackPoint[1], blackPoint[2]);
		result += buffer;
	
		if(catagory == "CalRGB") {
			sprintf(buffer, " gamma=\"%.2f,%.2f,%.2f\"", gamma[0], gamma[1], gamma[2]);
			result += buffer;
			sprintf(buffer, " matrix=\"%.2f,%.2f,%.2f,%.2f,%.2f,%.2f,%.2f,%.2f,%.2f\"", matrix[0], matrix[1], matrix[2], matrix[3], matrix[4], matrix[5], matrix[6], matrix[7], matrix[8]);
			result += buffer;
		}
		else if(catagory == "Lab") {
			sprintf(buffer, " range=\"%.2f,%.2f,%.2f,%.2f\"", range[0], range[1], range[2], range[3]);
			result += buffer;
		}
	}
	else if(catagory == "Indexed") {
		if(base->isSimpleCatagory()) {
			result += " base=\"";
			result += base->catagory;
			result += "\"";
		}
		else {
			base->refId = getNextResourceId();
			sprintf(buffer, " base_id=\"%d\"", base->refId);
			result += buffer;
		}

		sprintf(buffer, " hival=\"%d\"", hival);
		result += buffer;
	}
	else if(catagory == "Separation") {
		result += " separation_name=\"";
		result += separationName;
		result += "\"";

		if(altSpace->isSimpleCatagory()) {
			result += " alt_space=\"";
			result += altSpace->catagory;
			result += "\"";
		}
		else {
			altSpace->refId = getNextResourceId();
			sprintf(buffer, " alt_space_id=\"%d\"", altSpace->refId);
			result += buffer;
		}

		tintTransform->refId = functionNum++;
		sprintf(buffer, " tint_transform_id=\"%d\"", tintTransform->refId);
		result += buffer;
	}


	if(catagory == "Indexed" && dataLength > 0) {
		if(filters != NULL) {
			result += " filter=\"";
			StringList *filter = filters;
			while(filter != NULL) {
				result += filter->str;
				if(filter->next != NULL)
					result += ",";
				filter = filter->next;
			}
			result += "\"";
		}

		result += " encoding=\"Base85\">";
		result += DataFilter::encodeXML(DataFilter::encode85(String(data, dataLength)));
	}
	else {
		result += ">";
	}

	result += "</colorspace>\n";

	if(catagory == "Indexed" && ! base->isSimpleCatagory()) 
		result += base->render_xml(depth);
	else if(catagory == "Separation") {
		if( ! altSpace->isSimpleCatagory()) 
			result += altSpace->render_xml(depth);
		result += tintTransform->render_xml(depth);
	}

	return result;
}

PObj* PdfColorSpace::renderBasicObjects(PDF_root *root) {
	PObj *result = NULL;

	if(catagory == "DeviceGray" || catagory == "DeviceRGB" || catagory == "DeviceCMYK") {
		result = new PDF_name(catagory);
		result->id = id;
	}
	else  {
		PDF_array *array = new PDF_array();
		array->addsub(new PDF_name(catagory));

		result = array;
		result->id = id;

		if(catagory == "CalGray" || catagory == "CalRGB" || catagory == "Lab") {
			PDF_dictionary *dict = new PDF_dictionary();
			array->addsub(dict);

			array = new PDF_array();
			array->addsub(new PDF_real(whitePoint[0]));
			array->addsub(new PDF_real(whitePoint[1]));
			array->addsub(new PDF_real(whitePoint[2]));
			dict->setValueForKey("WhitePoint", array);

			if(blackPoint[0] != 0 || blackPoint[1] != 0 || blackPoint[2] != 0) {
				array = new PDF_array();
				array->addsub(new PDF_real(blackPoint[0]));
				array->addsub(new PDF_real(blackPoint[1]));
				array->addsub(new PDF_real(blackPoint[2]));
				dict->setValueForKey("BlackPoint", array);
			}

			if(catagory == "CalRGB") {
				if(gamma[0] != 1 || gamma[1] != 1 || gamma[2] != 1) {
					array = new PDF_array();
					array->addsub(new PDF_real(gamma[0]));
					array->addsub(new PDF_real(gamma[1]));
					array->addsub(new PDF_real(gamma[2]));
					dict->setValueForKey("Gamma", array);
				}

				if(matrix[0] != 1 || matrix[1] != 0 || matrix[2] != 0 ||
				   matrix[3] != 0 || matrix[4] != 1 || matrix[5] != 0 ||
				   matrix[6] != 0 || matrix[7] != 0 || matrix[8] != 1) {
					array = new PDF_array();
					array->addsub(new PDF_real(matrix[0]));
					array->addsub(new PDF_real(matrix[1]));
					array->addsub(new PDF_real(matrix[2]));
					array->addsub(new PDF_real(matrix[3]));
					array->addsub(new PDF_real(matrix[4]));
					array->addsub(new PDF_real(matrix[5]));
					array->addsub(new PDF_real(matrix[6]));
					array->addsub(new PDF_real(matrix[7]));
					array->addsub(new PDF_real(matrix[8]));
					dict->setValueForKey("Matrix", array);
				}
			}
			else if(catagory == "Lab") {
				if(range[0] != -100 || range[1] != 100 || range[2] != -100 || range[3] != 100) {
					array = new PDF_array();
					dict->setValueForKey("Range", array);
				}
			}
		}
		else if(catagory == "Indexed") {

			if(base == NULL)
				printerr("Error - cannot render basic objects for indexed colorspace without base oject\n");
			base->id = 0;
			array->addsub(base->renderBasicObjects(root));
			array->addsub(new PDF_int(hival));

			PDF_stream *stream = new PDF_stream();
			stream->id = getNextId();

			// Add filter list
			if(filters != NULL) {
				PDF_array *filterArray = new PDF_array();
				StringList *nextFilter = filters;

				while(nextFilter != NULL) {
					filterArray->addsub(new PDF_name(nextFilter->str));
					nextFilter = nextFilter->next;
				}
				stream->dictionary->setValueForKey("Filter", filterArray);
			}

			stream->dataObject->setValue(data, dataLength);
			array->addsub(new PDF_ref(stream));
			root->addsub(stream);
		}
		else if(catagory == "Separation") {
			array->addsub(new PDF_name(separationName));

			if(altSpace == NULL)
				printerr("Error - missing altSpace\n");
			altSpace->id = 0;
			PObj* altSpaceObj = altSpace->renderBasicObjects(root);
			altSpaceObj->id = getNextId();
			array->addsub(new PDF_ref(altSpaceObj));
			root->addsub(altSpaceObj);

			if(tintTransform == NULL)
				printerr("Error - missing tintTransform\n");
			PObj* tintTransformObj = tintTransform->renderBasicObjects();
			tintTransformObj->id = getNextId();
			array->addsub(new PDF_ref(tintTransformObj));
			root->addsub(tintTransformObj);
		}
		else if(catagory == "Pattern") {
			delete array;
			printerr("Pattern ColorSpaces not yet supported\n");
			result = NULL;
		}
		else {
			delete array;
			printerr("Error rendering unknown ColorSpace catagory: %s\n", catagory.nullterm());
			result = NULL;
		}

	}

	return result;
}

bool PdfColorSpace::isSimpleCatagory() {
	if(catagory == "DeviceGray" || catagory == "DeviceRGB" || catagory == "DeviceCMYK") 
		return true;
	else
		return false;
}

//------------------------------------------

PdfImage::PdfImage() : PdfResource(), width(0), height(0), bitsPerPixel(0), colorSpace(NULL), colorSpaceId(0) {
	type = ("XObject");
}

PdfImage::~PdfImage() {
	if(colorSpace != NULL)
		delete colorSpace;
}

PdfObject* PdfImage::copy() {
	PdfImage *beta = new PdfImage();

	beta->copyFields(this);
	return beta;
}

void PdfImage::copyFields(PdfImage *alpha) {
	PdfResource::copyFields(alpha);

    format = alpha->format;
    width = alpha->width;
    height = alpha->height;
    bitsPerPixel = alpha->bitsPerPixel;
    name = alpha->name;

	if(alpha->colorSpace != NULL)
    	colorSpace = (PdfColorSpace*)alpha->colorSpace->copy();

    colorSpaceId = alpha->colorSpaceId;
}

String PdfImage::render_xml(int depth) {
	String str;
	char buffer[256];
	StringList *imageFilter;

	if(colorSpace != NULL && ! colorSpace->isSimpleCatagory()) {
		colorSpace->refId = getNextResourceId();
		str += colorSpace->render_xml(depth);
	}
		
	str += "<image";

	sprintf(buffer, " id=\"%d\"", refId);
	str += buffer;

	if(format == "native") {
		sprintf(buffer, " width=\"%d\"", width);
		str += buffer;

		sprintf(buffer, " height=\"%d\"", height);
		str += buffer;

		sprintf(buffer, " bits_per_pixel=\"%d\"", bitsPerPixel);
		str += buffer;

		if(colorSpace != NULL) {
			if(colorSpace->isSimpleCatagory()) {
				str += " colorspace=\"";
				str += colorSpace->catagory;
				str += "\"";
			}
			else {
				sprintf(buffer, " colorspace_id=\"%d\"", colorSpace->refId);
				str += buffer;
			}
		}

		str += " format=\"";
		str += format;
		str += "\"";
	}

	if(filters != NULL) {
		str += " filter=\"";
		imageFilter = filters;
		while(imageFilter != NULL) {
			str += imageFilter->str;
			if(imageFilter->next != NULL)
				str += ",";
			imageFilter = imageFilter->next;
		}
		str += "\"";
	}

	if(filename.getlength() > 0 ) {
	}
	else if(decompose && dataLength > 0) {
		// Save Font to file
		char fbuffer[256];
		sprintf(fbuffer, "image_%ld-%d.img", time(NULL), imageNum++);

		filename = fbuffer;
 
		FILE *fp = fopen(fbuffer, "wb");

		if(fp == NULL) {
			fprintf(stderr, "Error while trying to open file %s\n", fbuffer);
			exit(1);
		}
		if(fwrite(data, 1, dataLength, fp)!=dataLength) {
			fprintf(stderr, "Error while trying to write file %s\n", fbuffer);
			exit(1);
		}
		fclose(fp);
	}

	if(filename.getlength() > 0) {
		str += " filename=\"";
		str += filename;
		str += "\"";
		str += ">";
	}
	else if(dataLength > 0){
		if(0) {
			str += " encoding=\"Hex\">";
			// str += DataFilter::encodeXML(String(data, dataLength));
			str += DataFilter::encodeXML(DataFilter::encodeHex(String(data, dataLength)));
		}
		else {
			str += " encoding=\"Base85\">";
			// str += DataFilter::encodeXML(String(data, dataLength));
			str += DataFilter::encodeXML(DataFilter::encode85(String(data, dataLength)));
		}
	}
	else {
		str += ">";
	}

	str += "</image>\n";

	return str;
}

PObj* PdfImage::renderBasicObjects(PDF_root *root) {
	PDF_stream *stream = new PDF_stream();
	PDF_dictionary *dict;
	PDF_blob *blob;
	PObj *filterArray, *obj;
	StringList *nextFilter;
 
	if(OUTPUT_LEVEL > 2) printf("Starting PdfImage::renderBasicObjects\n");
 
	stream->id = id;
 
	// Add attributes to dictionary
	dict = stream->dictionary;
	dict->setValueForKey("Type", new PDF_name("XObject"));
	dict->setValueForKey("Subtype", new PDF_name("Image"));

	if(name.getlength() > 0)
		dict->setValueForKey("Name", new PDF_name(name));
 
	dict->setValueForKey("Height", new PDF_int(height));
	dict->setValueForKey("Width", new PDF_int(width));
	dict->setValueForKey("BitsPerComponent", new PDF_int(bitsPerPixel));
 
	if(colorSpace != NULL) {
		colorSpace->id = 0;
		obj = colorSpace->renderBasicObjects(root);
		if(obj != NULL)
			dict->setValueForKey("ColorSpace", obj);
	}
	else {
		printerr("Error - empty colorspace for image\n");
	}
 
	// Get data from file if nessecary
	if(filename.getlength() > 0 && dataLength < 1) {
		loadData();
	}
 
	dict->setValueForKey("Length", new PDF_int(dataLength));
 
	blob = stream->dataObject;

	if(COMPRESS_DATA > 0 && filters == NULL) {
		// printf("Testing Image Compression\n");
		String output;
		output.setval(data, dataLength);
		output = DataFilter::encode(output, "FlateDecode");
		addFilter("FlateDecode");
 
		blob->setValue((char*)(output.getrawstring()), output.getlength() );
	}
	else {
		blob->setValue(data, dataLength);
	}
 
	// Add filter list
	if(filters != NULL) {
		filterArray = new PDF_array();
		dict->setValueForKey("Filter", filterArray);
		nextFilter = filters;
		while(nextFilter != NULL) {
			filterArray->addsub(new PDF_name(nextFilter->str));
			nextFilter = nextFilter->next;
		}
	}
 
	if(OUTPUT_LEVEL > 2) printf("Ending PdfImage::renderBasicObjects\n");
 
	return stream;
}

//------------------------------------------

PdfResourceRef::PdfResourceRef() : PdfObject(), prev(NULL), next(NULL), resource(NULL), resourceId(0) {
}

PdfObject* PdfResourceRef::copy() {
	return copy(false);
}

PdfObject* PdfResourceRef::copy(bool copyResourcePointer) {
	PdfResourceRef *beta = new PdfResourceRef();

	beta->copyFields(this, copyResourcePointer);
	return beta;
}

void PdfResourceRef::copyFields(PdfResourceRef *alpha, bool copyResourcePointer) {
	PdfObject::copyFields(alpha);

	if(alpha->resource != NULL) {
		if(copyResourcePointer)
			resource = alpha->resource;
		else
			resourceId = alpha->resource->resourceId;
	}

    name = alpha->name;
    type = alpha->type;
}

String PdfResourceRef::render_xml(int depth) {
	String str;
	char buffer[256];

	str += "<page_resource";

	if(resource == NULL)
		printerr("Error rendering resource ref - no resource found\n");

	sprintf(buffer, " id=\"%d\"", resource->refId);
	str += buffer;

	str += " name=\"";
	str += name;
	str += "\"";

	str += "></page_resource>\n";

	return str;
}

PObj* PdfResourceRef::renderBasicObjects() {
	return NULL;
}

//------------------------------------------

PdfAnnot::PdfAnnot() : PdfObject(), prev(NULL), next(NULL) {
	rect[0] = 0;
	rect[1] = 0;
	rect[2] = 0;
	rect[3] = 0;
}

PdfObject* PdfAnnot::copy() {
	PdfAnnot *beta = new PdfAnnot();

	beta->copyFields(this);
	return beta;
}

void PdfAnnot::copyFields(PdfAnnot *alpha) {
	PdfObject::copyFields(alpha);

	int i;

    type = alpha->type;
	for(i=0; i<4; i++)
    	rect[i] = alpha->rect[i];
}

String PdfAnnot::render_xml(int depth) {
	String str;
	char buffer[256];

	str += "<annotation";

	str += " type=\"";
	str += type;
	str += "\"";

	sprintf(buffer, " llx=\"%f\"", rect[0]);
	str += buffer;
	sprintf(buffer, " lly=\"%f\"", rect[1]);
	str += buffer;
	sprintf(buffer, " urx=\"%f\"", rect[2]);
	str += buffer;
	sprintf(buffer, " ury=\"%f\"", rect[3]);
	str += buffer;

	str += ">";

	str += "</annotation>\n";

	return str;
}

PObj* PdfAnnot::renderBasicObjects() {
	PObj *obj;
	int i;
	PObj *annotRect;

	obj = new PDF_dictionary();
	((PDF_dictionary*) obj)->setValueForKey("Type", new PDF_name("Annot"));
	((PDF_dictionary*) obj)->setValueForKey("Subtype", new PDF_name(type));

	annotRect = new PDF_array();
	for(i=0; i<4; i++) {
		annotRect->addsub(new PDF_int(rect[i]));
	}
	((PDF_dictionary*) obj)->setValueForKey("Rect", annotRect);

	return obj;
}

//------------------------------------------

PdfTextAnnot::PdfTextAnnot() : PdfAnnot(), isOpen(false) {
	type = "Text";
}

PdfObject* PdfTextAnnot::copy() {
	PdfTextAnnot *beta = new PdfTextAnnot();

	beta->copyFields(this);
	return beta;
}

void PdfTextAnnot::copyFields(PdfTextAnnot *alpha) {
	PdfAnnot::copyFields(alpha);

    contents = alpha->contents;
    isOpen = alpha->isOpen;
}

String PdfTextAnnot::render_xml(int depth) {
	String str;
	char buffer[256];

	str += "<text_annotation";

	sprintf(buffer, " llx=\"%f\"", rect[0]);
	str += buffer;
	sprintf(buffer, " lly=\"%f\"", rect[1]);
	str += buffer;
	sprintf(buffer, " urx=\"%f\"", rect[2]);
	str += buffer;
	sprintf(buffer, " ury=\"%f\"", rect[3]);
	str += buffer;

	if(isOpen)
		str += " open=\"true\"";

	str += ">";

	str += contents;

	str += "</text_annotation>\n";

	return str;
}

PObj* PdfTextAnnot::renderBasicObjects() {
	PObj *obj;
	PDF_boolean *openObj = new PDF_boolean();
	PDF_string *msg = new PDF_string();

	msg->value = contents;
	openObj->value = isOpen;

	obj = PdfAnnot::renderBasicObjects();

	((PDF_dictionary*) obj)->setValueForKey("Contents", msg);
	((PDF_dictionary*) obj)->setValueForKey("Open", openObj);

	return obj;
}

//------------------------------------------

PdfLink::PdfLink(PdfRoot *root) : PdfAnnot(), destination(NULL) {
	type = "Link";
	rootObj = root;
}

PdfObject* PdfLink::copy() {
	PdfLink *beta = new PdfLink(NULL);

	beta->copyFields(this);
	return beta;
}

void PdfLink::copyFields(PdfLink *alpha) {
	PdfAnnot::copyFields(alpha);

    destKey = alpha->destKey;
    // destination;
}

String PdfLink::render_xml(int depth) {
	String str;
	char buffer[256];

	str += "<link";

	if(destination == NULL) 
		printerr("Error - link found with no destination\n");
	sprintf(buffer, " page=\"%d\"", destination->refId);
	str += buffer;

	sprintf(buffer, " llx=\"%f\"", rect[0]);
	str += buffer;
	sprintf(buffer, " lly=\"%f\"", rect[1]);
	str += buffer;
	sprintf(buffer, " urx=\"%f\"", rect[2]);
	str += buffer;
	sprintf(buffer, " ury=\"%f\"", rect[3]);
	str += buffer;

	str += ">";

	str += "</link>\n";

	return str;
}

PObj* PdfLink::renderBasicObjects() {
	// see "PdfLink -- setting page links" in PdfRoot::renderBasicObjects()
	PObj *link = PdfAnnot::renderBasicObjects();

	return link;
}

//------------------------------------------

PdfBookmark::PdfBookmark() : PdfObject(), destination(NULL), parent(NULL), prev(NULL), next(NULL), first(NULL), last(NULL), isOpen(false), count(0) {
}

PdfBookmark::PdfBookmark(String newTitle, PdfPage *newPage) : PdfObject(), parent(NULL), prev(NULL), next(NULL), first(NULL), last(NULL), isOpen(false), count(0) {
	title = newTitle;
	destination = newPage;
}

PdfBookmark::~PdfBookmark() {
	PdfBookmark *ptr, *temp;

	ptr = first;
	while(ptr != NULL) {
		temp = ptr;
		ptr = ptr->next;
		delete ptr;
	}
}

PdfObject* PdfBookmark::copy() {
	PdfBookmark *beta = new PdfBookmark();

	beta->copyFields(this);
	return beta;
}

void PdfBookmark::copyFields(PdfBookmark *alpha) {
	PdfObject::copyFields(alpha);

    title = alpha->title;
    destKey = alpha->destKey;

    // destination;

	PdfBookmark *bookmarkPtr = alpha->first;
	while(bookmarkPtr != NULL) {
		addBookmark((PdfBookmark*)bookmarkPtr->copy());
		bookmarkPtr = bookmarkPtr->next;
	}

    isOpen = alpha->isOpen;
    count = alpha->count;
}

String PdfBookmark::render_xml(int depth) {
	String str;
	char buffer[256];
	PdfBookmark *bPtr = first;

	str += "<bookmark";

	str += " title=\"";
	str += title;
	str += "\"";

	if(destination == NULL) 
		printerr("Error - bookmark found with no destination\n");
	sprintf(buffer, " page=\"%d\"", destination->refId);
	str += buffer;

	if(isOpen)
		str += " status=\"open\"";

	str += ">";

	if(bPtr != NULL) { 
		str += "\n";
		while(bPtr != NULL) {
			str += bPtr->render_xml(depth+1);
			bPtr = bPtr->next;
		}
	}

	str += "</bookmark>\n";

	return str;
}

PObj* PdfBookmark::renderBasicObjects() {
	printerr("Error - invalid call to PdfBookmark::renderBasicObjects()\n");

	return NULL;
}

PObj* PdfBookmark::renderBasicObjects(PObj *root) {
	PDF_dictionary *bookmarkRoot = new PDF_dictionary();

	PDF_array *dest = new PDF_array();

	if(destination != NULL) {
		dest->addsub( new PDF_ref(destination->id) );
		dest->addsub( new PDF_name(String("XYZ")) );
		dest->addsub(new PDF_int(0) );
		dest->addsub(new PDF_int( destination->mediaBox[3] ) );
		dest->addsub(new PDF_null() );
	}

	bookmarkRoot->setValueForKey("Title", new PDF_string(title));
	bookmarkRoot->setValueForKey("Dest", dest);

	bookmarkRoot->id = getNextId();
	root->addsub(bookmarkRoot);

	count = 0;
	// Sub-entries 
	if(first != NULL) {
		PObj *bookmark, *prevBookmark;

		PdfBookmark *ptr = first;
		prevBookmark = NULL;
		while(ptr != NULL) {
			bookmark = ptr->renderBasicObjects(root);
			((PDF_dictionary*)bookmark)->setValueForKey("Parent", new PDF_ref(bookmarkRoot));

			if(prevBookmark == NULL) {
				((PDF_dictionary*)bookmarkRoot)->setValueForKey("First", new PDF_ref(bookmark));
			}
			else {
				((PDF_dictionary*)prevBookmark)->setValueForKey("Next", new PDF_ref(bookmark));
				((PDF_dictionary*)bookmark)->setValueForKey("Prev", new PDF_ref(prevBookmark));
			}

			count++;

			count += ptr->count;
			prevBookmark = bookmark;
			ptr = ptr->next;
		}
		((PDF_dictionary*)bookmarkRoot)->setValueForKey("Last", new PDF_ref(prevBookmark));

		if(count > 0) {
			if(isOpen)
				bookmarkRoot->setValueForKey("Count", new PDF_int(count));
			else
				bookmarkRoot->setValueForKey("Count", new PDF_int(-count));
		}
	}

	return bookmarkRoot;
}

void PdfBookmark::addBookmark(PdfBookmark *newBookmark) {
	if(newBookmark == NULL)
		return;

	if(first == NULL) {
		first = newBookmark;
		last = newBookmark;
	}
	else {
		PdfBookmark *ptr = first;
		while(ptr->next != NULL) 
			ptr = ptr->next;
		ptr->next = newBookmark;
		last = newBookmark;
	}
}

void PdfBookmark::addBookmark(String newTitle, PdfPage *newPage) {
	addBookmark(new PdfBookmark(newTitle, newPage));
}

void PdfBookmark::findPages(PdfRoot *root) {
	if(destKey.getlength() > 0) {
		destination = (PdfPage*) root->pageHash.get(destKey);
	}
	
	PdfBookmark *ptr = first;
	while(ptr != NULL) {
		ptr->findPages(root);
		ptr = ptr->next;
	}
}

//------------------------------------------

PdfFunction::PdfFunction() : PdfResource(), functionType(0), numIn(0), domain(NULL), numOut(0), range(NULL), size(NULL), bitsPerSample(0), order(1), encode(NULL), decode(NULL) {
}

PdfFunction::~PdfFunction() {
	if(domain != NULL)
		delete [] domain;
	if(range != NULL)
		delete [] range;
	if(size != NULL)
		delete [] size;
	if(encode != NULL)
		delete [] encode;
	if(decode != NULL)
		delete [] decode;
}

PdfObject* PdfFunction::copy() {
	PdfFunction *beta = new PdfFunction();

	beta->copyFields(this);
	return beta;
}

void PdfFunction::copyFields(PdfFunction* alpha) {
	PdfResource::copyFields(alpha);

	int i;

	functionType = alpha->functionType;

	numIn = alpha->numIn;
	domain = new double[2*numIn];
	for(i=0; i<2*numIn; i++) 
		domain[i] = alpha->domain[i];

	numOut = alpha->numOut;
	range = new double[2*numOut];
	for(i=0; i<2*numOut; i++) 
		range[i] = alpha->range[i];

	size = new int[numIn];
	for(i=0; i<numIn; i++) 
		size[i] = alpha->size[i];

    bitsPerSample = alpha->bitsPerSample;
    order = alpha->order;

	encode = new double[2*numIn];
	for(i=0; i<2*numIn; i++) 
		encode[i] = alpha->encode[i];

	decode = new double[2*numOut];
	for(i=0; i<2*numOut; i++) 
		decode[i] = alpha->decode[i];

/*
	int dataLength = 1;
	for(i=0; i<numIn; i++) 
		dataLength *= size[i];
	dataLength *= numOut;
	dataLength *= bitsPerSample;
	if(dataLength % 8 > 0)
		dataLength += 8;
	dataLength /= 8;
*/

}

String PdfFunction::render_xml(int depth) {
	String str;
	char buffer[256];
	StringList *imageFilter;
	int i;

	str += "<function";

	sprintf(buffer, " id=\"%d\"", refId);
	str += buffer;

	sprintf(buffer, " funtion_type=\"%d\"", functionType);
	str += buffer;

	str += " domain=\"";
	for(i=0; i<2*numIn; i++) {
		if(i < 2*numIn-1)
			sprintf(buffer, "%f,", domain[i]);
		else
			sprintf(buffer, "%f", domain[i]);
		str += buffer;
	}
	str += "\"";

	str += " range=\"";
	for(i=0; i<2*numOut; i++) {
		if(i < 2*numOut-1)
			sprintf(buffer, "%f,", range[i]);
		else
			sprintf(buffer, "%f", range[i]);
		str += buffer;
	}
	str += "\"";

	str += " size=\"";
	for(i=0; i<numIn; i++) {
		if(i < numIn-1)
			sprintf(buffer, "%d,", size[i]);
		else
			sprintf(buffer, "%d", size[i]);
		str += buffer;
	}
	str += "\"";

	sprintf(buffer, " bits_per_sample=\"%d\"", bitsPerSample);
	str += buffer;

	if(order != 1) {
		sprintf(buffer, " order=\"%d\"", order);
		str += buffer;
	}

	if(encode != NULL) {
		str += " encode=\"";
		for(i=0; i<numIn; i++) {
			if(i < numIn-1)
				sprintf(buffer, "%f,", encode[i]);
			else
				sprintf(buffer, "%f", encode[i]);
			str += buffer;
		}
		str += "\"";
	}

	if(decode != NULL) {
		str += " decode=\"";
		for(i=0; i<numOut; i++) {
			if(i < numOut-1)
				sprintf(buffer, "%f,", decode[i]);
			else
				sprintf(buffer, "%f", decode[i]);
			str += buffer;
		}
		str += "\"";
	}

	if(filters != NULL) {
		str += " filter=\"";
		imageFilter = filters;
		while(imageFilter != NULL) {
			str += imageFilter->str;
			if(imageFilter->next != NULL)
				str += ",";
			imageFilter = imageFilter->next;
		}
		str += "\"";
	}

	if(dataLength > 0){
		if(0) {
			str += " encoding=\"Hex\">";
			// str += DataFilter::encodeXML(String(data, dataLength));
			str += DataFilter::encodeXML(DataFilter::encodeHex(String(data, dataLength)));
		}
		else {
			str += " encoding=\"Base85\">";
			// str += DataFilter::encodeXML(String(data, dataLength));
			str += DataFilter::encodeXML(DataFilter::encode85(String(data, dataLength)));
		}
	}
	else {
		str += ">";
	}

	str += "</function>\n";

	return str;
}

PObj* PdfFunction::renderBasicObjects() {
	PDF_stream *stream = new PDF_stream();
	PDF_dictionary *dict;
	PDF_blob *blob;
	PObj *filterArray, *obj;
	StringList *nextFilter;
	PDF_array *array;
	int i;
 
	if(OUTPUT_LEVEL > 2) printf("Starting PdfFunction::renderBasicObjects\n");
 
	// Add attributes to dictionary
	dict = stream->dictionary;
	dict->setValueForKey("FunctionType", new PDF_int(functionType));
 
	if(domain == NULL)
		printerr("Error - cannot render empty function domain\n");
	array = new PDF_array();
	dict->setValueForKey("Domain", array);
	for(i=0; i<2*numIn; i++) 
		array->addsub(new PDF_real(domain[i]));

	if(range == NULL)
		printerr("Error - cannot render empty function range\n");
	array = new PDF_array();
	dict->setValueForKey("Range", array);
	for(i=0; i<2*numOut; i++) 
		array->addsub(new PDF_real(range[i]));

	if(size == NULL)
		printerr("Error - cannot render empty function size\n");
	array = new PDF_array();
	dict->setValueForKey("Size", array);
	for(i=0; i<numIn; i++) 
		array->addsub(new PDF_int(size[i]));
	
	dict->setValueForKey("BitsPerSample", new PDF_int(bitsPerSample));
	if(order != 1)
		dict->setValueForKey("Order", new PDF_int(order));

	if(encode != NULL) {
		array = new PDF_array();
		dict->setValueForKey("Encode", array);
		for(i=0; i<2*numIn; i++) 
			array->addsub(new PDF_real(encode[i]));
	}

	if(decode != NULL) {
		array = new PDF_array();
		dict->setValueForKey("Decode", array);
		for(i=0; i<2*numOut; i++) 
			array->addsub(new PDF_real(decode[i]));
	}
 
	dict->setValueForKey("Length", new PDF_int(dataLength));
 
	blob = stream->dataObject;
	if(COMPRESS_DATA > 0 && filters == NULL) {
		// printf("Function Compression\n");
		String output;
		output.setval(data, dataLength);
		output = DataFilter::encode(output, "FlateDecode");
		addFilter("FlateDecode");
 
		blob->setValue((char*)(output.getrawstring()), output.getlength() );
	}
	else {
		blob->setValue(data, dataLength);
	}
 
	// Add filter list
	if(filters != NULL) {
		filterArray = new PDF_array();
		dict->setValueForKey("Filter", filterArray);
		nextFilter = filters;
		while(nextFilter != NULL) {
			filterArray->addsub(new PDF_name(nextFilter->str));
			nextFilter = nextFilter->next;
		}
	}
 
	if(OUTPUT_LEVEL > 2) printf("Ending PdfFunction::renderBasicObjects\n");
 
	return stream;
}

//------------------------------------------

PdfExtGraphicsState::PdfExtGraphicsState() : PdfResource(), sa(-1), op(-1), bg(NULL), ucr(NULL), htp(NULL) {
	type = ("ExtGState");
}

PdfExtGraphicsState::~PdfExtGraphicsState() {
	if(bg != NULL)
		delete bg;
	if(ucr != NULL)
		delete ucr;
	if(htp != NULL)
		delete [] htp;
}

PdfObject* PdfExtGraphicsState::copy() {
	PdfExtGraphicsState *beta = new PdfExtGraphicsState();

	beta->copyFields(this);
	return beta;
}

void PdfExtGraphicsState::copyFields(PdfExtGraphicsState* alpha) {
	PdfResource::copyFields(alpha);

	sa = alpha->sa;
	op = alpha->op;
	bg = (PdfFunction*) alpha->bg->copy();
	ucr = (PdfFunction*) alpha->ucr->copy();
	if(alpha->htp != NULL) {
		htp = new int[2];
		htp[0] = alpha->htp[0];
		htp[1] = alpha->htp[1];
	}
}

String PdfExtGraphicsState::render_xml(int depth) {
	String str;
	char buffer[256];

	// printf("outputing ExtGraphicsState to XML\n");
	str += "<ext_graphics_state";

	sprintf(buffer, " id=\"%d\"", refId);
	str += buffer;

	if(sa == 0)
		str += " sa=\"false\"";
	else if(sa == 1)
		str += " sa=\"true\"";

	if(op == 0)
		str += " op=\"false\"";
	else if(sa == 1)
		str += " op=\"true\"";

	if(bg != NULL) {
		bg->refId = functionNum++;
		sprintf(buffer, " gb_id=\"%d\"", bg->refId);
		str += buffer;
	}
	if(ucr != NULL) {
		ucr->refId = functionNum++;
		sprintf(buffer, " ucr_id=\"%d\"", ucr->refId);
		str += buffer;
	}

	if(htp != NULL) {
		sprintf(buffer, " htp=\"%d,%d\"", htp[0], htp[1]);
		str += buffer;
	}

	str += ">";
	str += "</ext_graphics_state>\n";
	
	if(bg != NULL) 
		str += bg->render_xml(depth);
	if(ucr != NULL) 
		str += ucr->render_xml(depth);

	return str;
}

PObj* PdfExtGraphicsState::renderBasicObjects(PDF_root *root) {
	PDF_dictionary *dict = new PDF_dictionary();

	dict->id = id;

	// printf("rendering ExtGraphicsState\n");
	dict->setValueForKey("Type", new PDF_name("ExtGState"));

	if(sa == 0)
		dict->setValueForKey("SA", new PDF_boolean(false));
	else if(sa == 1)
		dict->setValueForKey("SA", new PDF_boolean(true));

	if(op == 0)
		dict->setValueForKey("OP", new PDF_boolean(false));
	else if(op == 1)
		dict->setValueForKey("OP", new PDF_boolean(true));

	if(bg != NULL) {
		PObj* bgObj = bg->renderBasicObjects();
		bgObj->id = getNextId();
		dict->setValueForKey("BG", bgObj);
		root->addsub(bgObj);
	}
	if(ucr != NULL) {
		PObj* ucrObj = ucr->renderBasicObjects();
		ucrObj->id = getNextId();
		dict->setValueForKey("UCR", ucrObj);
		root->addsub(ucrObj);
	}

	if(htp != NULL) {
		PObj* array = new PDF_array();
		array->addsub(new PDF_int(htp[0]));
		array->addsub(new PDF_int(htp[1]));
		dict->setValueForKey("HTP", array);
	}

	return dict;
}

//------------------------------------------

PdfLinkList::~PdfLinkList() {
}

PdfLinkList::PdfLinkList(PdfLink *newLink) {
	next = NULL;
	link = newLink;
	obj = NULL;
}

PdfLinkList::PdfLinkList(PdfLink *newLink, PObj *newObj) {
	next = NULL;
	link = newLink;
	obj = newObj;
}

PdfLinkList* PdfLinkList::copy() {
	PdfLinkList *beta = new PdfLinkList(NULL);

	beta->copyFields(this);
	return beta;
}

void PdfLinkList::copyFields(PdfLinkList *alpha) {
	link = (PdfLink*)alpha->link->copy();
	obj = alpha->obj->copy();
}

//------------------------------------------

StringList::StringList() : next(NULL) {
}

StringList::StringList(String newString) : next(NULL) {
	str = newString;
}

StringList::~StringList() {
	if(next != NULL)
		delete next;
}

StringList* StringList::copy() {
	StringList *beta = new StringList();

	beta->copyFields(this);
	return beta;
}

void StringList::copyFields(StringList *alpha) {
	str = alpha->str;
}

//------------------------------------------

PdfList::PdfList() : object(NULL), next(NULL) {
}

PdfList::PdfList(PdfObject *newObject) : object(newObject), next(NULL) {
}

//------------------------------------------

