#include "PsCommands.hpp"

// High Level Commands ------------------------------------------
PObj* PsCommands::drawimage(PObj *box, String imageName) {
	PDF_pscommand *command = new PDF_pscommand("drawimage");

	command->addsub(box);
	command->contentstring = new char[imageName.getlength()+1];
	memcpy(command->contentstring, imageName.getrawstring(), imageName.getlength());
	command->contentstring[imageName.getlength()] = '\0';

	return command;
}

PObj* PsCommands::drawstring(String fontId, int fontSize, PObj *box, String text) {
	PDF_pscommand *command = new PDF_pscommand("drawstring");

	command->addsub(new PDF_name(fontId));
	command->addsub(new PDF_int(fontSize));
	command->addsub(box);

	command->contentstring = new char[text.getlength()+1];
	memcpy(command->contentstring, text.getrawstring(), text.getlength());
	command->contentstring[text.getlength()] = '\0';

	return command;
}
 
// Basic Commands -----------------------------------------------
PObj* PsCommands::closepath_eofill_and_stroke() {
	PDF_pscommand *command = new PDF_pscommand("b*");

	return command;
}

PObj* PsCommands::closepath_fill_and_stroke() {
	PDF_pscommand *command = new PDF_pscommand("b");

	return command;
}

PObj* PsCommands::begin_marked_content_with_propertylist() {
	PDF_pscommand *command = new PDF_pscommand("BDC");

	// FIX ME

	return command;
}

PObj* PsCommands::eofill_and_stroke() {
	PDF_pscommand *command = new PDF_pscommand("B*");

	return command;
}

PObj* PsCommands::fill_and_stroke() {
	PDF_pscommand *command = new PDF_pscommand("B");

	return command;
}

PObj* PsCommands::begin_image() {
	PDF_pscommand *command = new PDF_pscommand("BI");

	return command;
}

PObj* PsCommands::begin_marked_content(const String& name0) {
	PDF_pscommand *command = new PDF_pscommand("BMC");

	command->addsub(new PDF_name(name0));

	return command;
}

PObj* PsCommands::begin_text_object() {
	PDF_pscommand *command = new PDF_pscommand("BT");

	return command;
}

PObj* PsCommands::begin_hiding_undefined_operators() {
	PDF_pscommand *command = new PDF_pscommand("BX");

	return command;
}

PObj* PsCommands::appendbeziercurve(double num0, double num1, double num2, double num3, double num4, double num5) {
	PDF_pscommand *command = new PDF_pscommand("c");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));
	command->addsub(new PDF_real(num3));
	command->addsub(new PDF_real(num4));
	command->addsub(new PDF_real(num5));

	return command;
}

PObj* PsCommands::concatenate_CTM_matrix(double num0, double num1, double num2, double num3, double num4, double num5) {
	PDF_pscommand *command = new PDF_pscommand("cm");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));
	command->addsub(new PDF_real(num3));
	command->addsub(new PDF_real(num4));
	command->addsub(new PDF_real(num5));

	return command;
}

PObj* PsCommands::set_fillcolorspace(const String& name0) {
	PDF_pscommand *command = new PDF_pscommand("cs");

	command->addsub(new PDF_name(name0));

	return command;
}

PObj* PsCommands::set_strokecolorspace(const String& name0) {
	PDF_pscommand *command = new PDF_pscommand("CS");

	command->addsub(new PDF_name(name0));

	return command;
}

PObj* PsCommands::set_character_width(double num0, double num1) {
	PDF_pscommand *command = new PDF_pscommand("D0");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));

	return command;
}

PObj* PsCommands::set_cache_device(double num0, double num1, double num2, double num3, double num4, double num5) {
	PDF_pscommand *command = new PDF_pscommand("D1");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));
	command->addsub(new PDF_real(num3));
	command->addsub(new PDF_real(num4));
	command->addsub(new PDF_real(num5));

	return command;
}

PObj* PsCommands::execute_xobject(const String& xobject0) {
	PDF_pscommand *command = new PDF_pscommand("Do");

	command->addsub(new PDF_name(xobject0));

	return command;
}

PObj* PsCommands::mark_point_with_propertylist() {
	PDF_pscommand *command = new PDF_pscommand("DP");

	// FIX ME

	return command;
}

PObj* PsCommands::set_dashpattern(PDF_array *array0, double num1) {
	PDF_pscommand *command = new PDF_pscommand("d");

	command->addsub(array0);
	command->addsub(new PDF_real(num1));

	return command;
}

PObj* PsCommands::end_image() {
	PDF_pscommand *command = new PDF_pscommand("EI");

	return command;
}

PObj* PsCommands::end_marked_content() {
	PDF_pscommand *command = new PDF_pscommand("EMC");

	return command;
}

PObj* PsCommands::end_text_object() {
	PDF_pscommand *command = new PDF_pscommand("ET");

	return command;
}

PObj* PsCommands::end_hiding_undefined_operators() {
	PDF_pscommand *command = new PDF_pscommand("EX");

	return command;
}

PObj* PsCommands::fillpath() {
	PDF_pscommand *command = new PDF_pscommand("f");

	return command;
}

PObj* PsCommands::fillpath_oddeven() {
	PDF_pscommand *command = new PDF_pscommand("f*");

	return command;
}

PObj* PsCommands::set_grayfilltint(double num0) {
	PDF_pscommand *command = new PDF_pscommand("g");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::set_graystroketint(double num0) {
	PDF_pscommand *command = new PDF_pscommand("G");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::set_graphicsstate(const String& name0) {
	PDF_pscommand *command = new PDF_pscommand("gs");

	command->addsub(new PDF_name(name0));

	return command;
}

PObj* PsCommands::closecurrentsubpath() {
	PDF_pscommand *command = new PDF_pscommand("h");

	return command;
}

PObj* PsCommands::begin_image_data() {
	PDF_pscommand *command = new PDF_pscommand("ID");

	return command;
}

PObj* PsCommands::set_flatness(double num0) {
	PDF_pscommand *command = new PDF_pscommand("i");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::set_linecap(double num0) {
	PDF_pscommand *command = new PDF_pscommand("J");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::set_linejoin(double num0) {
	PDF_pscommand *command = new PDF_pscommand("j");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::set_fillcolorCMYK(double num0, double num1, double num2, double num3) {
	PDF_pscommand *command = new PDF_pscommand("k");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));
	command->addsub(new PDF_real(num3));

	return command;
}

PObj* PsCommands::set_strokecolorCMYK(double num0, double num1, double num2, double num3) {
	PDF_pscommand *command = new PDF_pscommand("K");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));
	command->addsub(new PDF_real(num3));

	return command;
}

PObj* PsCommands::lineto(double num0, double num1) {
	PDF_pscommand *command = new PDF_pscommand("l");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));

	return command;
}

PObj* PsCommands::moveto(double num0, double num1) {
	PDF_pscommand *command = new PDF_pscommand("m");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));

	return command;
}

PObj* PsCommands::mark_point(const String& name0) {
	PDF_pscommand *command = new PDF_pscommand("MP");

	command->addsub(new PDF_name(name0));

	return command;
}

PObj* PsCommands::set_miterlimit(double num0) {
	PDF_pscommand *command = new PDF_pscommand("M");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::endpath() {
	PDF_pscommand *command = new PDF_pscommand("n");

	return command;
}

PObj* PsCommands::newline_and_setspacing_and_show_text_string(double num0, double num1, const String& string2) {
	PDF_pscommand *command = new PDF_pscommand("\"");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_string(string2));

	return command;
}

PObj* PsCommands::newline_and_show_text_string(const String& string0) {
	PDF_pscommand *command = new PDF_pscommand("'");

	command->addsub(new PDF_string(string0));

	return command;
}

PObj* PsCommands::exec_postscript(const String& string0) {
	PDF_pscommand *command = new PDF_pscommand("PS");

	command->addsub(new PDF_string(string0));

	return command;
}

PObj* PsCommands::restore_graphics_state() {
	PDF_pscommand *command = new PDF_pscommand("Q");

	return command;
}

PObj* PsCommands::save_graphics_state() {
	PDF_pscommand *command = new PDF_pscommand("q");

	return command;
}

PObj* PsCommands::add_rectangle(double num0, double num1, double num2, double num3) {
	PDF_pscommand *command = new PDF_pscommand("re");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));
	command->addsub(new PDF_real(num3));

	return command;
}

PObj* PsCommands::set_fillcolorRGB(double num0, double num1, double num2) {
	PDF_pscommand *command = new PDF_pscommand("rg");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));

	return command;
}

PObj* PsCommands::set_strokecolorRGB(double num0, double num1, double num2) {
	PDF_pscommand *command = new PDF_pscommand("RG");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));

	return command;
}

PObj* PsCommands::set_colorrenderingintent(const String& name0) {
	PDF_pscommand *command = new PDF_pscommand("ri");

	command->addsub(new PDF_name(name0));

	return command;
}

PObj* PsCommands::close_and_strokepath() {
	PDF_pscommand *command = new PDF_pscommand("s");

	return command;
}

PObj* PsCommands::SCN() {
	PDF_pscommand *command = new PDF_pscommand("SCN");

	// FIX ME

	return command;
}

PObj* PsCommands::scn() {
	PDF_pscommand *command = new PDF_pscommand("scn");

	// FIX ME

	return command;
}

PObj* PsCommands::set_fillcolor() {
	PDF_pscommand *command = new PDF_pscommand("sc");

	// FIX ME

	return command;
}

PObj* PsCommands::set_strokecolor() {
	PDF_pscommand *command = new PDF_pscommand("SC");

	// FIX ME

	return command;
}

PObj* PsCommands::shade(double num0) {
	PDF_pscommand *command = new PDF_pscommand("sh");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::strokepath() {
	PDF_pscommand *command = new PDF_pscommand("S");

	return command;
}

PObj* PsCommands::set_character_spacing(double num0) {
	PDF_pscommand *command = new PDF_pscommand("Tc");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::newline(double num0, double num1) {
	PDF_pscommand *command = new PDF_pscommand("Td");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));

	return command;
}

PObj* PsCommands::newline_and_set_leading(double num0, double num1) {
	PDF_pscommand *command = new PDF_pscommand("TD");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));

	return command;
}

PObj* PsCommands::set_font_and_size(const String& name0, double num1) {
	PDF_pscommand *command = new PDF_pscommand("Tf");

	command->addsub(new PDF_name(name0));
	command->addsub(new PDF_real(num1));

	return command;
}

PObj* PsCommands::show_text_string(const String& string0) {
	PDF_pscommand *command = new PDF_pscommand("Tj");

	command->addsub(new PDF_string(string0));

	return command;
}

PObj* PsCommands::show_text_string_with_displacements(PDF_array *array0) {
	PDF_pscommand *command = new PDF_pscommand("TJ");

	command->addsub(array0);

	return command;
}

PObj* PsCommands::set_text_leading(double num0) {
	PDF_pscommand *command = new PDF_pscommand("TL");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::set_text_matrix(double num0, double num1, double num2, double num3, double num4, double num5) {
	PDF_pscommand *command = new PDF_pscommand("Tm");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));
	command->addsub(new PDF_real(num3));
	command->addsub(new PDF_real(num4));
	command->addsub(new PDF_real(num5));

	return command;
}

PObj* PsCommands::newline_with_defaultoffset() {
	PDF_pscommand *command = new PDF_pscommand("T*");

	return command;
}

PObj* PsCommands::set_text_rendering_mode(double num0) {
	PDF_pscommand *command = new PDF_pscommand("Tr");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::set_text_size(double num0) {
	PDF_pscommand *command = new PDF_pscommand("Ts");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::set_word_spacing(double num0) {
	PDF_pscommand *command = new PDF_pscommand("Tw");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::set_horizontal_scaling(double num0) {
	PDF_pscommand *command = new PDF_pscommand("Tz");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::appendbeziercurve_current(double num0, double num1, double num2, double num3) {
	PDF_pscommand *command = new PDF_pscommand("v");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));
	command->addsub(new PDF_real(num3));

	return command;
}

PObj* PsCommands::clip() {
	PDF_pscommand *command = new PDF_pscommand("W");

	return command;
}

PObj* PsCommands::clip_evenodd() {
	PDF_pscommand *command = new PDF_pscommand("W*");

	return command;
}

PObj* PsCommands::set_linewidth(double num0) {
	PDF_pscommand *command = new PDF_pscommand("w");

	command->addsub(new PDF_real(num0));

	return command;
}

PObj* PsCommands::appendbeziercurve_current_secondversion(double num0, double num1, double num2, double num3) {
	PDF_pscommand *command = new PDF_pscommand("y");

	command->addsub(new PDF_real(num0));
	command->addsub(new PDF_real(num1));
	command->addsub(new PDF_real(num2));
	command->addsub(new PDF_real(num3));

	return command;
}

