#include "XMLHandler.hpp"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>

XMLHandler::XMLHandler(XMLModule *newModule) {
	stack = new XMLElement*[MAX_STACK_SIZE];
	level = -1;
	DEBUG_OUTPUT = false;

	module = newModule;
}

void XMLHandler::startElement(const xmlChar *name, const xmlChar **attributes) {
	int i;
	char *type;

    if(level >= MAX_STACK_SIZE-1)
		return;

	level++;

	if(DEBUG_OUTPUT) indent();

    stack[level] = new XMLElement(name, attributes);                                          

	type = stack[level]->name;

	if(DEBUG_OUTPUT) printf("Element: %s\n", stack[level]->name);
	for(i=0; i<stack[level]->attrLen; i++) {
        if(DEBUG_OUTPUT) indent();
        if(DEBUG_OUTPUT) printf(" **Attribute: %s = %s\n", stack[level]->attrName[i], stack[level]->attrValue[i]);
    } 

	if(DEBUG_OUTPUT) indent();
	if(DEBUG_OUTPUT) printf("Processing %s\n", type);

	//---------------------
	module->startElement(stack, level);

}

void XMLHandler::characters( const xmlChar *ch, int len ) {
	stack[level]->appendToCData(ch, len);
}

void XMLHandler::endElement (const xmlChar *name) {

	if(level < 0)
		return;

	if(stack[level]->cdata.getlength() > 0) {
		if(DEBUG_OUTPUT) indent();
		if(DEBUG_OUTPUT) printf("Data: %s\n", stack[level]->cdata.nullterm());
	}

	//---------------------
	module->endElement(stack, level);

	if(DEBUG_OUTPUT) indent();
	if(DEBUG_OUTPUT) printf("end %s\n", stack[level]->name);

    delete stack[level];
	level--;
}

/* FIX - Xerces error handling code
void XMLHandler::error(const SAXParseException& exception) {
	const XMLCh *xMsg = exception.getMessage();
	int i;
	char buf[1024]="\0";
	char *ptr;

	strcat(buf, "Error: ");
	ptr = buf + pdfc_strlen(buf);
	for(i=0; xMsg[i] != 0; i++)
		*ptr++ = xMsg[i] & 0xff;
	sprintf(ptr, " at line number %d\n", exception.getLineNumber() );
	printerr(buf);
}

void XMLHandler::fatalError(const SAXParseException& exception) {
	const XMLCh *xMsg = exception.getMessage();
	int i;
	char buf[1024]="\0";
	char *ptr;

	strcat(buf, "Fatal Error: ");
	ptr = buf + pdfc_strlen(buf);
	for(i=0; xMsg[i] != 0; i++)
		*ptr++ = xMsg[i] & 0xff;
	sprintf(ptr, " at line number %d column %d\n", exception.getLineNumber(), exception.getColumnNumber() );
	printerr(buf);
}

void XMLHandler::warning(const SAXParseException& exception) {
	const XMLCh *xMsg = exception.getMessage();
	int i;
	char buf[1024]="\0";
	char *ptr;

	strcat(buf, "Warning: ");
	ptr = buf + pdfc_strlen(buf);
	for(i=0; xMsg[i] != 0; i++)
		*ptr++ = xMsg[i] & 0xff;
	sprintf(ptr, " at line number %d\n", exception.getLineNumber() );
	printerr(buf);
}
*/

void XMLHandler::indent() {
	int i;

	for(i=0; i<level; i++) {
		putchar(' ');
	}
}

