#include "XMLModule_highLevel.hpp"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <sys/stat.h>

#include "pdfc_base.h"
#include "pdfc_basicobjects.h"
#include "FontReader.hpp"
#include "ImageReader.hpp"
#include "stringclass.h"

XMLModule_highLevel::XMLModule_highLevel() : root(NULL), currentObj(NULL), inContentSection(false) {
	DEBUG_OUTPUT = false;
	contentModule.enablePostscriptParsing();
}

XMLModule_highLevel::~XMLModule_highLevel() {} // Intentionally NULL (Steve says so).

PdfRoot* XMLModule_highLevel::getRoot() {
	return root;
}

//================================================

void XMLModule_highLevel::startElement(XMLElement** stack, int level) {
	PdfObject *pdfObject = NULL;
	char *type, *value;
	char *filename = NULL; 
	int i;

	// printf("Creating XML Element\n");
	type = stack[level]->name;
	if(DEBUG_OUTPUT) 
		printf("XML Tag: %s\n", type);

	if(inContentSection) {
		contentModule.startElement(stack, level);
		return;
	}

	if(DEBUG_OUTPUT) {
		if(currentObj != NULL) printf("start pdfObject: %s, Current: %s\n", type, currentObj->objTypeStr());
		else printf("start pdfObject: %s, Current: %s\n", type, "NULL");
	}

	if(strcmp(type, "pdf") == 0) {
		root = new PdfRoot();
		pdfObject = root;

		value = stack[level]->getAttribute("author");
		if(value != NULL) 
			root->author = value;

		value = stack[level]->getAttribute("creationDate");
		if(value != NULL) 
			root->creationDate = value;

		value = stack[level]->getAttribute("creator");
		if(value != NULL) 
			root->creator = value;

		value = stack[level]->getAttribute("title");
		if(value != NULL) 
			root->title = value;

		value = stack[level]->getAttribute("subject");
		if(value != NULL) 
			root->subject = value;

		value = stack[level]->getAttribute("keywords");
		if(value != NULL) 
			root->keywords = value;

		value = stack[level]->getAttribute("mode");
		if(value != NULL) {
			if(strcmp(value, "show_bookmarks") == 0)
				root->showBookmarks = true;
			else
				printerr("Error - unknown document mode\n");
		}

		if(currentObj != NULL)
			printerr("HL Error - pdf tags may not be contained within other tags\n");
	}
	else if(strcmp(type, "page") == 0) {
		pdfObject = new PdfPage();

		if(currentObj->objType() != p_root) 
			printerr("HL Error - pages must be contained within pdf tags\n");
		if(root == NULL)
			printerr("Error - trying to add page without root object\n");

		value = stack[level]->getAttribute("id");
		if(value != NULL) 
			((PdfPage*)pdfObject)->id = atoi(value);

		root->addPage( (PdfPage*) pdfObject );
	}
	else if(strcmp(type, "contents") == 0) {
		inContentSection = true;
		if(DEBUG_OUTPUT) printf("Parsing Contents with low level XMLModule\n");
	}
	else if(strcmp(type, "page_resource") == 0) {
		pdfObject = new PdfResourceRef();
		PdfResource *resource;

		value = stack[level]->getAttribute("id");
		if(value == NULL) 
			printerr("HL Error - page resources must have an id\n");

		resource = (PdfResource*) resourceHash.get(value);
		if(resource == NULL)
			printerr("HL Error - could not find resource specified by page_resource - %s\n", value);

		((PdfResourceRef*)pdfObject)->resource = resource;
		((PdfResourceRef*)pdfObject)->type = resource->type;

		value = stack[level]->getAttribute("name");
		if(value == NULL) 
			printerr("HL Error - page resources must have an name\n");

		((PdfResourceRef*)pdfObject)->name = (value);

		if(currentObj->objType() != p_page) 
			printerr("HL Error - page_resources must be contained within pages\n");
		((PdfPage*)currentObj)->addResourceRef( (PdfResourceRef*) pdfObject );
	}
	else if(strcmp(type, "image") == 0) {
		pdfObject = new PdfImage();

		value = stack[level]->getAttribute("id");
		if(value == NULL) 
			printerr("HL Error - images must have an id\n");
		((PdfImage*)pdfObject)->id = atoi(value);

		if(currentObj->objType() == p_root) {
			resourceHash.add(value, pdfObject);
		}
		else if(currentObj->objType() == p_page) {
			PdfResourceRef *ref = new PdfResourceRef();
			ref->resource = (PdfResource*)pdfObject;
			ref->type = ((PdfResource*)pdfObject)->type;
			ref->name = (value);

			((PdfPage*)currentObj)->addResourceRef( ref );
		}
		else {
			printerr("HL Error - images must be contained within pdf tags or page tags\n");
		}

		((PdfRoot*)root)->addResource( (PdfResource*) pdfObject );

		value = stack[level]->getAttribute("format");
		if(value == NULL) 
			((PdfImage*)pdfObject)->format = "external";
		else
			((PdfImage*)pdfObject)->format = value;

		if(((PdfImage*)pdfObject)->format == "native") {
			value = stack[level]->getAttribute("width");
			if(value == NULL) 
				printerr("HL Error - images must have a width\n");
			((PdfImage*)pdfObject)->width = atoi(value);

			value = stack[level]->getAttribute("height");
			if(value == NULL) 
				printerr("HL Error - images must have a height\n");
			((PdfImage*)pdfObject)->height = atoi(value);

			value = stack[level]->getAttribute("bits_per_pixel");
			if(value == NULL) 
				printerr("HL Error - images must specify a number of bits per pixel\n");
			((PdfImage*)pdfObject)->bitsPerPixel = atoi(value);

			value = stack[level]->getAttribute("colorspace");
			if(value != NULL) {
				((PdfImage*)pdfObject)->colorSpace = new PdfColorSpace();
				((PdfImage*)pdfObject)->colorSpace->catagory = value;
			}
			value = stack[level]->getAttribute("colorspace_id");
			if(value != NULL) {
				// printf("ColorSpace ID: %s\n", value);
				if( ((PdfImage*)pdfObject)->colorSpace == NULL)
					((PdfImage*)pdfObject)->colorSpaceId = atoi(value);
				else
					printerr("Error - images may not specify both a colorspace and colorspace_id\n");
			}
		}
		else if(((PdfImage*)pdfObject)->format == "external") {
		}
		else {
			printerr("Error - format of image is unkown - %s\n", ((PdfImage*)pdfObject)->format.nullterm());
		}

		value = stack[level]->getAttribute("filter");
		if(value != NULL) {
			char *ptr; 

			while( (ptr=strchr(value, ',')) != NULL ) {
				*ptr = 0;
				((PdfImage*)pdfObject)->addFilter(value);
				value = ptr+1;
			}
			((PdfImage*)pdfObject)->addFilter(value);
		}
	
		filename = stack[level]->getAttribute("filename");
		if(filename != NULL) {
			((PdfResource*)pdfObject)->filename = filename;
			if(DEBUG_OUTPUT) printf("Filename: %s\n", filename);
		}
	}
	else if(strcmp(type, "colorspace") == 0) {
		if(currentObj->objType() != p_root) 
			printerr("HL Error - colorspaces must be contained within the document root\n");

		// printf("Found ColorSpace entry\n");
		PdfColorSpace *colorSpace = new PdfColorSpace();
		pdfObject = colorSpace;

		value = stack[level]->getAttribute("is_page_resource");
		if(value != NULL) {
			if(strcmp(value, "true") == 0)
				colorSpace->isPageResource = true;
			else if(strcmp(value, "false") == 0)
				colorSpace->isPageResource = false;
			else
				printerr("Error - invalid value for colorspace page_resource attribute (must be true or false)\n");
		}

		value = stack[level]->getAttribute("id");
		if(value == NULL) 
			printerr("HL Error - colorspaces must specify an id\n");
		colorSpace->id = atoi(value);
		// printf("colorspace id: %s\n", value);

		if(colorSpace->isPageResource) {
			if(currentObj->objType() == p_root) {
				resourceHash.add(value, pdfObject);
				((PdfRoot*)root)->addResource( (PdfResource*) pdfObject );
			}
			else if(currentObj->objType() == p_page) {
				PdfResourceRef *ref = new PdfResourceRef();
				ref->resource = (PdfResource*)pdfObject;
				ref->type = ((PdfResource*)pdfObject)->type;
				ref->name = (value);

				((PdfPage*)currentObj)->addResourceRef( ref );

				((PdfRoot*)root)->addResource( (PdfResource*) pdfObject );
			}
			else {
				printerr("Error - page resource colorspaces must be contained in the root or page objects, not %s\n", currentObj->objTypeStr());
			}
		}

		root->addColorSpace(colorSpace);

		value = stack[level]->getAttribute("catagory");
		if(value == NULL) 
			printerr("HL Error - colorspaces must specify a catagory\n");
		colorSpace->catagory = value;
		// printf("colorspace catagory: %s\n", value);

		if(colorSpace->catagory == "CalGray" || colorSpace->catagory == "CalRGB" || colorSpace->catagory == "Lab") {
			// printf("Device independent\n");

			value = stack[level]->getAttribute("white_point");
			if(value == NULL) 
				printerr("HL Error - device independent colorspaces must specify a white_point\n");
			splitNumsIntoArray(colorSpace->whitePoint, 3, value);

			value = stack[level]->getAttribute("black_point");
			if(value != NULL) 
				splitNumsIntoArray(colorSpace->blackPoint, 3, value);

			if(colorSpace->catagory == "CalRGB") {
				value = stack[level]->getAttribute("gamma");
				if(value != NULL) 
					splitNumsIntoArray(colorSpace->gamma, 3, value);

				value = stack[level]->getAttribute("matrix");
				if(value != NULL) 
					splitNumsIntoArray(colorSpace->matrix, 9, value);
			}
			else if(colorSpace->catagory == "Lab") {
				value = stack[level]->getAttribute("range");
				if(value != NULL) 
					splitNumsIntoArray(colorSpace->range, 4, value);
			}
		}
		else if(colorSpace->catagory == "Indexed") {
			// printf("Indexed\n");

			value = stack[level]->getAttribute("base");
			if(value != NULL) {
				colorSpace->base = new PdfColorSpace(value);
			}
			else {
				value = stack[level]->getAttribute("base_id");
				if(value == NULL) 
					printerr("HL Error - indexed colorspaces must specify a base\n");
				colorSpace->baseId = atoi(value);
			}

			value = stack[level]->getAttribute("hival");
			if(value == NULL) 
				printerr("HL Error - indexed colorspaces must specify a hival\n");
			colorSpace->hival = atoi(value);

			value = stack[level]->getAttribute("filter");
			if(value != NULL) {
				char *ptr; 

				while( (ptr=strchr(value, ',')) != NULL ) {
					*ptr = 0;
					colorSpace->addFilter(value);
					value = ptr+1;
				}
				colorSpace->addFilter(value);
			}
		}
		else if(colorSpace->catagory == "Separation") {
			// printf("Separation\n");

			value = stack[level]->getAttribute("separation_name");
			if(value == NULL) 
				printerr("HL Error - indexed colorspaces must specify a separation_name\n");
			colorSpace->separationName = value;

			value = stack[level]->getAttribute("alt_space");
			if(value != NULL) {
				colorSpace->altSpace = new PdfColorSpace(value);
			}
			else {
				value = stack[level]->getAttribute("alt_space_id");
				if(value == NULL) 
					printerr("HL Error - indexed colorspaces must specify an alt_space\n");
				colorSpace->altSpaceId = atoi(value);
			}

			value = stack[level]->getAttribute("tint_transform_id");
			if(value == NULL) 
				printerr("HL Error - indexed colorspaces must specify a tint_transform_id\n");
			colorSpace->tintTransformId = atoi(value);

			value = stack[level]->getAttribute("filter");
			if(value != NULL) {
				char *ptr; 

				while( (ptr=strchr(value, ',')) != NULL ) {
					*ptr = 0;
					colorSpace->addFilter(value);
					value = ptr+1;
				}
				colorSpace->addFilter(value);
			}
		}
		else if(colorSpace->catagory == "DeviceGray" || colorSpace->catagory == "DeviceRGB" || colorSpace->catagory == "DeviceCMYK") {
			// printf("Device dependent\n");
			// Do nothing
		}
		else {
			printerr("Error unsupported colorspace catagory - %s\n", colorSpace->catagory.nullterm());
		}
	}
	else if(strcmp(type, "function") == 0) {
		PdfFunction *function = new PdfFunction();
		pdfObject = function;

		// Get id
		value = stack[level]->getAttribute("id");
		if(value == NULL) 
			printerr("HL Error - functions must have an id\n");
		function->id = atoi(value);

		// Get domain
		value = stack[level]->getAttribute("domain");
		if(value == NULL) 
			printerr("HL Error - functions must have a domain\n");

		function->numIn = 1;
		char *ptr = value; 
		while( (ptr=strchr(ptr, ',')) != NULL ) {
			function->numIn++;
			ptr++;
		}
		if(function->numIn %2)
			printerr("Error - domain has mismatched set of pairs\n");
		i = 0;
		function->domain = new double[function->numIn];
		function->numIn /= 2;
		while( (ptr=strchr(value, ',')) != NULL ) {
			*ptr = 0;
			function->domain[i++] = atof(value);
			value = ptr+1;
		}
		function->domain[i++] = atof(value);

		// Get range
		value = stack[level]->getAttribute("range");
		if(value == NULL) 
			printerr("HL Error - functions must have a range\n");

		function->numOut = 1;
		ptr = value; 
		while( (ptr=strchr(ptr, ',')) != NULL ) {
			function->numOut++;
			ptr++;
		}
		i = 0;
		if(function->numOut %2)
			printerr("Error - range has mismatched set of pairs\n");
		function->range = new double[function->numOut];
		function->numOut /= 2;
		while( (ptr=strchr(value, ',')) != NULL ) {
			*ptr = 0;
			function->range[i++] = atof(value);
			value = ptr+1;
		}
		function->range[i++] = atof(value);

		// Get size
		value = stack[level]->getAttribute("size");
		if(value == NULL) 
			printerr("HL Error - functions must have a size\n");

		i = 1;
		ptr = value; 
		while( (ptr=strchr(ptr, ',')) != NULL ) {
			i++;
			ptr++;
		}
		if(i != function->numIn)
			printerr("Error - function size array does not match up with domain array\n");
		i = 0;
		function->size = new int[function->numIn];
		while( (ptr=strchr(value, ',')) != NULL ) {
			*ptr = 0;
			function->size[i++] = atoi(value);
			value = ptr+1;
		}
		function->size[i++] = atoi(value);

		// Get bitsPerSample
		value = stack[level]->getAttribute("bits_per_sample");
		if(value == NULL) 
			printerr("HL Error - functions must have an bits_per_sample\n");
		function->bitsPerSample = atoi(value);

		// Get order
		value = stack[level]->getAttribute("order");
		if(value != NULL) 
			function->order = atoi(value);

		// Get encode
		value = stack[level]->getAttribute("encode");
		if(value != NULL) {
			i = 1;
			ptr = value; 
			while( (ptr=strchr(ptr, ',')) != NULL ) {
				i++;
				ptr++;
			}
			if(i != 2*function->numIn)
				printerr("Error - function encode array does not match up with domain array\n");
			i = 0;
			function->encode = new double[function->numIn];
			while( (ptr=strchr(value, ',')) != NULL ) {
				*ptr = 0;
				function->encode[i++] = atof(value);
				value = ptr+1;
			}
			function->encode[i++] = atof(value);
		}

		// Get decode
		value = stack[level]->getAttribute("decode");
		if(value != NULL) {
			i = 1;
			ptr = value; 
			while( (ptr=strchr(ptr, ',')) != NULL ) {
				i++;
				ptr++;
			}
			if(i != 2*function->numOut)
				printerr("Error - function decode array does not match up with domain array\n");
			i = 0;
			function->decode = new double[function->numOut];
			while( (ptr=strchr(value, ',')) != NULL ) {
				*ptr = 0;
				function->decode[i++] = atof(value);
				value = ptr+1;
			}
			function->decode[i++] = atof(value);
		}

		value = stack[level]->getAttribute("filter");
		if(value != NULL) {
			char *ptr; 

			while( (ptr=strchr(value, ',')) != NULL ) {
				*ptr = 0;
				function->addFilter(value);
				value = ptr+1;
			}
			function->addFilter(value);
		}

		root->addFunction( function );
	}
	else if(strcmp(type, "ext_graphics_state") == 0) {
		PdfExtGraphicsState *eGState = new PdfExtGraphicsState();
		pdfObject = eGState;

		// printf("Found XML ExtGraphicsState\n");

		// Get id
		value = stack[level]->getAttribute("id");
		if(value == NULL) 
			printerr("HL Error - functions must have an id\n");
		eGState->id = atoi(value);

		if(currentObj->objType() == p_root) {
			resourceHash.add(value, pdfObject);
		}
		else if(currentObj->objType() == p_page) {
			PdfResourceRef *ref = new PdfResourceRef();
			ref->resource = (PdfResource*)pdfObject;
			ref->type = ((PdfResource*)pdfObject)->type;
			ref->name = (value);

			((PdfPage*)currentObj)->addResourceRef( ref );
		}
		else {
			printerr("HL Error - ExtGStates must be contained within pdf tags or page tags\n");
		}

		((PdfRoot*)root)->addResource( (PdfResource*) pdfObject );

		// Get sa
		value = stack[level]->getAttribute("sa");
		if(value != NULL) {
			if(strcmp(value, "true") == 0)
				eGState->sa = 1;
			else if(strcmp(value, "false") == 0)
				eGState->sa = 0;
			else
				printerr("Error - invalid value for ExtGState sa attribute (must be true or false)\n");
		}

		// Get op
		value = stack[level]->getAttribute("op");
		if(value != NULL) {
			if(strcmp(value, "true") == 0)
				eGState->op = 1;
			else if(strcmp(value, "false") == 0)
				eGState->op = 0;
			else
				printerr("Error - invalid value for ExtGState op attribute (must be true or false)\n");
		}

/*
		// Get bg id
		value = stack[level]->getAttribute("bg_id");
		if(value != NULL) 
			eGState->bg_id = atoi(value);

		// Get ucr id
		value = stack[level]->getAttribute("ucr_id");
		if(value != NULL) 
			eGState->bg_id = atoi(value);

		// Get htp
*/
	}
	else if(strcmp(type, "font") == 0) {
		char *subType = NULL; 
		pdfObject = new PdfFont();

		value = stack[level]->getAttribute("id");
		if(value == NULL) 
			printerr("HL Error - fonts must have an id\n");
		((PdfFont*)pdfObject)->id = atoi(value);

		if(currentObj->objType() == p_root) {
			resourceHash.add(value, pdfObject);
		}
		else if(currentObj->objType() == p_page) {
			PdfResourceRef *ref = new PdfResourceRef();
			ref->resource = (PdfResource*)pdfObject;
			ref->type = ((PdfResource*)pdfObject)->type;
			ref->name = (value);

			((PdfPage*)currentObj)->addResourceRef( ref );
		}
		else {
			printerr("HL Error - fonts must be contained within pdf tags or page tags\n");
		}

		((PdfRoot*)root)->addResource( (PdfResource*) pdfObject );

		subType = stack[level]->getAttribute("type");
		if(subType == NULL) 
			printerr("HL Error - fonts must specify a type\n");
		((PdfFont*)pdfObject)->subType = subType;

		filename = stack[level]->getAttribute("filename");
		if(filename != NULL) {
			((PdfResource*)pdfObject)->filename = filename;
			if(DEBUG_OUTPUT) printf("Filename: %s\n", filename);
		}

		value = stack[level]->getAttribute("data_subtype");
		if(value != NULL) {
			((PdfFont*)pdfObject)->dataSubType = value;
			if(DEBUG_OUTPUT) printf("DataSubType: %s\n", value);
		}

	}
	else if(strcmp(type, "font_metrics") == 0) {
		if(currentObj->objType() != p_font) 
			printerr("HL Error - font_metrics must be contained within a font\n");

		if(((PdfFont*)currentObj)->subType != "TrueType" || ((PdfFont*)currentObj)->filename.getlength() < 1) {
			value = stack[level]->getAttribute("first_char");
			if(value == NULL) 
				printerr("HL Error - fonts must have an first_char\n");
			((PdfFont*)currentObj)->firstChar = atoi(value);

			value = stack[level]->getAttribute("last_char");
			if(value == NULL) 
				printerr("HL Error - fonts must have an last_char\n");
			((PdfFont*)currentObj)->lastChar = atoi(value);

			if((((PdfFont*)currentObj)->lastChar - ((PdfFont*)currentObj)->firstChar) > -1) {
				((PdfFont*)currentObj)->widths = new long[((PdfFont*)currentObj)->lastChar - ((PdfFont*)currentObj)->firstChar+1];
			}
			value = stack[level]->getAttribute("ascent");
			if(value == NULL) 
				printerr("HL Error - fonts must have an ascent\n");
			((PdfFont*)currentObj)->ascent = atoi(value);

			value = stack[level]->getAttribute("descent");
			if(value == NULL) 
				printerr("HL Error - fonts must have a descent\n");
			((PdfFont*)currentObj)->descent = atoi(value);

			value = stack[level]->getAttribute("flags");
			if(value == NULL) 
				printerr("HL Error - fonts must have a flags\n");
			((PdfFont*)currentObj)->flags = atoi(value);
		}
	}
	else if(strcmp(type, "font_widths") == 0) {
		if(level < 1 || strcmp(stack[level-1]->name, "font_metrics") != 0)  
			printerr("HL Error - font_widths tags must be contained within font_metrics tags\n"); 

		if(currentObj->objType() != p_font) 
			printerr("HL Error - a font_widths must be contained within a font\n");

		value = stack[level]->getAttribute("encoding");
		if(value != NULL) 
			((PdfFont*)currentObj)->encoding = value;
	}
	else if(strcmp(type, "width") == 0) {
		if(level < 1 || strcmp(stack[level-1]->name, "font_widths") != 0)  
			printerr("HL Error - width tags must be contained within font_widths tags\n"); 
	}
	else if(strcmp(type, "font_bounding_box") == 0) {
		if(currentObj->objType() != p_font) 
			printerr("HL Error - font_bounding_box must be embedded within a font\n");

		if(((PdfFont*)currentObj)->subType != "TrueType" || ((PdfFont*)currentObj)->filename.getlength() < 1) {
			if(currentObj->objType() != p_font) 
				printerr("HL Error - a font_bounding_box must be contained within a font\n");

			double *bBox = ((PdfFont*)currentObj)->fontBBox;

			value = stack[level]->getAttribute("llx");
			if(value == NULL) 
				printerr("HL Error - font_bounding_box must have a llx coordinate\n");
			bBox[0] = atof(value);

			value = stack[level]->getAttribute("lly");
			if(value == NULL) 
				printerr("HL Error - font_bounding_box must have a lly coordinate\n");
			bBox[1] = atof(value);

			value = stack[level]->getAttribute("urx");
			if(value == NULL) 
				printerr("HL Error - font_bounding_box must have a urx coordinate\n");
			bBox[2] = atof(value);

			value = stack[level]->getAttribute("ury");
			if(value == NULL) 
				printerr("HL Error - font_bounding_box must have a ury coordinate\n");
			bBox[3] = atof(value);
		}
	}
	else if(strcmp(type, "font_data") == 0) {
	}
	else if(strcmp(type, "bookmark") == 0) {
		PdfBookmark *bookmark = new PdfBookmark();

		value = stack[level]->getAttribute("title");
		if(value == NULL) 
			printerr("HL Error - a bookmark must have a title\n");
		bookmark->title = value;

		value = stack[level]->getAttribute("status");
		if(value != NULL) {
			if(strcmp(value, "open") == 0) {
				bookmark->isOpen = true;
			}
			else if(strcmp(value, "closed") == 0) {
			}
			else {
				printerr("Error - unknown value for bookmark status\n");
			}
		}

		value = stack[level]->getAttribute("page");
		if(value == NULL) 
			printerr("Error - bookmarks must specify a page\n");
		bookmark->destKey = value;

		if(currentObj->objType() == p_root) {
			((PdfRoot*)currentObj)->addBookmark(bookmark);
		}
		else if(currentObj->objType() == p_bookmark) {
			((PdfBookmark*)currentObj)->addBookmark(bookmark);
		}
		else {
			printerr("Bookmarks must be contained within the main pdf tag or nested within other bookmarks\n");
		}

		pdfObject = bookmark;
	}
	else if(strcmp(type, "annotation") == 0 || 
	        strcmp(type, "text_annotation") == 0 ||
	        strcmp(type, "link") == 0) {
		if(currentObj->objType() != p_page) 
			printerr("HL Error - a annotation must be contained within a page\n");

		if(strcmp(type, "annotation") == 0) {
			pdfObject = new PdfAnnot();

			value = stack[level]->getAttribute("type");
			if(value == NULL) 
				printerr("HL Error - annotations must have a type\n");
			((PdfAnnot*)pdfObject)->type = (value);

		}
		else if(strcmp(type, "text_annotation") == 0) {
			pdfObject = new PdfTextAnnot();

			((PdfAnnot*)pdfObject)->type = "Text";

			value = stack[level]->getAttribute("open");
			if(value != NULL) {
				if(strcmp(value, "true") == 0)
					((PdfTextAnnot*)pdfObject)->isOpen = true;
				else if(strcmp(value, "false") == 0)
					((PdfTextAnnot*)pdfObject)->isOpen = false;
				else
					printerr("Error - unknown value for text_annotation Open value\n");
			}
		}
		else if(strcmp(type, "link") == 0) {
			pdfObject = new PdfLink(root);

			((PdfAnnot*)pdfObject)->type = "Link";

			value = stack[level]->getAttribute("page");
			if(value == NULL) 
				printerr("Error - links must specify a page\n");
			((PdfLink*)pdfObject)->destKey = value;

			root->addLink( ((PdfLink*)pdfObject) );
		}
		else {
			printerr("Error - unknown annotation type %s\n", type);
		}

		value = stack[level]->getAttribute("llx");
		if(value == NULL) 
			printerr("HL Error - annotations must have a llx coordinate\n");
		((PdfAnnot*) pdfObject)->rect[0] = atof(value);

		value = stack[level]->getAttribute("lly");
		if(value == NULL) 
			printerr("HL Error - annotations must have a lly coordinate\n");
		((PdfAnnot*) pdfObject)->rect[1] = atof(value);

		value = stack[level]->getAttribute("urx");
		if(value == NULL) 
			printerr("HL Error - annotations must have a urx coordinate\n");
		((PdfAnnot*) pdfObject)->rect[2] = atof(value);

		value = stack[level]->getAttribute("ury");
		if(value == NULL) 
			printerr("HL Error - annotations must have a ury coordinate\n");
		((PdfAnnot*) pdfObject)->rect[3] = atof(value);

		((PdfPage*)currentObj)->addAnnotation( (PdfAnnot*) pdfObject );
	}
	else if(strcmp(type, "media_box") == 0) {
		if(currentObj->objType() != p_page) 
			printerr("HL Error - a media_box must be contained within a page\n");

		double *mBox = ((PdfPage*)currentObj)->mediaBox;

		value = stack[level]->getAttribute("llx");
		if(value == NULL) 
			printerr("HL Error - media_boxes must have a llx coordinate\n");
		mBox[0] = atof(value);

		value = stack[level]->getAttribute("lly");
		if(value == NULL) 
			printerr("HL Error - media_boxes must have a lly coordinate\n");
		mBox[1] = atof(value);

		value = stack[level]->getAttribute("urx");
		if(value == NULL) 
			printerr("HL Error - media_boxes must have a urx coordinate\n");
		mBox[2] = atof(value);

		value = stack[level]->getAttribute("ury");
		if(value == NULL) 
			printerr("HL Error - media_boxes must have a ury coordinate\n");
		mBox[3] = atof(value);
	}
	else {
		printerr("\nERROR - Unknown XML type: %s\n", type);
	}

	if(pdfObject != NULL) {
		stack[level]->mapsToElement = true;

		pdfObject->parent = currentObj;
		currentObj = pdfObject;
	}
}

//-----------------------------------

void XMLModule_highLevel::endElement(XMLElement** stack, int level) {
	char *value;

	if(strcmp(stack[level]->name, "contents") == 0) {
		inContentSection = false;
		if(DEBUG_OUTPUT) printf("Finished parsing Contents with low level XMLModule\n");
	
		if(currentObj->objType() != p_page) 
			printerr("HL Error - content must be contained within a page\n");

		((PdfPage*)currentObj)->contentObj = contentModule.getRoot();
		((PdfPage*)currentObj)->contentObj->parentPage=((PdfPage*)currentObj);
		contentModule.reset(false);
	}
	if(inContentSection) {
		contentModule.endElement(stack, level);
		return;
	}

	if(strcmp(stack[level]->name, "font_data") == 0 && 
	   currentObj != NULL && 
	   currentObj->objType() == p_font && 
	   ((PdfResource*)currentObj)->filename.getlength() > 0) {
	}
	else if(strcmp(stack[level]->name, "image") == 0 && 
	   currentObj != NULL && 
	   currentObj->objType() == p_image && 
	   ((PdfResource*)currentObj)->filename.getlength() > 0) {
	}
	else if(stack[level]->cdata.getlength() > 0) {
		if(strcmp(stack[level]->name, "width") == 0) {
		if(DEBUG_LOG_NULLTERM) printf("nullterm() #6.\n");
			char *data = stack[level]->cdata.nullterm();

			if(currentObj->objType() != p_font) 
				printerr("HL Error - a font_widths must be contained within a font\n");
			
			if(((PdfFont*)currentObj)->subType != "TrueType" || ((PdfFont*)currentObj)->filename.getlength() < 1) {
				if( (((PdfFont*)currentObj)->lastChar - ((PdfFont*)currentObj)->firstChar + 1) <= ((PdfFont*)currentObj)->numWidths)
					printerr("HL Error - too many widths in font_widths array\n");

				((PdfFont*) currentObj)->widths[((PdfFont*)currentObj)->numWidths++] = atoi(data);
			}
			delete [] data;
		}
		else if( (currentObj->objType() == p_page && strcmp(stack[level]->name, "contents") == 0) ||
		        currentObj->objType() == p_image ||
		        currentObj->objType() == p_colorspace ||
		        (currentObj->objType() == p_font && strcmp(stack[level]->name, "font_data") == 0) ||
				currentObj->objType() == p_function ) {
			String dataStr = stack[level]->cdata;

			value = stack[level]->getAttribute("encoding");
			if(value != NULL) {
				if(strcmp(value, "Base85") == 0)
					dataStr = DataFilter::decode85(dataStr);
				else if(strcmp(value, "Hex") == 0)
            		dataStr = DataFilter::decodeHex(dataStr);
				else
					printerr("HL Error - unknown encoding type - %s\n", value);
			}
	
			if(strcmp(stack[level]->name, "contents") == 0) {
				if(currentObj->objType() != p_page) 
					printerr("HL Error - contents must be contained within pages\n");
	
				// ((PdfPage*)currentObj)->content = dataStr;
			}
			else if( currentObj->objType() == p_image || 
			         currentObj->objType() == p_font || 
			         currentObj->objType() == p_colorspace ||
					 currentObj->objType() == p_function) {
				((PdfResource*)currentObj)->dataLength = dataStr.getlength();
				((PdfResource*)currentObj)->data = new char[dataStr.getlength()];
				memcpy( ((PdfResource*)currentObj)->data, dataStr.getrawstring(), dataStr.getlength() );
			}
		}
	}

	if(strcmp(stack[level]->name, "image") == 0) {
		if(currentObj == NULL || currentObj->objType() != p_image)
			printerr("Error - problem parsing image object\n");

		if( ((PdfImage*) currentObj)->format == "external") {
			PdfImage *pdfImage = (PdfImage*) currentObj;

    		ImageReader iReader;
    		iReader.readImage(pdfImage);
		}
	}
	else if(strcmp(stack[level]->name, "font") == 0) {
		if(currentObj == NULL || currentObj->objType() != p_font)
			printerr("Error - problem parsing font object\n");

		PdfFont *pdfFont = (PdfFont*) currentObj;

		if(pdfFont->subType == "TrueType" && pdfFont->filename.getlength() > 0) {
			pdfFont->loadData();
		}

		if(pdfFont->subType == "TrueType" && pdfFont->dataLength > 0) {
			FontReader::readFont( pdfFont );
		}
		else {
			value = stack[level]->getAttribute("font_name");
			if(value == NULL) 
				printerr("HL Error - fonts must specify a font_name\n");
			pdfFont->fontName = value;

			if(pdfFont->isBase14Font())
				pdfFont->initBase14Font();
		}
	}
	else if(strcmp(stack[level]->name, "font_widths") == 0) {
		if(currentObj->objType() != p_font) 
			printerr("HL Error - a font_widths must be contained within a font\n");
			
		if(((PdfFont*)currentObj)->subType != "TrueType" || ((PdfFont*)currentObj)->filename.getlength() < 1) {
			if(((PdfFont*) currentObj)->numWidths != (((PdfFont*) currentObj)->lastChar - ((PdfFont*) currentObj)->firstChar + 1))
				printerr("HL Error - wrong number of widths in font_widths array\n");
		}
	}
	else if(strcmp(stack[level]->name, "text_annotation") == 0) {
		if(currentObj->objType() != p_text_annot)
			printerr("Error parsing annotation");
		if(stack[level]->cdata.getlength() < 1) 
			printerr("HL Error - text_annotations must have a contents\n");

		((PdfTextAnnot*)currentObj)->contents = stack[level]->cdata;
	}
	else if(strcmp(stack[level]->name, "pdf") == 0) {
		PdfLinkList *lptr = root->linkList;
		PdfLink *link;
		PdfPage *page;

		while(lptr != NULL) {
			link = lptr->link;
			page = (PdfPage*) root->pageHash.get(link->destKey);
			if(page == NULL) 
				printerr("Error - invalid link id\n");
			
			link->destination = page;

			lptr = lptr->next;
		}
		if(root->linkList != NULL) {
			delete root->linkList;
			root->linkList = NULL;
		}

		PdfBookmark *bptr = root->bookmarks;
		while(bptr != NULL) {
			bptr->findPages(root);
			bptr = bptr->next;
		}

		PdfList *cptr = root->colorSpaces;
		while(cptr != NULL) {
			// fprintf(stderr, "Found Indexed colorspace\n");
			if( ((PdfColorSpace*)cptr->object)->catagory == "Indexed") {
				if( ((PdfColorSpace*)cptr->object)->base == NULL) {
					 ((PdfColorSpace*)cptr->object)->base = root->getColorSpace(((PdfColorSpace*)cptr->object)->baseId);
					 ((PdfColorSpace*)cptr->object)->baseId = 0;

					if( ((PdfColorSpace*)cptr->object)->base == NULL)
						printerr("Error - Indexed colorspaces must have a base colorSpace\n");
				}
			}
			else if( ((PdfColorSpace*)cptr->object)->catagory == "Separation") {
				if( ((PdfColorSpace*)cptr->object)->altSpace == NULL) {
					 ((PdfColorSpace*)cptr->object)->altSpace = root->getColorSpace(((PdfColorSpace*)cptr->object)->altSpaceId);
					 ((PdfColorSpace*)cptr->object)->altSpaceId = 0;

					if( ((PdfColorSpace*)cptr->object)->altSpace == NULL)
						printerr("Error - Indexed colorspaces must have a alt_space colorSpace\n");
				}
				if( ((PdfColorSpace*)cptr->object)->tintTransform == NULL) {
					 ((PdfColorSpace*)cptr->object)->tintTransform = root->getFunction(((PdfColorSpace*)cptr->object)->tintTransformId);
					 ((PdfColorSpace*)cptr->object)->tintTransformId = 0;

					if( ((PdfColorSpace*)cptr->object)->tintTransform == NULL)
						printerr("Error - Indexed colorspaces must have a tint_transform colorSpace\n");
				}
			}

			cptr = cptr->next;
		}

		PdfResource *rptr = root->resources;
		while(rptr != NULL) {
			if(rptr->type == "XObject") {
				if( ((PdfImage*)rptr)->colorSpace == NULL) {
					((PdfImage*)rptr)->colorSpace = root->getColorSpace(((PdfImage*)rptr)->colorSpaceId);
					((PdfImage*)rptr)->colorSpaceId = 0;
				}
				if( ((PdfImage*)rptr)->colorSpace == NULL) {
					printerr("Error - images must specify a valid colorspace\n");
				}
			}
			rptr = rptr->next;
		}
	}

	if(currentObj != NULL && stack[level]->mapsToElement) {
		currentObj = currentObj->parent;
	}
}

void XMLModule_highLevel::splitNumsIntoArray(double *array, int numVals, char *str) {
	char *numStr = str;
	char *ptr = numStr;
	double *numPtr = array;

	while(*numStr != '\0') {
		while(*ptr != '\0' && *ptr != ',')
			ptr++;
		if(*ptr == ',') {
			*ptr++ = '\0';
		}

		*numPtr++ = atoi(numStr);
		numStr = ptr;
	}
}

