#include "XMLReader.hpp"

#include <stdlib.h>
#include <stdio.h>

#include "parser.h"
#include "parserInternals.h"

#include "pdfc_base.h"
#include "PdfObjects.hpp"
#include "XMLModule.hpp"
#include "XMLModule_lowLevel.hpp"
#include "XMLModule_highLevel.hpp"

xmlSAXHandler SAXHandlerStruct = {
	NULL, /* internalSubset */
	NULL, /* isStandalone */
	NULL, /* hasInternalSubset */
	NULL, /* hasExternalSubset */
	NULL, /* resolveEntity */
	NULL, /* getEntity */
	NULL, /* entityDecl */
	NULL, /* notationDecl */
	NULL, /* attributeDecl */
	NULL, /* elementDecl */
	NULL, /* unparsedEntityDecl */
	NULL, /* setDocumentLocator */
	NULL, /* startDocument */
	NULL, /* endDocument */
	startElement, /* startElement */
	endElement, /* endElement */
	NULL, /* reference */
	characters, /* characters */
	NULL, /* ignorableWhitespace */
	NULL, /* processingInstruction */
	NULL, /* comment */
	warning, /* xmlParserWarning */
	error, /* xmlParserError */
	fatal, /* xmlParserError */
	NULL, /* getParameterEntity */
	NULL, /* cdataBlock; */
}; 

xmlSAXHandlerPtr SAXHandler = &SAXHandlerStruct;

extern int xmlDoValidityCheckingDefaultValue;

//------------------------------------------------

void startElement(void *handlerPtr, const xmlChar *name, const xmlChar **atts)
{
	XMLHandler *handler = (XMLHandler*) handlerPtr;
 
	// printf("SAX.startElement(%s", (char *) name);
	// if (atts != NULL) {
		// for (i = 0;(atts[i] != NULL);i++) {
			// printf(", %s='", atts[i++]);
			// printf("%s'", atts[i]);
		// }
	// }
	// printf(")\n");
 
	handler->startElement(name, atts);
} 

void endElement(void *handlerPtr, const xmlChar *name)
{
	XMLHandler *handler = (XMLHandler*) handlerPtr;
 
	// printf("SAX.endElement(%s)\n", (char *) name);
	handler->endElement(name);
}
 
void characters(void *handlerPtr, const xmlChar *ch, int len)
{
	XMLHandler *handler = (XMLHandler*) handlerPtr;
	// int i;
 
	// printf("SAX.characters(");
	// for (i = 0;(i < len) && (i < 30);i++)
		// printf("%c", ch[i]);
	// printf(", %d)\n", len);
 
	handler->characters(ch, len);
} 

//==========================================================

void warning(void *ctx, const char *msg, ...) {
	printerr("XML Warning: %s\n", msg);
}

void error(void *ctx, const char *msg, ...) {
	printerr("XML Error: %s\n", msg);
}

void fatal(void *ctx, const char *msg, ...) {
	printerr("XML Fatal: %s\n", msg);
}

//==========================================================

void XMLReader::parseXmlFilePointer(XMLHandler *handler, FILE *fp) {
	if (fp != NULL) {
		int res;
		char chars[XMLREADER_BUFFER_SIZE];
		xmlParserCtxtPtr ctxt;
 
		res = fread(chars, 1, XMLREADER_BUFFER_SIZE, fp);
		if (res > 0) {
			ctxt = xmlCreatePushParserCtxt(SAXHandler, (void*)handler, chars, res, NULL);
			if(ctxt == NULL)
				printerr("ctxt is null in XMLReader::parseXmlFilePointer.\n");

			while ((res = fread(chars, 1, XMLREADER_BUFFER_SIZE, fp)) > 0) {
				xmlParseChunk(ctxt, chars, res, 0);
			}
			res = xmlParseChunk(ctxt, chars, 0, 1);
			xmlFreeParserCtxt(ctxt);
			if (res != 0) {
				printerr("xmlSAXUserParseFile returned error %d\n", res);
			}
		}
	}
}

void XMLReader::parseXmlFile(XMLHandler *handler, char *filename) {
	xmlParserCtxtPtr ctxt;

	ctxt = xmlCreateFileParserCtxt(filename);
	if (ctxt == NULL) 
		printerr("ctxt is null in XMLReader::parseXmlFile.\n");

	ctxt->sax = SAXHandler;
	ctxt->userData = (void*)handler;

	xmlParseDocument(ctxt);

	if (! ctxt->wellFormed)
		printerr("Error - not well formed\n");

	ctxt->sax = NULL;
	ctxt->userData = NULL;
// HACK: test comment out next line.  libxml blows up under Windows if xmlFreeParserCtxt is called.
	//xmlFreeParserCtxt(ctxt);
}

void XMLReader::parseXmlMemory(XMLHandler *handler, char *buffer, long bufferLength) {
	xmlParserCtxtPtr ctxt;

	ctxt = xmlCreateMemoryParserCtxt(buffer, bufferLength);
	if (ctxt == NULL)  {
		printerr("ctxt is null in XMLReader::parseXmlMemory.\n");
	}

	ctxt->sax = SAXHandler;
	ctxt->userData = (void*)handler;

	xmlParseDocument(ctxt);

	if (! ctxt->wellFormed)
		printerr("Error - not well formed\n");

	ctxt->sax = NULL;
	xmlFreeParserCtxt(ctxt);
}

//-----------------------------------------------------------

