#include <stdio.h>
#include <stdarg.h> // For variable argument list functions (only used in printerr()).
#include "pdfc_errors.h"

Err::Err(uint newcode): code(newcode), message(NULL), next(NULL) {}

Err::Err(uint newcode, char* newmessage): code(newcode), message(newmessage), next(NULL) {}

Err::Err(uint newcode, char* newmessage, Err* nexterror): code(newcode), message(newmessage), next(nexterror) {}

Err::~Err() {
	// if(message) delete [] message; // Intentionally not doing this; message normally points to string literals.
	if(next) delete next;
	message = NULL;
}

void Err::print(FILE* destination) {
	fprintf(destination, "Error code 0x%.8X: %s\n", code, message==NULL?"(no descriptive error message available.)":message);
	if(next) {
		fprintf(destination, "Encapsulated error: ");
		next->print(destination);
	}
}

FILE* errstream=stderr;

void printerr_getbufferspace(char*& buf) {
	buf=new char[1024];
	if(buf==NULL) {
		fprintf(errstream, "FATAL error: failed to allocate a buffer for an error message in printerr().\n");
		exit(1);
	}
}

void printerr_guts(int code, const char* first, va_list& args) {
	char* buf;
	switch(code&ESEVERITY) {
	case EMINORWARN:
EMW:
		vfprintf(errstream, first, args);
		break;
	case ESTRICTSYNTAX:
		if(ESTRICTSYNTAX_IS_FATAL) goto EF;
		else goto EMW;
	case EFATAL:
EF:
		vfprintf(errstream, first, args);
		exit(1);
		// printerr_getbufferspace(buf);
		// vsprintf(buf, first, args);
		// throw new Err(code, buf);
	default:
		printerr_getbufferspace(buf);
		vsprintf(buf, first, args);
		throw new Err(EFATAL|EOTHER|EINTERNALERROR, "Error in printerr(): invalid severity code.\n",
			new Err(code, buf));
	}
}

void printerr(int code, const char* first, ... ) {
	va_list args;
	va_start(args, first);
	printerr_guts(code, first, args);
	va_end(args);
}

void printerr(const char* first, ... ) {
	va_list args;
	va_start(args, first);
	printerr_guts(EFATAL|EOTHER_GENERICERROR, first, args);
	va_end(args);
}

#ifdef REDEFINE_NEW_OPERATOR

void* operator new(size_t size) {
	return malloc(size);
}

void operator delete(void* ptr) {
	free(ptr);
}

#endif
