#include "stringformatter.h"

#define DEBUG false

uint StringFormatter::countCharsThatFit(char* source, PdfFont* font, uint boxwidth, bool wordwrap, int fontSize) {
	uint currcharwidth;

	 // Convert from points to thousands of an EM-Square (the latter is the native unit for font character width info).
	boxwidth*=1000;
	boxwidth=(boxwidth/fontSize);
	// boxwidth /= currentpage->textmatrix_x;

	uint linelength=0;
	uint charsThatFit=0;
	char* lastDiscoveredSpace=NULL;
	if(source==NULL) return charsThatFit;
	for(char* c=source; true; c++) {
		if((*c=='\0')||(*c=='\n')) {
			return charsThatFit; // Reached end of source or a newline.
		}
		if(*c==' ') 
			lastDiscoveredSpace=c;

		if(*c < font->firstChar || *c > font->lastChar) 
			currcharwidth = (uint) (font->fontBBox[2] - font->fontBBox[0]);
		else 
			currcharwidth = font->widths[*c - font->firstChar];

		linelength+=currcharwidth;

		if(linelength>boxwidth) {
			if(wordwrap) {
				if(lastDiscoveredSpace!=NULL) return lastDiscoveredSpace-source+1;
				else return charsThatFit; // Current space-delimited word was too big to fit on one line,
					// so it must be broken up, despite the setting of wordwrap.
			} else return charsThatFit;
		}
		charsThatFit++;
	}
}

String StringFormatter::wrapstring(char* source, PdfFont* font, uint boxwidth, uint boxheight, bool wordwrap, char*& leftovers, int fontSize) {
	if(DEBUG) printf("wrapstring(): %s\n", source);

	// boxwidth*=currentpage->textmatrix_x; // Convert units to points (assuming user space is currently in points).
	// boxheight*=currentpage->textmatrix_y;
	leftovers=NULL;
	String output;
	if(source==NULL) return output;
	if(font==NULL) printerr(EFATAL|EPDFOUT|EINTERNALERROR, "null font passed to wrapstring().\n");
	if((boxwidth<1)||(boxheight<1)) printerr(EFATAL|EPDFOUT|EARGOUTOFRANGE, "boxwidth and boxheight cannot be zero.\n");

	if((fontSize)==0) printerr(EFATAL|EPDFOUT|EDIVBYZERO, "attempted to divide by a current font size of zero.\n");
	uint lines=boxheight/(fontSize);
	char* currchar=source;
	for(uint c=0; c<lines; c++) {
		uint chars=countCharsThatFit(currchar, font, boxwidth, wordwrap, fontSize);
		output.append(currchar, chars); // Append processed line to output.
		output.append("\n");
		currchar+=chars;
		if(*currchar=='\n') currchar++; // countCharsThatFit() doesn't count newlines toward the total that fits,
			// so manually advance past the newline here if that was the last character on the line that fit.
		if(*currchar=='\0') return output; // Finished processing entire source.
	}
	leftovers=currchar; // for() loop above didn't finish processing all of the source.
	return output;
}

String StringFormatter::stripCR(char* source) {
	if(DEBUG) printf("stripCR(): %s\n", source);
	String output;
	if(source==NULL) return output;
	char prevchar='\0';
	for(char* c=source; *c!='\0'; c++) {
		if(prevchar==0xd) {
			if(*c!=0xd) output.append("\n"); // Saw a CR followed by a non-CR.
			else output.append("\n\n"); // Saw a CR followed by a CR.
		} else if(*c!=0xd) output.append(c, 1); // Directly append everything except CR.
		prevchar=*c;
	}
	return output;
}

PObj* StringFormatter::renderLF(char* source, int x, int y, PdfPage* currentpage, String fontId, int fontSize) {
	if(DEBUG) printf("renderLF(): %s\n", source);

	// x*=currentpage->textmatrix_x;
	// y*=currentpage->textmatrix_x;

	if(source==NULL) return NULL;
	String currline;
	char* currchar=source;
	PDF_collection *retval = new PDF_collection();

	// Move to the specified coordinates.
	PDF_pscommand* p=new PDF_pscommand;
	p->setValue("BT"); // Begin a new text object.
	retval->addsub(p);

	PDF_int* i=NULL;

	p=new PDF_pscommand;
	p->setValue("Tm"); // Set the text position.
	retval->addsub(p);

	i=new PDF_int;
	i->value=1;
	p->addsub(i);
	i=new PDF_int;
	i->value=0;
	p->addsub(i);
	i=new PDF_int;
	i->value=0;
	p->addsub(i);
	i=new PDF_int;
	i->value=1;
	p->addsub(i);
	i=new PDF_int;
	i->value=0;
	p->addsub(i);
	i=new PDF_int;
	i->value=0;
	p->addsub(i);

	p=new PDF_pscommand;
	p->setValue("Tf"); // Set the text position.
	retval->addsub(p);

	PDF_name *str=new PDF_name(fontId);
	p->addsub(str);

	i=new PDF_int;
	i->value=fontSize;
	p->addsub(i);
/*
	p=new PDF_pscommand;
	p->setValue("k"); // Set the color.
	retval->addsub(p);

	// Set the four arguments to command "k".
	i=new PDF_int;
	i->value=0;
	p->addsub(i);
	i=new PDF_int;
	i->value=0;
	p->addsub(i);
	i=new PDF_int;
	i->value=0;
	p->addsub(i);
	i=new PDF_int;
	i->value=1;
	p->addsub(i);
*/

	p=new PDF_pscommand;
	p->setValue("Td"); // Set the text position.
	retval->addsub(p);

	// Set the two arguments to command "Td".
	i=new PDF_int;
	i->value=x;
	p->addsub(i);
	i=new PDF_int;
	i->value=y-fontSize; // y coordinate specifies top of box, but text is drawn with the current
		// position taken as the _bottom_ of the line of text.
	p->addsub(i);

	while(true) {
		currline=String();
		while(((*currchar)!='\0')&&((*currchar)!=0xa)) { // Load the next LF-delimited line into currline.
			currline.append(currchar, 1);
 			currchar++;
		}
		
		// And now, append a new show_text_string command to the end of the command list.
		p=new PDF_pscommand;
		p->setValue("show_text_string");
		retval->addsub(p);

		PDF_string* s=new PDF_string;
		s->setValue(currline);
		p->addsub(s);

		if(*currchar==0xa) { // Reached a newline.
			// And now, append a new newline command to the end of the command list.
			p=new PDF_pscommand;
			p->setValue("newline");
			retval->addsub(p);

			i=new PDF_int;
			i->setNumericValue(0); // First argument to pscommand "newline".
			p->addsub(i);

			i=new PDF_int;
			// i->setNumericValue(0-(fontSize+currentpage->currentleading)); // Second argument to pscommand "newline".  Indicates move DOWN one text line.
			i->setNumericValue(0-fontSize); // Second argument to pscommand "newline".  Indicates move DOWN one text line.
			p->addsub(i);

		} else if(*currchar==0) 
			break; // Finished processing all of source (reached null terminator).
		else 
			printerr(EFATAL|EPDFOUT|EINTERNALERROR, "Bug in renderLF()!  Expecting a newline or null terminator, but found something else.\n");
		currchar++;

		if(*currchar == '\0')
			break;
	}

	p=new PDF_pscommand;
	p->setValue("ET"); // End the text object.
	retval->addsub(p);

	return retval;
}

int StringFormatter::getUnwrappedStringWidth(String text, PdfFont *font, int size) {
	int maxWidth = 0;
	int currentWidth = 0;
	int strLen = text.getlength();
	int i;
	const char *str = text.getrawstring();

	for(i=0; i<strLen; i++) {
		if(str[i] == '\n') {
			if(maxWidth < currentWidth)
				maxWidth = currentWidth;
			currentWidth = 0;
		}
		else {
			if(str[i] < font->firstChar || str[i] > font->lastChar)
				currentWidth = (int) (font->fontBBox[2] - font->fontBBox[0]);
			else
				currentWidth = font->widths[str[i] - font->firstChar];
		}
	}
	if(maxWidth < currentWidth)
		maxWidth = currentWidth;

	maxWidth = size * maxWidth;
	if((maxWidth % 1000) != 0)
		maxWidth += 1000;

	maxWidth /= 1000;
	return maxWidth;
}

int StringFormatter::getWrappedStringHeight(String text, PdfFont *font, int size, int width) {
	int scaledWidth = width * 1000 / size;
	int currentWidth = 0;
	int cWidth = 0;
	int strLen = text.getlength();
	int i;
	const char *str = text.getrawstring();
	int numLines = 0;
	int startOfLine = 0;
	int lastSpace = 0;

	for(i=0; i<strLen; i++) {
		if(str[i] == '\n') {
			startOfLine = i+1;
			currentWidth = 0;
			numLines++;
		}
		else {
			if(str[i] < font->firstChar || str[i] > font->lastChar)
				cWidth = (int) (font->fontBBox[2] - font->fontBBox[0]);
			else
				cWidth = font->widths[str[i] - font->firstChar];

			if(str[i] == ' ') {
				lastSpace = i;
			}
			else if(currentWidth + cWidth > scaledWidth) {
				if(lastSpace > startOfLine) {
					i = lastSpace;
					startOfLine = i+1;
					cWidth = 0;
				}
				else {
					startOfLine = i;
				}

				currentWidth = 0;
				numLines++;
			}
			currentWidth += cWidth;
		}
	}
	numLines++;

	return numLines*size;
}

