// pdfcwrapper.cpp : Implementation of Cpdfcwrapper
#include "stdafx.h"
#include "LibpdfcCOM.h"
#include "pdfcwrapper.h"

/////////////////////////////////////////////////////////////////////////////
// Cpdfcwrapper

STDMETHODIMP Cpdfcwrapper::OnStartPage (IUnknown* pUnk)  
{
	if(!pUnk)
		return E_POINTER;

	CComPtr<IScriptingContext> spContext;
	HRESULT hr;

	// Get the IScriptingContext Interface
	hr = pUnk->QueryInterface(IID_IScriptingContext, (void **)&spContext);
	if(FAILED(hr))
		return hr;

	// Get Request Object Pointer
	hr = spContext->get_Request(&m_piRequest);
	if(FAILED(hr))
	{
		spContext.Release();
		return hr;
	}

	// Get Response Object Pointer
	hr = spContext->get_Response(&m_piResponse);
	if(FAILED(hr))
	{
		m_piRequest.Release();
		return hr;
	}
	
	// Get Server Object Pointer
	hr = spContext->get_Server(&m_piServer);
	if(FAILED(hr))
	{
		m_piRequest.Release();
		m_piResponse.Release();
		return hr;
	}
	
	// Get Session Object Pointer
	hr = spContext->get_Session(&m_piSession);
	if(FAILED(hr))
	{
		m_piRequest.Release();
		m_piResponse.Release();
		m_piServer.Release();
		return hr;
	}

	// Get Application Object Pointer
	hr = spContext->get_Application(&m_piApplication);
	if(FAILED(hr))
	{
		m_piRequest.Release();
		m_piResponse.Release();
		m_piServer.Release();
		m_piSession.Release();
		return hr;
	}
	m_bOnStartPageCalled = TRUE;
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::OnEndPage ()  
{
	m_bOnStartPageCalled = FALSE;
	// Release all interfaces
	m_piRequest.Release();
	m_piResponse.Release();
	m_piServer.Release();
	m_piSession.Release();
	m_piApplication.Release();

	return S_OK;
}


STDMETHODIMP Cpdfcwrapper::pdf2xml(BSTR infilename, BSTR outfilename)
{
	String s_ifile=String(infilename);
	char* c_ifile=s_ifile.nullterm();
	String s_ofile=String(outfilename);
	char* c_ofile=s_ofile.nullterm();

	FileReader data(c_ifile);
	root=preader.parse(&data);
	pdfRoot=objParser.parse(root);

	FILE* ofilehandle=fopen(c_ofile, "wb");
	if(ofilehandle==NULL) throw new Err(EFATAL|EPDFIN|EOTHER, "unable to open output file");
	char* theoutput=pdfRoot->render_xml(0).nullterm();
	fprintf(ofilehandle, "%s", theoutput);
	delete [] theoutput;
	fclose(ofilehandle);

	delete [] c_ifile;
	delete [] c_ofile;

	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::xml2pdf(BSTR infilename, BSTR outfilename)
{
	String s_ifile=String(infilename);
	char* c_ifile=s_ifile.nullterm();
	String s_ofile=String(outfilename);
	char* c_ofile=s_ofile.nullterm();

	XMLModule_highLevel module;
	XMLHandler handler(&module);
	XMLReader::parseXmlFile(&handler, c_ifile);
	pdfRoot=module.getRoot();

	PObj *rootObj = pdfRoot->renderBasicObjects();
	if(rootObj!=NULL) {
		if(root!=NULL)
			delete root;
		root=rootObj;
	}
	else {
		throw new Err(EFATAL|EPDFOUT|EOTHER, "Error - could not convert high level representation to low level");
	}
	delete pdfRoot;

	String r;
	root->render_pdf(NULL, r);
	FILE* ofilehandle=fopen(c_ofile, "wb");
	if(ofilehandle==NULL) throw new Err(EFATAL|EPDFOUT|EOTHER, "unable to open output file");
	uint written=fwrite((void*)(r.getrawstring()), 1, r.getlength(), ofilehandle);
	fclose(ofilehandle);
	if(written!=r.getlength()) {
		throw new Err(EFATAL|EPDFOUT|EOTHER, "Error writing to output file");
		exit(1);
	}

	delete [] c_ifile;
	delete [] c_ofile;

	return S_OK;
}

/////////////////////////////////////////////////////////////////////////////

STDMETHODIMP Cpdfcwrapper::closepath_eofill_and_stroke(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::closepath_eofill_and_stroke();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::closepath_fill_and_stroke(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::closepath_fill_and_stroke();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::begin_marked_content_with_propertylist(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::begin_marked_content_with_propertylist();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::eofill_and_stroke(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::eofill_and_stroke();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::fill_and_stroke(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::fill_and_stroke();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::begin_image(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::begin_image();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::begin_marked_content(IUnknown** objRef, BSTR name0) {
	*objRef = (IUnknown*) PsCommands::begin_marked_content(name0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::begin_text_object(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::begin_text_object();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::begin_hiding_undefined_operators(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::begin_hiding_undefined_operators();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::appendbeziercurve(IUnknown** objRef, double num0, double num1, double num2, double num3, double num4, double num5) {
	*objRef = (IUnknown*) PsCommands::appendbeziercurve(num0, num1, num2, num3, num4, num5);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::concatenate_CTM_matrix(IUnknown** objRef, double num0, double num1, double num2, double num3, double num4, double num5) {
	*objRef = (IUnknown*) PsCommands::concatenate_CTM_matrix(num0, num1, num2, num3, num4, num5);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_fillcolorspace(IUnknown** objRef, BSTR name0) {
	*objRef = (IUnknown*) PsCommands::set_fillcolorspace(name0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_strokecolorspace(IUnknown** objRef, BSTR name0) {
	*objRef = (IUnknown*) PsCommands::set_strokecolorspace(name0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_character_width(IUnknown** objRef, double num0, double num1) {
	*objRef = (IUnknown*) PsCommands::set_character_width(num0, num1);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_cache_device(IUnknown** objRef, double num0, double num1, double num2, double num3, double num4, double num5) {
	*objRef = (IUnknown*) PsCommands::set_cache_device(num0, num1, num2, num3, num4, num5);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::execute_xobject(IUnknown** objRef, BSTR xobject0) {
	*objRef = (IUnknown*) PsCommands::execute_xobject(xobject0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::mark_point_with_propertylist(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::mark_point_with_propertylist();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_dashpattern(IUnknown** objRef, IUnknown* array0, double num1) {
	*objRef = (IUnknown*) PsCommands::set_dashpattern((PDF_array*)array0, num1);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::end_image(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::end_image();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::end_marked_content(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::end_marked_content();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::end_text_object(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::end_text_object();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::end_hiding_undefined_operators(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::end_hiding_undefined_operators();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::fillpath(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::fillpath();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::fillpath_oddeven(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::fillpath_oddeven();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_grayfilltint(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_grayfilltint(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_graystroketint(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_graystroketint(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_graphicsstate(IUnknown** objRef, BSTR name0) {
	*objRef = (IUnknown*) PsCommands::set_graphicsstate(name0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::closecurrentsubpath(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::closecurrentsubpath();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::begin_image_data(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::begin_image_data();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_flatness(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_flatness(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_linecap(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_linecap(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_linejoin(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_linejoin(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_fillcolorCMYK(IUnknown** objRef, double num0, double num1, double num2, double num3) {
	*objRef = (IUnknown*) PsCommands::set_fillcolorCMYK(num0, num1, num2, num3);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_strokecolorCMYK(IUnknown** objRef, double num0, double num1, double num2, double num3) {
	*objRef = (IUnknown*) PsCommands::set_strokecolorCMYK(num0, num1, num2, num3);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::lineto(IUnknown** objRef, double num0, double num1) {
	*objRef = (IUnknown*) PsCommands::lineto(num0, num1);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::moveto(IUnknown** objRef, double num0, double num1) {
	*objRef = (IUnknown*) PsCommands::moveto(num0, num1);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::mark_point(IUnknown** objRef, BSTR name0) {
	*objRef = (IUnknown*) PsCommands::mark_point(name0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_miterlimit(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_miterlimit(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::endpath(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::endpath();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::newline_and_setspacing_and_show_text_string(IUnknown** objRef, double num0, double num1, BSTR string2) {
	*objRef = (IUnknown*) PsCommands::newline_and_setspacing_and_show_text_string(num0, num1, string2);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::newline_and_show_text_string(IUnknown** objRef, BSTR string0) {
	*objRef = (IUnknown*) PsCommands::newline_and_show_text_string(string0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::exec_postscript(IUnknown** objRef, BSTR string0) {
	*objRef = (IUnknown*) PsCommands::exec_postscript(string0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::restore_graphics_state(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::restore_graphics_state();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::save_graphics_state(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::save_graphics_state();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::add_rectangle(IUnknown** objRef, double num0, double num1, double num2, double num3) {
	*objRef = (IUnknown*) PsCommands::add_rectangle(num0, num1, num2, num3);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_fillcolorRGB(IUnknown** objRef, double num0, double num1, double num2) {
	*objRef = (IUnknown*) PsCommands::set_fillcolorRGB(num0, num1, num2);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_strokecolorRGB(IUnknown** objRef, double num0, double num1, double num2) {
	*objRef = (IUnknown*) PsCommands::set_strokecolorRGB(num0, num1, num2);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_colorrenderingintent(IUnknown** objRef, BSTR name0) {
	*objRef = (IUnknown*) PsCommands::set_colorrenderingintent(name0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::close_and_strokepath(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::close_and_strokepath();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::SCN_(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::SCN();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::scn(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::scn();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_fillcolor(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::set_fillcolor();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_strokecolor(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::set_strokecolor();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::shade(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::shade(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::strokepath(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::strokepath();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_character_spacing(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_character_spacing(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::newline(IUnknown** objRef, double num0, double num1) {
	*objRef = (IUnknown*) PsCommands::newline(num0, num1);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::newline_and_set_leading(IUnknown** objRef, double num0, double num1) {
	*objRef = (IUnknown*) PsCommands::newline_and_set_leading(num0, num1);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_font_and_size(IUnknown** objRef, BSTR name0, double num1) {
	*objRef = (IUnknown*) PsCommands::set_font_and_size(name0, num1);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::show_text_string(IUnknown** objRef, BSTR string0) {
	*objRef = (IUnknown*) PsCommands::show_text_string(string0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::show_text_string_with_displacements(IUnknown** objRef, IUnknown* array0) {
	*objRef = (IUnknown*) PsCommands::show_text_string_with_displacements((PDF_array*)array0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_text_leading(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_text_leading(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_text_matrix(IUnknown** objRef, double num0, double num1, double num2, double num3, double num4, double num5) {
	*objRef = (IUnknown*) PsCommands::set_text_matrix(num0, num1, num2, num3, num4, num5);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::newline_with_defaultoffset(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::newline_with_defaultoffset();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_text_rendering_mode(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_text_rendering_mode(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_text_size(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_text_size(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_word_spacing(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_word_spacing(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_horizontal_scaling(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_horizontal_scaling(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::appendbeziercurve_current(IUnknown** objRef, double num0, double num1, double num2, double num3) {
	*objRef = (IUnknown*) PsCommands::appendbeziercurve_current(num0, num1, num2, num3);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::clip(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::clip();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::clip_evenodd(IUnknown** objRef) {
	*objRef = (IUnknown*) PsCommands::clip_evenodd();
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::set_linewidth(IUnknown** objRef, double num0) {
	*objRef = (IUnknown*) PsCommands::set_linewidth(num0);
	return S_OK;
}

STDMETHODIMP Cpdfcwrapper::appendbeziercurve_current_secondversion(IUnknown** objRef, double num0, double num1, double num2, double num3) {
	*objRef = (IUnknown*) PsCommands::appendbeziercurve_current_secondversion(num0, num1, num2, num3);
	return S_OK;
}

